/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#ifndef __GAME_H
#define __GAME_H

#include <stdio.h>

#include <set>
#include <list>
#include <vector>
#include <string>

typedef list<string> path_container;

enum merge_t {
	merge_no, // All required roms in every set
	merge_differential, // Only the unique roms in every set
	merge_parent, // All roms in the parent set
	merge_any, // Any merge allowed
	merge_disable // Check disabled
};

// ------------------------------------------------------------------------
// Group

#define GROUP_UNDEFINED "<undefined>"

typedef list<string> group_container;

#define TYPE_UNDEFINED "<undefined>"

typedef list<string> type_container;

// ------------------------------------------------------------------------
// Game

class game {
	// game information
	string name;
	string romof;
	string cloneof;
	string sampleof;
	string description;
	string year;
	string manufacturer;
	bool working;
	bool resource;
	bool vector;
	bool sound;
	bool color;
	unsigned sizex;
	unsigned sizey;

	mutable string group; // nome del gruppo
	mutable string type; // nome del gruppo
	mutable unsigned time; // tempo di gioco in secondi

	mutable unsigned size; // cached size in bytes of the roms
	mutable unsigned clone; // cached number of clones

        mutable path_container rzs; // set of rom zip for the game
        mutable path_container szs; // set of sample zip for the game
public:
	game();
	game(const string& Aname);
	game(const game&);
	~game();

	void name_set(const string& Aname) { name = Aname; }
	void cloneof_set(const string& Acloneof) { cloneof = Acloneof; }
	void romof_set(const string& Aromof) { romof = Aromof; }
	void sampleof_set(const	char* Asampleof) { sampleof = Asampleof; }
	void description_set(const string& Adescription) { description = Adescription; }
	void year_set(const string& Ayear) { year = Ayear; }
	void manufacturer_set(const string& Amanufacturer) { manufacturer = Amanufacturer; }
	void working_set(bool Aworking) { working = Aworking; }
	void resource_set(bool Aresource) { resource = Aresource; }
	void time_set(unsigned Atime) const { time = Atime; }
	void group_set(const string& Agroup) const { group = Agroup; }
	void type_set(const string& Atype) const { type = Atype; }
	void vector_set(bool Avector) { vector = Avector; }
	void size_set(unsigned Asize) const { size = Asize; }
	void clone_set(unsigned Aclone) const { clone = Aclone; }
	void sound_set(bool Asound) { sound = Asound; }
	void color_set(bool Acolor) { color = Acolor; }
	void sizex_set(unsigned A) { sizex = A; }
	void sizey_set(unsigned A) { sizey = A; }

	void rom_zip_set_add(const string& Afile) const;
        void sample_zip_set_add(const string& Afile) const;

	const string& name_get() const { return name; }
	const string& cloneof_get() const { return cloneof; }
	const string& romof_get() const { return romof; }
	const string& sampleof_get() const { return sampleof; }
	const string& description_get() const { return description; }
	const string& year_get() const { return year; }
	const string& manufacturer_get() const { return manufacturer; }
        bool working_get() const { return working; }
	bool resource_get() const { return resource; }
	unsigned time_get() const { return time; }
	string group_get() const { return group; }
	string type_get() const { return type; }
	bool vector_get() const { return vector; }
	unsigned size_get() const { return size; }
	unsigned clone_get() const { return clone; }
	bool sound_get() const { return sound; }
	bool color_get() const { return color; }
	unsigned sizex_get() const { return sizex; }
	unsigned sizey_get() const { return sizey; }

        const path_container& rom_zip_set_get() const { return rzs; }
	const path_container& sample_zip_set_get() const { return szs; }

	bool is_present() const {
		return size_get() == 0 || rom_zip_set_get().size() > 0;
	}
};

struct game_by_name_less : binary_function<game,game,bool> {
	bool operator()(const game& A, const game& B) const {
		return A.name_get().compare(B.name_get()) < 0;
	}
};

typedef	set<game,game_by_name_less> game_by_name_set;
typedef vector<const game*> game_container;

class game_set : public	game_by_name_set {
public:
	typedef game_by_name_set::const_iterator const_iterator;
	typedef game_by_name_set::iterator iterator;

	bool test() const;
	bool load(FILE* f);

	bool is_tree_rom_of_present(const string& name, merge_t type) const;
	bool is_game_neogeo(const string& name) const;
	game_set::const_iterator root_rom_of_get(const string& name) const;
	game_set::const_iterator root_clone_of_get(const string& name) const;
	bool is_game_rom_of(const string& name_son, const string& name_parent) const;
	bool is_game_clone_of(const string& name_son, const string& name_parent) const;
};

struct pgame_by_desc_less : binary_function<game*,game*,bool> {
	pgame_by_desc_less(const game_set&) { }

	bool operator()(const game* A, const game* B) const {
		return A->description_get() < B->description_get();
	}
};

struct pgame_by_rootdesc_less : binary_function<game*,game*,bool> {
	const game_set& gar;

	pgame_by_rootdesc_less(const game_set& Agar) : gar(Agar) { }

	bool operator()(const game* A, const game* B) const {
		return gar.root_clone_of_get(A->name_get())->description_get() < gar.root_clone_of_get(B->name_get())->description_get();
	}
};

struct pgame_by_manufacturer_less : binary_function<game*,game*,bool> {
	pgame_by_manufacturer_less(const game_set&) { }

	bool operator()(const game* A, const game* B) const {
		return A->manufacturer_get() < B->manufacturer_get();
	}
};

struct pgame_by_year_less : binary_function<game*,game*,bool> {
	pgame_by_year_less(const game_set&) { }

	bool operator()(const game* A, const game* B) const {
		return A->year_get() < B->year_get();
	}
};

struct pgame_by_time_less : binary_function<game*,game*,bool> {
	pgame_by_time_less(const game_set&) { }

	bool operator()(const game* A, const game* B) const {
		return A->time_get() > B->time_get();
	}
};

struct pgame_by_group_less : binary_function<game*,game*,bool> {
	pgame_by_group_less(const game_set&) { }

	bool operator()(const game* A, const game* B) const {
		return A->group_get() < B->group_get();
	}
};

struct pgame_by_type_less : binary_function<game*,game*,bool> {
	pgame_by_type_less(const game_set&) { }

	bool operator()(const game* A, const game* B) const {
		return A->type_get() < B->type_get();
	}
};

struct pgame_by_size_less : binary_function<game*,game*,bool> {
	pgame_by_size_less(const game_set&) { }

	bool operator()(const game* A, const game* B) const {
		return A->size_get() > B->size_get();
	}
};

struct pgame_by_clone_less : binary_function<game*,game*,bool> {
	pgame_by_clone_less(const game_set&) { }

	bool operator()(const game* A, const game* B) const {
		return (A->cloneof_get().length()!=0) < (B->cloneof_get().length()!=0);
	}
};


template<typename T, typename T1, typename T2>
struct combine_less : binary_function<T,T,bool> {
	T1 c1;
	T2 c2;

	combine_less(const game_set& gar) : c1(gar), c2(gar) { }

	bool operator()(const T A, const T B) const {
		if (c1(A,B)) return true;
		if (c1(B,A)) return false;
		return c2(A,B);
	}
};

#endif
