/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "text.h"
#include "font.h"
#include "backdrop.h"
#include "screen.h"
#include "allegro.h"

#include <iostream.h>
#include <unistd.h>
#include <math.h>
#include <sys/farptr.h>
#include <sys/segments.h>
#include <sys/stat.h>
#include <conio.h>
#include <pc.h>

#define VIDEO_MAX_X 1600
#define VIDEO_MAX_Y 1280

// Orientation flag
static unsigned text_orientation = 0;

// Text buffer (NOT orientation corrected)
static unsigned text_dx;
static unsigned text_dy;

// Text (ALREADY orientation corrected)
static unsigned real_text_map[32768];
static unsigned real_text_dx;
static unsigned real_text_dy;

// Font (already orientation corrected)
static unsigned real_font_dx;
static unsigned real_font_dy;
static struct bitmapfont* real_font_map;

// Backdrop (already orientation corrected)
static struct bitmap* backdrop;
static char backdrop_path[256];
static double backdrop_brightness;

// RGB palette for 8 bit backdrop
static video_rgb backdrop_color_map[256];
static unsigned	backdrop_index_map[256];

// RGB conversion table
unsigned backdrop_convert[BITMAP_INDEX_MAX];

// Backdrop expansion factor (also used for MIRROR effect)
static unsigned backdrop_x_map[VIDEO_MAX_X];
static unsigned backdrop_y_map[VIDEO_MAX_Y];

// Color used for print backdrop
static unsigned backdrop_color_mask;

// Position of the backdrop in the screen
static int real_backdrop_x;
static int real_backdrop_y;
static int real_backdrop_dx;
static int real_backdrop_dy;

// -----------------------------------------------------------------------
// Joystick

struct { char *name; int id; } joy_table[] =
{
	{ "none",                       JOY_TYPE_NONE },
	{ "auto",                       JOY_TYPE_AUTODETECT },
	{ "standard",           JOY_TYPE_STANDARD },
	{ "dual",                       JOY_TYPE_2PADS },
	{ "4button",            JOY_TYPE_4BUTTON },
	{ "6button",            JOY_TYPE_6BUTTON },
	{ "8button",            JOY_TYPE_8BUTTON },
	{ "fspro",                      JOY_TYPE_FSPRO },
	{ "wingex",                     JOY_TYPE_WINGEX },
	{ "sidewinder",         JOY_TYPE_SIDEWINDER },
	{ "gamepadpro",         JOY_TYPE_GAMEPAD_PRO },
	{ "grip",               JOY_TYPE_GRIP },
	{ "grip4",              JOY_TYPE_GRIP4 },
	{ "sneslpt1",           JOY_TYPE_SNESPAD_LPT1 },
	{ "sneslpt2",           JOY_TYPE_SNESPAD_LPT2 },
	{ "sneslpt3",           JOY_TYPE_SNESPAD_LPT3 },
	{ "psxlpt1",            JOY_TYPE_PSXPAD_LPT1 },
	{ "psxlpt2",            JOY_TYPE_PSXPAD_LPT2 },
	{ "psxlpt3",            JOY_TYPE_PSXPAD_LPT3 },
	{ "n64lpt1",            JOY_TYPE_N64PAD_LPT1 },
	{ "n64lpt2",            JOY_TYPE_N64PAD_LPT2 },
	{ "n64lpt3",            JOY_TYPE_N64PAD_LPT3 },
	{ "wingwarrior",        JOY_TYPE_WINGWARRIOR },
	{ 0, 0 }
};

static int joystick = -1;

static void joystick_init() {
	char* joyname = get_config_string("config", "joystick", 0);
	if (!joyname)
		joyname = get_config_string("config", "joy", "none");

	unsigned i;
	for(i=0;joy_table[i].name;++i)
		if (strcmp(joy_table[i].name,joyname)==0)
			break;

	if (joy_table[i].name) {
		joystick = joy_table[i].id;
		if (install_joystick(joystick)!=0) {
			joystick = -1;
		}
	}
}

static void joystick_done() {
	if (joystick>=0)
		remove_joystick();
}

static int joystick_raw_poll() {
	if (joystick>=0) {
		poll_joystick();

		for(int i=0;i<num_joysticks;++i) {
			for(int j=0;j<joy[i].num_sticks;++j) {
				if (joy[i].stick[j].num_axis > 0) {
					if (joy[i].stick[j].axis[0].d1)
						return TEXT_KEY_LEFT;
					if (joy[i].stick[j].axis[0].d2)
						return TEXT_KEY_RIGHT;
				}
				if (joy[i].stick[j].num_axis > 1) {
					if (joy[i].stick[j].axis[1].d1)
						return TEXT_KEY_UP;
					if (joy[i].stick[j].axis[1].d2)
						return TEXT_KEY_DOWN;
				}
			}
			for(int j=0;j<joy[i].num_buttons;++j) {
				if (joy[i].button[j].b) {
					switch (j) {
						case 0 : return TEXT_KEY_ENTER;
						case 1 : return TEXT_KEY_ESC;
						case 2 : return TEXT_KEY_SPACE;
						case 3 : return TEXT_KEY_MODE;
					}
				}
			}
		}

	}
	return TEXT_KEY_NONE;
}

// -------------------------------------------------------------------------
// Size

__inline__ void swap(unsigned& a, unsigned& b) {
	unsigned t = a;
	a = b;
	b = t;
}

__inline__ void swap(int& a, int& b) {
	int t = a;
	a = b;
	b = t;
}

int text_dx_get() {
	return text_dx;
}

int text_dy_get() {
	return text_dy;
}

// -------------------------------------------------------------------------
// Cursor

void text_cursor(bool visible) {
}

void text_gotoxy(int x, int y) {
}

// -------------------------------------------------------------------------
// Video mode choice

#define DEFAULT_GRAPH_MODE "default_graph" // default video mode
#define DEFAULT_TEXT_MODE "default_text" // default video mode

// comparing for graphics mode
bool mode_graphics_less(const video_mode* A, const video_mode* B) {
	int bitsA = video_mode_bytes_per_pixel(A);
	int bitsB = video_mode_bytes_per_pixel(B);
	if (bitsA < bitsB) return true;
	if (bitsA > bitsB) return false;

	int areaA = A->size_x * A->size_y;
	int areaB = B->size_x * B->size_y;
	int difA = abs( areaA - 640*480 );
	int difB = abs( areaB - 640*480 );
	if (difA > difB) return true;
	if (difA < difB) return false;

	return false;
}

// comparing for text mode
bool mode_text_less(const video_mode* A, const video_mode* B) {
	int areaA = A->size_x * A->size_y;
	int areaB = B->size_x * B->size_y;
	int difA = abs( areaA - 720*400 );
	int difB = abs( areaB - 720*400 );
	return difA > difB;
}

static video_mode mode; // selected video mode
static video_monitor monitor; // monitor info
static video_error monitor_error; // monitor info status

// -------------------------------------------------------------------------
// Inizialization

bool text_init() {
	char* cfg = "mame.cfg";

	bool mode_found = false; // selected video mode found
	video_mode_container vm_sel;
	video_mode_container vm_def;
	video_mode_container_iterator i;

	if (allegro_init() != 0) {
		cerr << "Error inizializing the Allegro library" << endl;
		goto out;
	}
	set_config_file(cfg);

	joystick_init();

	if (video_options_load() != 0) {
		cerr << "Error loading the video options" << endl << video_error_description_get() << endl;
		goto out_joy;
	}

	video_init();

	monitor_error = video_monitor_load(&monitor);
	if (monitor_error < 0) {
		cerr << "Error loading the monitor options" << endl << video_error_description_get() << endl;
		goto out_video;
	}

	// load selected
	video_mode_container_init(&vm_sel);
	video_mode_container_load(&vm_sel);

	// load default
	video_mode_container_init(&vm_def);
	video_mode_container_insert_vga(&vm_def);
	video_mode_container_insert_vbe(&vm_def);

	// seach default mode in selected
	for(video_mode_container_iterator_begin(&i,&vm_sel);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
		const video_mode* vm = video_mode_container_iterator_get(&i);
		if (strcmp(vm->name,DEFAULT_GRAPH_MODE)==0) {
			mode = *vm;
			mode_found = true;
			break;
		}
	}

	// search graphics mode in selected
	if (!mode_found) {
		for(video_mode_container_iterator_begin(&i,&vm_sel);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
			const video_mode* vm = video_mode_container_iterator_get(&i);
			if ((monitor_error>0 || video_monitor_mode_check(&monitor,vm)) && video_mode_is_graphics(vm) && (video_mode_bytes_per_pixel(vm)==1 || video_mode_bytes_per_pixel(vm)==2)) {
				if (!mode_found || mode_graphics_less(&mode,vm)) {
					mode = *vm;
					mode_found = true;
				}
			}
		}
	}

	// search graphics mode in default
	if (!mode_found) {
		for(video_mode_container_iterator_begin(&i,&vm_def);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
			const video_mode* vm = video_mode_container_iterator_get(&i);
			if ((monitor_error>0 || video_monitor_mode_check(&monitor,vm)) && video_mode_is_graphics(vm) && (video_mode_bytes_per_pixel(vm)==1 || video_mode_bytes_per_pixel(vm)==2)) {
				if (!mode_found || mode_graphics_less(&mode,vm)) {
					mode = *vm;
					mode_found = true;
				}
			}
		}
	}

	// seach default mode in text selected
	if (!mode_found) {
		for(video_mode_container_iterator_begin(&i,&vm_sel);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
			const video_mode* vm = video_mode_container_iterator_get(&i);
			if (strcmp(vm->name,DEFAULT_TEXT_MODE)==0) {
				mode = *vm;
				mode_found = true;
				break;
			}
		}
	}

	// search text mode in default
	if (!mode_found) {
		for(video_mode_container_iterator_begin(&i,&vm_def);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
			const video_mode* vm = video_mode_container_iterator_get(&i);
			if ((monitor_error>0 || video_monitor_mode_check(&monitor,vm)) && video_mode_is_text(vm)) {
				if (!mode_found || mode_text_less(&mode,vm)) {
					mode = *vm;
					mode_found = true;
				}
			}
		}
	}

	video_mode_container_done(&vm_def);
	video_mode_container_done(&vm_sel);

	if (!mode_found) {
		cerr << "No video mode avaliable" << endl;
		goto out_video;
	}

	return true;

out_video:
        video_done();
out_joy:
	joystick_done();
	allegro_exit();
out:
	return false;
}

void text_done(void) {
	video_mode_reset();
	video_done();
	allegro_exit();
}

static double text_gamma;
static double text_brightness;
static unsigned text_idle_0;
static unsigned text_idle_0_rep;
static unsigned text_idle_1;
static unsigned text_idle_1_rep;
static clock_t text_idle_clock;

// Inizialize the submenu
bool text_sub_init(const char* font, unsigned orientation, double gamma, double brightness, unsigned idle_0, unsigned idle_0_rep,unsigned idle_1, unsigned idle_1_rep) {
	text_idle_clock = clock();
	text_idle_0 = idle_0;
	text_idle_1 = idle_1;
	text_idle_0_rep = idle_0_rep;
	text_idle_1_rep = idle_1_rep;

	text_orientation = orientation;
	if (gamma < 0.1) gamma = 0.1;
	if (gamma > 10) gamma = 10;
	text_gamma = 1.0 / gamma;
	text_brightness = brightness;

	// load the font
	real_font_map = bitmapfont_load(font);
	if (!real_font_map)
		real_font_map = bitmapfont_default();

	// set the orientation
	bitmapfont_orientation(real_font_map,text_orientation,0);

	// compute font size
	real_font_dx = bitmapfont_size_x(real_font_map);
	real_font_dy = bitmapfont_size_y(real_font_map);

	// startup keyboard
	if (install_keyboard() != 0) {
		goto out_font;
	}

	// set the video mode
	if (video_mode_set(&mode) != 0) {
		goto out_key;
	}

	// clear the screen
	if (!video_is_text()) {
		for(unsigned y=0;y<video_size_y();++y)
			for(unsigned x=0;x<video_size_x();++x)
				video_put_pixel(x,y,0);
	}

	// set font
	bitmapfont_set(real_font_map);

	// compute text video size
	if (video_is_text()) {
		real_text_dx = video_size_x() / video_font_size_x();
		real_text_dy = video_size_y() / video_font_size_y();
	 } else {
		real_text_dx = video_size_x() / real_font_dx;
		real_text_dy = video_size_y() / real_font_dy;
	}

	text_dx = real_text_dx;
	text_dy = real_text_dy;
	if (!video_is_text()) {
		if (text_orientation & ORIENTATION_FLIP_XY)
			swap(text_dx,text_dy);
	}

	return true;

out_key:
	remove_keyboard();
out_font:
	bitmapfont_free(real_font_map);
	return false;
}

// Deinizialize the sub menu
void text_sub_done(void) {
	video_mode_reset();
	remove_keyboard();
	bitmapfont_free(real_font_map);
}

//---------------------------------------------------------------------------
// Put

void text_put(int x, int y, char c, int color) {
	if (x>=0 && y>=0 && x<text_dx_get() && y<text_dy_get()) {
		// orientation
		if (text_orientation & ORIENTATION_FLIP_XY) {
			swap(x,y);
		}
		if (text_orientation & ORIENTATION_MIRROR_X)
			x = real_text_dx - x - 1;
		if (text_orientation & ORIENTATION_MIRROR_Y)
			y = real_text_dy - y - 1;
		real_text_map[x + y * real_text_dx] = (color << 8) | (unsigned)c;
	}
}

void text_put(int x, int y, const string& s, int color) {
        for(unsigned i=0;i<s.length();++i)
	        text_put(x+i,y,s[i],color);
}

void text_fill(int x, int y, int dx, char c, int color) {
	for(int i=0;i<dx;++i)
	        text_put(x+i,y,c,color);
}

void text_fill(int x, int y, int dx, int dy, char c, int color) {
	for(int i=0;i<dy;++i)
	        text_fill(x,y+i,dx,c,color);
}

//---------------------------------------------------------------------------
// Color

#define PALETTE_RESERVED_MAX 16

static video_color PALETTE_RESERVED[PALETTE_RESERVED_MAX] = {
{   0,  0,  0 },
{ 192,  0,  0 },
{   0,192,  0 },
{ 192,192,  0 },
{   0,  0,192 },
{ 192,  0,192 },
{   0,192,192 },
{ 192,192,192 },
{ 128,128,128 },
{ 255,  0,  0 },
{   0,255,  0 },
{ 255,255,  0 },
{   0,  0,255 },
{ 255,  0,255 },
{   0,255,255 },
{ 255,255,255 }
};

static video_color text_palette[256];

int palette_dist(const video_color& A, const video_color& B) {
	int r = (int)A.red - B.red;
	int g = (int)A.green - B.green;
	int b = (int)A.blue - B.blue;
	return r*r + g*g + b*b;
}

void palette_set(video_color* pal, unsigned pal_max, unsigned* conv) {
	unsigned i;
	for(i=0;i<PALETTE_RESERVED_MAX;++i)
		text_palette[i] = PALETTE_RESERVED[i];

	for(i=0;i<256-PALETTE_RESERVED_MAX && i<pal_max;++i) {
		text_palette[i+PALETTE_RESERVED_MAX] = pal[i];
		conv[i] = i+PALETTE_RESERVED_MAX;
	}

	// search for best colors
	while (i<pal_max) {
		unsigned best = PALETTE_RESERVED_MAX;
		int dist = palette_dist(pal[i],text_palette[best]);
		for(unsigned j=best+1;j<256;++j) {
			int ndist = palette_dist(pal[i],text_palette[j]);
			if (ndist < dist) {
				dist = ndist;
				best = j;
			}
		}
		conv[i] = best;
		++i;
	}

	for(i=0;i<256;++i) {
		double f = 255.0 * text_brightness;
		if (i>=PALETTE_RESERVED_MAX)
			f *= backdrop_brightness;
		text_palette[i].red = (uint8)(f * pow(text_palette[i].red / 255.0, text_gamma));
		text_palette[i].green = (uint8)(f * pow(text_palette[i].green / 255.0, text_gamma));
		text_palette[i].blue = (uint8)(f * pow(text_palette[i].blue / 255.0, text_gamma));
	}
}

inline unsigned color2palette(unsigned color) {
	return color & 0xF;
}

// Conversion from text color to video color
inline video_rgb color2rgb(unsigned color) {
	video_rgb r;
	color &= 0xF;
	video_rgb_make(&r, PALETTE_RESERVED[color].red, PALETTE_RESERVED[color].green, PALETTE_RESERVED[color].blue);
	return r;
}

//---------------------------------------------------------------------------
// Backdrop

extern "C" {

int fast_exit_handler(void) {
	return text_keypressed() != 0;
}

};

// Select the backdrop
void text_backdrop(const char* file) {
	strcpy(backdrop_path,file);
}

// Clear the backdrop
void text_backdrop_clear() {
	backdrop_path[0] = 0;
	if (backdrop) {
		bitmap_free(backdrop);
		backdrop = 0;
	}
}

// Set the backdrop position and size
void text_backdrop_set(int x, int y, int dx, int dy, unsigned back_color, double Abackdrop_brightness) {
	text_fill(x,y,dx,dy,' ',back_color);

	backdrop_brightness = Abackdrop_brightness;
	real_backdrop_x = x;
	real_backdrop_y = y;
	real_backdrop_dx = dx;
	real_backdrop_dy = dy;
	if (text_orientation & ORIENTATION_FLIP_XY) {
		swap(real_backdrop_x,real_backdrop_y);
		swap(real_backdrop_dx,real_backdrop_dy);
	}
	if (text_orientation & ORIENTATION_MIRROR_X) {
		real_backdrop_x = real_text_dx - real_backdrop_x - real_backdrop_dx;
	}
	if (text_orientation & ORIENTATION_MIRROR_Y) {
		real_backdrop_y = real_text_dy - real_backdrop_y - real_backdrop_dy;
	}

	backdrop_color_mask = back_color;
}

// Update the backdrop image
bool text_backdrop_update() {
	if (backdrop_path[0]) {
		if (access(backdrop_path,F_OK | R_OK)!=0) {
			backdrop_path[0] = 0;
			if (backdrop) {
				bitmap_free(backdrop);
				backdrop = 0;
			}
		} else {
			video_color rgb[256];
			unsigned rgb_max;
			struct bitmap* bitmap = backdrop_load(backdrop_path, rgb, &rgb_max, fast_exit_handler);
			if (!bitmap)
				return false;

			if (video_index() == VIDEO_FLAGS_INDEX_PACKED && bitmap->bytes_per_pixel != 1) {
				rgb_max = 256 - PALETTE_RESERVED_MAX;
				rgb_max = bitmap_reduction(backdrop_convert, rgb, rgb_max, bitmap, fast_exit_handler);
			}

			// set orientation (only flip)
			if (bitmap_orientation(bitmap,text_orientation & ORIENTATION_FLIP_XY, fast_exit_handler) != 0) {
				bitmap_free(bitmap);
				return false;
			}

			bitmap_free(backdrop);
			backdrop = bitmap;

			backdrop_path[0] = 0;

			// palette
			if (backdrop->bytes_per_pixel == 1) {
				if (video_index() == VIDEO_FLAGS_INDEX_RGB) {
					for(unsigned i=0;i<rgb_max;++i)
						video_rgb_make(&backdrop_color_map[i],rgb[i].red,rgb[i].green,rgb[i].blue);
				} else if (video_index() == VIDEO_FLAGS_INDEX_PACKED) {
					palette_set(rgb,rgb_max,backdrop_index_map);
				}
			} else {
				if (video_index() == VIDEO_FLAGS_INDEX_PACKED) {
					unsigned i;
					palette_set(rgb,rgb_max,backdrop_index_map);
					for(i=0;i<BITMAP_INDEX_MAX;++i)
						backdrop_convert[i] = backdrop_index_map[backdrop_convert[i]];
				}
			}

			// set orientation (only mirror) and scaling
			for(unsigned i=0;i<video_size_x();++i) {
				unsigned x = i * (backdrop->size_x) / (real_backdrop_dx*real_font_dx);
				if (text_orientation & ORIENTATION_MIRROR_X) {
					x = backdrop->size_x - x - 1;
				}
				backdrop_x_map[i] = x;
			}
			for(unsigned i=0;i<video_size_y();++i) {
				unsigned y = i * (backdrop->size_y) / (real_backdrop_dy*real_font_dy);
				if (text_orientation & ORIENTATION_MIRROR_Y) {
					y = backdrop->size_y - y - 1;
				}
				backdrop_y_map[i] = y;
			}
		}
	} else {
		if (video_index() == VIDEO_FLAGS_INDEX_PACKED)
			palette_set(0,0,0);
	}
	return true;
}

//---------------------------------------------------------------------------
// Update

void text_put16_char_backdrop8(unsigned x, unsigned y) {
	for(unsigned cy=0;cy<real_font_dy;++cy) {
		unsigned dst_off = video_write_line(y+cy) + 2*x;
		unsigned by = y - real_backdrop_y * real_font_dy;
		unsigned char* back_off = bitmap_line(backdrop,backdrop_y_map[by+cy]);

		_farsetsel(video_write_selector());

		unsigned* bx = &backdrop_x_map[x - real_backdrop_x * real_font_dx];
		for(unsigned cx=0;cx<real_font_dx;++cx) {
			unsigned color = backdrop_color_map[back_off[*bx]];
			_farnspokew(dst_off, color);
			dst_off += 2;
			bx += 1;
		}
	}
}

void text_put8_char_backdrop8(unsigned x, unsigned y) {
	for(unsigned cy=0;cy<real_font_dy;++cy) {
		unsigned by = y - real_backdrop_y * real_font_dy;
		unsigned char* back_off = bitmap_line(backdrop,backdrop_y_map[by+cy]);
		unsigned* bx = &backdrop_x_map[x - real_backdrop_x * real_font_dx];
		if (video_unchained_active()) {
			unsigned dst_off = video_write_line(y+cy) + x/4;
			_farsetsel(video_write_selector());
			for(unsigned plane=0;plane<4;++plane) {
				video_unchained_plane_set(plane);
				unsigned* p_bx = bx + plane;
				unsigned p_dst_off = dst_off;
				for(unsigned cx=plane;cx<real_font_dx;cx += 4) {
					unsigned color = backdrop_index_map[back_off[*p_bx]];
					_farnspokeb(p_dst_off, color);
					p_dst_off += 1;
					p_bx += 4;
				}
			}
		} else {
			unsigned dst_off = video_write_line(y+cy) + x;
			_farsetsel(video_write_selector());
			for(unsigned cx=0;cx<real_font_dx;++cx) {
				unsigned color = backdrop_index_map[back_off[*bx]];
				_farnspokeb(dst_off, color);
				dst_off += 1;
				bx += 1;
			}
		}
	}
}

void text_put16_char_backdrop24(unsigned x, unsigned y) {
	for(unsigned cy=0;cy<real_font_dy;++cy) {
		unsigned dst_off = video_write_line(y+cy) + 2*x;
		unsigned by = y - real_backdrop_y * real_font_dy;
		unsigned char* back_off = bitmap_line(backdrop,backdrop_y_map[by+cy]);

		_farsetsel(video_write_selector());

		unsigned* bx = &backdrop_x_map[x - real_backdrop_x * real_font_dx];
		for(unsigned cx=0;cx<real_font_dx;++cx) {
			unsigned off = *bx * 3;
			video_rgb color;
			video_rgb_make(&color,back_off[off],back_off[off+1],back_off[off+2]);
			_farnspokew(dst_off, color);
			dst_off += 2;
			bx += 1;
		}
	}
}

void text_put8_char_backdrop24(unsigned x, unsigned y) {
	for(unsigned cy=0;cy<real_font_dy;++cy) {
		unsigned by = y - real_backdrop_y * real_font_dy;
		unsigned char* back_off = bitmap_line(backdrop,backdrop_y_map[by+cy]);
		unsigned* bx = &backdrop_x_map[x - real_backdrop_x * real_font_dx];
		if (video_unchained_active()) {
			unsigned dst_off = video_write_line(y+cy) + x/4;
			_farsetsel(video_write_selector());
			for(unsigned plane=0;plane<4;++plane) {
				video_unchained_plane_set(plane);
				unsigned* p_bx = bx + plane;
				unsigned p_dst_off = dst_off;
				for(unsigned cx=plane;cx<real_font_dx;cx += 4) {
					unsigned off = *p_bx * 3;
					unsigned color = backdrop_convert[BITMAP_COLOR_TO_INDEX(back_off[off],back_off[off+1],back_off[off+2])];
					_farnspokeb(p_dst_off, color);
					p_dst_off += 1;
					p_bx += 4;
				}
			}
		} else {
			unsigned dst_off = video_write_line(y+cy) + x;
			_farsetsel(video_write_selector());
			for(unsigned cx=0;cx<real_font_dx;++cx) {
				unsigned off = *bx * 3;
				unsigned color = backdrop_convert[BITMAP_COLOR_TO_INDEX(back_off[off],back_off[off+1],back_off[off+2])];
				_farnspokeb(dst_off, color);
				dst_off += 1;
				bx += 1;
			}
		}
	}
}

void text_put16_char_trasp8(unsigned x, unsigned y, unsigned v) {
	video_rgb color_fore = color2rgb(v >> 8);
	struct bitmap* src = real_font_map->data[v & 0xFF];
	for(unsigned cy=0;cy<real_font_dy;++cy) {
		unsigned dst_off = video_write_line(y+cy) + 2*x;
		unsigned char* src_off = bitmap_line(src,cy);
		unsigned by = y - real_backdrop_y * real_font_dy;
		unsigned char* back_off = bitmap_line(backdrop,backdrop_y_map[by+cy]);

		_farsetsel(video_write_selector());

		unsigned* bx = &backdrop_x_map[x - real_backdrop_x * real_font_dx];
		for(unsigned cx=0;cx<real_font_dx;++cx) {
			unsigned color;
 			if (*src_off)
 				color = color_fore;
 			else
 				color = backdrop_color_map[back_off[*bx]];
			_farnspokew(dst_off, color);
			dst_off += 2;
			src_off += 1;
			bx += 1;
		}
	}
}

void text_put8_char_trasp8(unsigned x, unsigned y, unsigned v) {
	unsigned color_fore = color2palette(v >> 8);
	struct bitmap* src = real_font_map->data[v & 0xFF];
	for(unsigned cy=0;cy<real_font_dy;++cy) {
		unsigned char* src_off = bitmap_line(src,cy);
		unsigned by = y - real_backdrop_y * real_font_dy;
		unsigned char* back_off = bitmap_line(backdrop,backdrop_y_map[by+cy]);
		unsigned* bx = &backdrop_x_map[x - real_backdrop_x * real_font_dx];
		if (video_unchained_active()) {
			unsigned dst_off = video_write_line(y+cy) + x/4;
			_farsetsel(video_write_selector());
			for(unsigned plane=0;plane<4;++plane) {
				video_unchained_plane_set(plane);
				unsigned char* p_src_off = src_off + plane;
				unsigned* p_bx = bx + plane;
				unsigned p_dst_off = dst_off;
				for(unsigned cx=plane;cx<real_font_dx;cx += 4) {
					unsigned color;
		 			if (*p_src_off)
		 				color = color_fore;
		 			else
	 					color = backdrop_index_map[back_off[*p_bx]];
					_farnspokeb(p_dst_off, color);
					p_dst_off += 1;
					p_src_off += 4;
					p_bx += 4;
				}
			}
		} else {
			unsigned dst_off = video_write_line(y+cy) + x;
			_farsetsel(video_write_selector());
			for(unsigned cx=0;cx<real_font_dx;++cx) {
				unsigned color;
	 			if (*src_off)
	 				color = color_fore;
	 			else
	 				color = backdrop_index_map[back_off[*bx]];
				_farnspokeb(dst_off, color);
				dst_off += 1;
				src_off += 1;
				bx += 1;
			}
		}
	}
}

void text_put16_char_trasp24(unsigned x, unsigned y, unsigned v) {
	video_rgb color_fore = color2rgb(v >> 8);
	struct bitmap* src = real_font_map->data[v & 0xFF];
	for(unsigned cy=0;cy<real_font_dy;++cy) {
		unsigned dst_off = video_write_line(y+cy) + 2*x;
		unsigned char* src_off = bitmap_line(src,cy);
		unsigned by = y - real_backdrop_y * real_font_dy;
		unsigned char* back_off = bitmap_line(backdrop,backdrop_y_map[by+cy]);

		_farsetsel(video_write_selector());

		unsigned* bx = &backdrop_x_map[x - real_backdrop_x * real_font_dx];
		for(unsigned cx=0;cx<real_font_dx;++cx) {
			video_rgb color;
 			if (*src_off)
 				color = color_fore;
 			else {
				unsigned off = *bx * 3;
				video_rgb_make(&color,back_off[off],back_off[off+1],back_off[off+2]);
			}
			_farnspokew(dst_off, color);
			dst_off += 2;
			src_off += 1;
			bx += 1;
		}
	}
}

void text_put16_char_font(unsigned x, unsigned y, unsigned v) {
	struct bitmap* src = real_font_map->data[v & 0xFF];
	video_rgb color_fore = color2rgb(v >> 8);
	video_rgb color_back = color2rgb(v >> 12);
	for(unsigned cy=0;cy<real_font_dy;++cy) {
		unsigned long dst_off = video_write_line(y+cy) + 2*x;
		unsigned char* src_off = bitmap_line(src,cy);

		_farsetsel(video_write_selector());

		for(unsigned cx=0;cx<real_font_dx;++cx) {
			unsigned color = *src_off ? color_fore : color_back;
			_farnspokew(dst_off, color);
			dst_off += 2;
			src_off += 1;
		}
	}
}

void text_put8_char_font(unsigned x, unsigned y, unsigned v) {
	struct bitmap* src = real_font_map->data[v & 0xFF];
	unsigned color_fore = color2palette(v >> 8);
	unsigned color_back = color2palette(v >> 12);
	for(unsigned cy=0;cy<real_font_dy;++cy) {
		unsigned char* src_off = bitmap_line(src,cy);
		if (video_unchained_active()) {
			unsigned dst_off = video_write_line(y+cy) + x/4;
			_farsetsel(video_write_selector());
			for(unsigned plane=0;plane<4;++plane) {
				video_unchained_plane_set(plane);
				unsigned char* p_src_off = src_off + plane;
				unsigned p_dst_off = dst_off;
				for(unsigned cx=plane;cx<real_font_dx;cx += 4) {
					unsigned color = *p_src_off ? color_fore : color_back;
					_farnspokeb(p_dst_off, color);
					p_dst_off += 1;
					p_src_off += 4;
				}
			}
		} else {
			unsigned long dst_off = video_write_line(y+cy) + x;
			_farsetsel(video_write_selector());
			for(unsigned cx=0;cx<real_font_dx;++cx) {
				unsigned color = *src_off ? color_fore : color_back;
				_farnspokeb(dst_off, color);
				dst_off += 1;
				src_off += 1;
			}
		}
	}
}

/*void text_put8_char_backdrop24(unsigned x, unsigned y) {
	text_put8_char_font(x,y, (0x7 << 8) | (unsigned char)'#');
}*/

void text_put8_char_trasp24(unsigned x, unsigned y, unsigned v) {
	text_put8_char_font(x,y,v);
}

void text_put_char_grahics(int x, int y, unsigned v) {
	if (backdrop &&
		real_backdrop_x<=x && x<real_backdrop_x+real_backdrop_dx &&
		real_backdrop_y<=y && y<real_backdrop_y+real_backdrop_dy &&
		((v >> 12) & 0xF) == backdrop_color_mask) {

		x *= real_font_dx;
		y *= real_font_dy;

		if (video_bytes_per_pixel() == 1) {
			if (backdrop->bytes_per_pixel==1) {
				if ((v & 0xFF)==' ')
					text_put8_char_backdrop8(x,y);
				else
					text_put8_char_trasp8(x,y,v);
			} else {
				text_put8_char_backdrop24(x,y);
			}
		} else {
			if (backdrop->bytes_per_pixel==1) {
				if ((v & 0xFF)==' ')
					text_put16_char_backdrop8(x,y);
				else
					text_put16_char_trasp8(x,y,v);
			} else {
				if ((v & 0xFF)==' ')
					text_put16_char_backdrop24(x,y);
				else
					text_put16_char_trasp24(x,y,v);
			}
		}
	} else {
		x *= real_font_dx;
		y *= real_font_dy;

		if (video_bytes_per_pixel() == 1)
			text_put8_char_font(x,y,v);
		else
			text_put16_char_font(x,y,v);
	}
}

static void text_update_graphics() {
	for(unsigned y=0;y<real_text_dy;++y) {
		for(unsigned x=0;x<real_text_dx;++x) {
			text_put_char_grahics(x,y,real_text_map[x+y*real_text_dx]);
		}
	}
}

static void text_update_text() {
	for(unsigned y=0;y<real_text_dy;++y) {
		unsigned long dst_off = video_write_line(y);
		unsigned* src_off = real_text_map + y * real_text_dx;

		_farsetsel(video_write_selector());

		for(unsigned x=0;x<real_text_dx;++x) {
			_farnspokew(dst_off, *src_off);
			++src_off;
			dst_off += 2;
		}
	}
}

static void text_update() {
	if (video_is_text()) {
		text_update_text();
	} else if (video_is_graphics()) {
		if (text_backdrop_update()) {
			/*
			static video_color BLANK_PALETTE[256];
			if (video_index() == VIDEO_FLAGS_INDEX_PACKED) {
				video_palette_set(BLANK_PALETTE,PALETTE_RESERVED_MAX,256 - PALETTE_RESERVED_MAX,1);
			}
			*/
			text_update_graphics();
			if (video_index() == VIDEO_FLAGS_INDEX_PACKED) {
				video_palette_set(text_palette,0,256,0);
			}
		}
	}
}

//---------------------------------------------------------------------------
// Key

struct key_cvt {
	const char* name;
	unsigned event;
	unsigned key0;
	unsigned key1;
};

static struct key_cvt KEYTAB[] = {
{"up", TEXT_KEY_UP, KEY_UP, KEY_MAX},
{"down", TEXT_KEY_DOWN, KEY_DOWN, KEY_MAX},
{"left", TEXT_KEY_LEFT, KEY_LEFT, KEY_MAX},
{"right",TEXT_KEY_RIGHT, KEY_RIGHT, KEY_MAX},
{"enter",TEXT_KEY_ENTER, KEY_ENTER, KEY_MAX},
{"esc",TEXT_KEY_ESC, KEY_ESC, KEY_MAX},
{"space",TEXT_KEY_SPACE, KEY_SPACE, KEY_MAX},
{"mode",TEXT_KEY_MODE, KEY_TAB, KEY_MAX},
{"home",TEXT_KEY_HOME, KEY_HOME, KEY_MAX},
{"end",TEXT_KEY_END, KEY_END, KEY_MAX},
{"pgup",TEXT_KEY_PGUP, KEY_PGUP, KEY_MAX},
{"pgdn",TEXT_KEY_PGDN, KEY_PGDN, KEY_MAX},
{"help",TEXT_KEY_HELP, KEY_F1, KEY_MAX},
{"group",TEXT_KEY_GROUP, KEY_F2, KEY_MAX},
{"type",TEXT_KEY_TYPE, KEY_F3, KEY_MAX},
{"exclude",TEXT_KEY_EXCLUDE, KEY_F4, KEY_MAX},
{"sort",TEXT_KEY_SORT, KEY_F5, KEY_MAX},
{"setgroup",TEXT_KEY_SETGROUP, KEY_F9, KEY_MAX},
{"settype",TEXT_KEY_SETTYPE, KEY_F10, KEY_MAX},
{"runclone",TEXT_KEY_RUN_CLONE, KEY_F12, KEY_MAX},
{0,0,0,0}
};

static int keyboard_raw_poll() {
	poll_keyboard();

	for(const struct key_cvt* i=KEYTAB;i->name;++i)
		if ((i->key0 < KEY_MAX && key[i->key0])
			|| (i->key1 < KEY_MAX && key[i->key1]))
			return i->event;

	for(unsigned i=KEY_A;i<=KEY_Z;++i)
		if (key[i])
			return 'a' + i - KEY_A;

	return TEXT_KEY_NONE;
}

/**************************************************************************/

// Number of event repeat for second
#define KEY_repeat_REPEAT_0 2
#define KEY_repeat_REPEAT_1 20

static int key_poll() {
	static int key_repeat_last = TEXT_KEY_NONE;
	static clock_t key_repeat_last_time = 0;
	static bool key_repeat_last_counter = 0;

	int r = keyboard_raw_poll();
	if (r == TEXT_KEY_NONE)
		r = joystick_raw_poll();

	if (r != key_repeat_last || r == TEXT_KEY_NONE) {
		key_repeat_last = r;
		key_repeat_last_counter = 0;
		key_repeat_last_time = clock();
		return r;
	} else {
		if ((key_repeat_last_counter == 0 && (clock() - key_repeat_last_time > CLOCKS_PER_SEC / KEY_repeat_REPEAT_0)) ||
			(key_repeat_last_counter > 0 && (clock() - key_repeat_last_time > CLOCKS_PER_SEC / KEY_repeat_REPEAT_1))) {
			key_repeat_last_time = clock();
			++key_repeat_last_counter;
			return r;
		} else {
			return TEXT_KEY_NONE;
		}
	}
}

static int key_saved = TEXT_KEY_NONE;
static int key_last = TEXT_KEY_NONE;

int text_keypressed() {
	static clock_t key_pressed_last_time = 0;

	if (key_saved != TEXT_KEY_NONE)
		return 1;

	clock_t now = clock();

	/* don't check too fast */
	if (now - key_pressed_last_time >= CLOCKS_PER_SEC / 25) {
		key_pressed_last_time = now;

		key_saved = key_poll();

		if (key_saved != TEXT_KEY_NONE)
			return 1;
	}

	return 0;
}

static void text_idle() {
	if (key_last == TEXT_KEY_IDLE_0 && text_idle_0_rep && clock() - text_idle_clock > text_idle_0_rep * CLOCKS_PER_SEC)
		key_saved = TEXT_KEY_IDLE_0;
	if (key_last == TEXT_KEY_IDLE_1 && text_idle_1_rep && clock() - text_idle_clock > text_idle_1_rep * CLOCKS_PER_SEC)
		key_saved = TEXT_KEY_IDLE_1;
	if (text_idle_0 && clock() - text_idle_clock > text_idle_0 * CLOCKS_PER_SEC)
		key_saved = TEXT_KEY_IDLE_0;
	if (text_idle_1 && clock() - text_idle_clock > text_idle_1 * CLOCKS_PER_SEC)
		key_saved = TEXT_KEY_IDLE_1;
}

unsigned text_getkey() {
	if (!text_keypressed()) {
		text_update();

		while (!text_keypressed()) {
			text_idle();
		}
	}

	text_idle_clock = clock();

	assert(	key_saved != TEXT_KEY_NONE);

	key_last = key_saved;
	key_saved = TEXT_KEY_NONE;

	return key_last;
}


static void text_key_insert(unsigned event, unsigned key0, unsigned key1) {
	unsigned i;
	for(i=0;KEYTAB[i].name && KEYTAB[i].event != event;++i);
	if (KEYTAB[i].name) {
		KEYTAB[i].key0 = key0;
		KEYTAB[i].key1 = key1;
	}
}

static unsigned text_string2event(const string& s) {
	unsigned i;
	for(i=0;KEYTAB[i].name && KEYTAB[i].name != s;++i);
	if (KEYTAB[i].name)
		return KEYTAB[i].event;
	else
		return TEXT_KEY_NONE;
}

bool text_key_in(const string& s) {
	string sevent;
	string skey0;
	string skey1;
	unsigned i = 0;

	while (i < s.length() && !isspace(s[i])) {
		sevent += s[i];
		++i;
	}
         
	while (i < s.length() && isspace(s[i]))
        	++i;

	while (i < s.length() && !isspace(s[i])) {
		skey0 += s[i];
		++i;
	}
         
	while (i < s.length() && isspace(s[i]))
		++i;


	while (i < s.length() && !isspace(s[i])) {
		skey1 += s[i];
        	++i;
	}
         
	while (i < s.length() && isspace(s[i]))
		++i;

	unsigned event = text_string2event(sevent);
	if (event == TEXT_KEY_NONE)
		return false;

	unsigned key0 = atoi(skey0.c_str());
	if (!key0)
		key0 = KEY_MAX;
	unsigned key1 = atoi(skey1.c_str());
	if (!key1)
		key1 = KEY_MAX;

	text_key_insert(event,key0,key1);

	if (i != s.length())
		return false;

	return true;
}

void text_key_out_all(ostream& os) {
	unsigned i;
	for(i=0;KEYTAB[i].name;++i) {
		os << "event " << KEYTAB[i].name << " ";
 		if (KEYTAB[i].key0 < KEY_MAX)
 			os << KEYTAB[i].key0;
		else
			os << "0";
		os << " ";
 		if (KEYTAB[i].key1 < KEY_MAX)
 			os << KEYTAB[i].key1;
		else
			os << "0";
		os << endl;
	}
}
