/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "screen.h"
#include "scrcard.h"
#include "scrblit.h"
#include "font.h"
#include "gtf.h"
#include "allegro.h"

#include <sys/movedata.h>
#include <dos.h>
#include <math.h>
#include <conio.h>
#include <time.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <pc.h>

/***************************************************************************/
/* video_mode_container extension */

int mode_cmp(const video_mode* a,const video_mode* b) {
	int av;
	int bv;

	if (a->size_x < b->size_x)
		return -1;
	if (a->size_x > b->size_x)
		return 1;
	if (a->size_y < b->size_y)
		return -1;
	if (a->size_y > b->size_y)
		return 1;
	if (a->bits < b->bits)
		return -1;
	if (a->bits > b->bits)
		return 1;

	av = a->flags & VIDEO_FLAGS_LINEAR;
	bv = b->flags & VIDEO_FLAGS_LINEAR;
	if (av > bv)
		return -1;
	if (av < bv)
		return 1;

	av = a->flags & VIDEO_FLAGS_DRIVER_MASK;
	bv = b->flags & VIDEO_FLAGS_DRIVER_MASK;
	if (av < bv)
		return -1;
	if (av > bv)
		return 1;

	return 0;
}

video_mode* video_mode_container_pos(video_mode_container* vmc, unsigned pos) {
	video_mode_container_iterator i;
	video_mode_container_iterator_begin(&i,vmc);
	while (pos && !video_mode_container_iterator_end(&i)) {
		--pos;
		video_mode_container_iterator_next(&i);
	}
	if (video_mode_container_iterator_end(&i))
		return 0;
	else
		return video_mode_container_iterator_get(&i);
}

unsigned video_mode_container_max(video_mode_container* vmc) {
	unsigned pos = 0;
	video_mode_container_iterator i;
	video_mode_container_iterator_begin(&i,vmc);
	while (!video_mode_container_iterator_end(&i)) {
		++pos;
		video_mode_container_iterator_next(&i);
	}
	return pos;
}

/***************************************************************************/
/* sound */

int sound_flag = 1;

void sound_error(void) {
	if (sound_flag) {
		sound(100);
		delay(120);
		nosound();
	}
}

void sound_warn(void) {
	if (sound_flag) {
		sound(900);
		delay(10);
		nosound();
	}
}

void sound_signal(void) {
	if (sound_flag) {
		unsigned i;
		for(i=0;i<10;++i) {
			sound(800);
			delay(10);
			nosound();
			delay(10);
		}
	}
}

/***************************************************************************/
/* Common variable */

int menu_base;
int menu_rel;
int menu_rel_max;
int menu_base_max;
int menu_max;

video_mode_container vm_all; /* main video mode container */

int vm_all_modified; /* container need to be saved */

/***************************************************************************/
/* Monitor */

static video_monitor monitor;

int monitor_check(const video_mode* mode) {
	if (!video_monitor_mode_check(&monitor, mode)) {
		return 0;
	}

	return 1;
}

/* Really SURE that the monitor support the video mode */
int monitor_check_strict(const video_mode* mode) {
	switch (video_mode_driver(mode)) {
		case VIDEO_FLAGS_DRIVER_VGA:
		case VIDEO_FLAGS_DRIVER_VBE :
			return 0;
		default:
			return monitor_check(mode);
	}
}

/* Probably that the monitor support the video mode */
int monitor_check_safe(const video_mode* mode) {
	switch (video_mode_driver(mode)) {
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VBE :
			return video_monitor_clock_check(&monitor, 31.5E3, 60);
		default:
			return monitor_check(mode);
	}
}

/***************************************************************************/
/* Text output */

#define DEFAULT_TEXT_MODE "default_text"

unsigned screen_size_x(void) {
	return video_size_x() / video_font_size_x();
}

unsigned screen_size_y(void) {
	return video_size_y() / video_font_size_y();
}

void screen_clear(void) {
	unsigned x,y;
	for(x=0;x<screen_size_x();++x)
		for(y=0;y<screen_size_y();++y)
			video_put_char(x,y,0,0);
}

static struct bitmapfont* font_map;

void screen_set_font(void) {
	if (video_is_text()) {
		if (font_map) {
			bitmapfont_set(font_map);
		} else {
			unsigned font_row = video_size_y() / 25;
			if (font_row <= 8)
				vga_font_set_default8();
			else if (font_row <= 14)
				vga_font_set_default14();
			else
				vga_font_set_default16();
		}
	}
}

/* Default video mode */
static int screen_default_flag = 0;
static video_mode screen_default;

int screen_default_set(void) {
	video_mode_container_iterator i;

	/* search the "default_text" mode in selected */
	for(video_mode_container_iterator_begin(&i,&vm_all);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
		const video_mode* vm = video_mode_container_iterator_get(&i);
		if ((vm->flags & VIDEO_FLAGS_USER_BIT0)!=0
			&& video_mode_is_text(vm)
			&& strcmp(vm->name,DEFAULT_TEXT_MODE)==0) {
			screen_default = *vm;
			if (video_mode_set(&screen_default)!=0) {
				return -1;
			}
			screen_default_flag = 1;
			screen_set_font();
			return 0;
		}
	}

	/* search the best mode in selected */
	for(video_mode_container_iterator_begin(&i,&vm_all);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
		const video_mode* vm = video_mode_container_iterator_get(&i);
		if ((vm->flags & VIDEO_FLAGS_USER_BIT0)!=0
			&& video_mode_is_text(vm)
			&& monitor_check_strict(vm)) {
			screen_default = *vm;
			if (video_mode_set(&screen_default)!=0) {
				return -1;
			}
			screen_default_flag = 1;
			screen_set_font();
			return 0;
		}
	}

	/* search the best mode in default */
	for(video_mode_container_iterator_begin(&i,&vm_all);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
		const video_mode* vm = video_mode_container_iterator_get(&i);
		if ((vm->flags & VIDEO_FLAGS_USER_BIT0)==0
			&& video_mode_is_text(vm)
			&& monitor_check_strict(vm)) {
			screen_default = *vm;
			if (video_mode_set(&screen_default)!=0) {
				return -1;
			}
			screen_default_flag = 1;
			screen_set_font();
			return 0;
		}
	}

	/* use the current */
	if (vgaline_mode_grab(&screen_default.driver.vgaline)==0 &&
		vgaline_mode_import(&screen_default)==0 &&
		video_mode_is_text(&screen_default)) {
		strcpy(screen_default.name,DEFAULT_TEXT_MODE);

		if (video_mode_set(&screen_default)!=0) {
			return -1;
		}
		screen_default_flag = 1;
		screen_set_font();
		video_mode_container_insert_sort(&vm_all,&screen_default,mode_cmp);
		return 0;
	}

	return 0;
}

void screen_init(void) {
	font_map = bitmapfont_load("mv.fnt");
#if 0
	if (!font_map) {
		video_mode_reset();
		printf("Error loading the font file <mv.fnt>\n");
		exit(EXIT_FAILURE);
	}
#endif

	if (screen_default_set()!=0) {
		video_mode_reset();
		printf("Error inizialing the default video mode\n\"%s\"",video_error_description_get());
		exit(EXIT_FAILURE);
	}

	if (screen_default_flag == 0) {
		video_mode_reset();
		printf("No video mode avaliable for your hardware\n");
		exit(EXIT_FAILURE);
	}

}

void screen_reset(void) {
	if (video_mode_set(&screen_default)!=0) {
		video_mode_reset();
		printf("Error in inizialing the default video mode\n");
		exit(EXIT_FAILURE);
	}
	screen_set_font();
}

void screen_gotoxy(int x, int y) {
	(void)x;
	(void)y;
	/* TODO */
}

void screen_done(void) {
	if (font_map)
		bitmapfont_free(font_map);

	video_mode_reset();
}

void screen_put(int x, int y, char c, int color) {
	if (y>=0 && y<screen_size_y() && x>=0 && x<screen_size_x()) {
		video_put_char(x,y,c,color);
	}
}

int screen_kbhit(void) {
	return kbhit();
}

unsigned screen_getkey(void) {
	char c = getch();
        if (!c) {
        	unsigned e = getch();
                return e << 8;
        } else
        	return c;
}

/**************************************************************************/
/* Keys */

#define SCREEN_KEY_UP 18432
#define SCREEN_KEY_DOWN 20480
#define SCREEN_KEY_LEFT 19200
#define SCREEN_KEY_RIGHT 19712
#define SCREEN_KEY_CTRLUP 36096
#define SCREEN_KEY_CTRLDOWN 37120
#define SCREEN_KEY_CTRLRIGHT 29696
#define SCREEN_KEY_CTRLLEFT 29440
#define SCREEN_KEY_ENTER 13
#define SCREEN_KEY_ESC 27
#define SCREEN_KEY_HOME 18176
#define SCREEN_KEY_END 20224
#define SCREEN_KEY_PGUP 18688
#define SCREEN_KEY_PGDN 20736
#define SCREEN_KEY_F1 15104
#define SCREEN_KEY_F2 15360
#define SCREEN_KEY_F3 15616
#define SCREEN_KEY_F4 15872
#define SCREEN_KEY_F5 16128
#define SCREEN_KEY_F6 16384
#define SCREEN_KEY_F7 16640
#define SCREEN_KEY_F8 16896
#define SCREEN_KEY_F9 17152
#define SCREEN_KEY_F10 17408
#define SCREEN_KEY_BACKSPACE 8
#define SCREEN_KEY_DEL 21248
#define SCREEN_KEY_INS 20992
#define SCREEN_KEY_SPACE ' '
#define SCREEN_KEY_TAB '\t'

/* Colors */
#define SCREEN_COLOR_TITLE 0x0B
#define SCREEN_COLOR_NORMAL 0x03
#define SCREEN_COLOR_INPUT 0x0F
#define SCREEN_COLOR_BAD 0x0C
#define SCREEN_COLOR_MARK 0x30
#define SCREEN_COLOR_MARK_BAD 0x34
#define SCREEN_COLOR_SELECTED 0x70
#define SCREEN_COLOR_SELECTED_MARK 0x7F

#define SCREEN_COLOR_INFO_NORMAL 0x08
#define SCREEN_COLOR_INFO_TITLE 0x07

#define SCREEN_COLOR_BAR 0x70
#define SCREEN_COLOR_ERROR 0x4F
#define SCREEN_COLOR_REVERSE 0x70

/***************************************************************************/
/* Draw generic graphics */

#define FONTX 6
#define FONTY 8

static unsigned char draw_font[] = {
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x30,0x48,0x84,0xb4,0xb4,0x84,0x48,0x30,0x30,0x48,0x84,0x84,0x84,0x84,0x48,0x30,
		0x00,0xfc,0x84,0x8c,0xd4,0xa4,0xfc,0x00,0x00,0xfc,0x84,0x84,0x84,0x84,0xfc,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x30,0x68,0x78,0x78,0x30,0x00,0x00,
		0x80,0xc0,0xe0,0xf0,0xe0,0xc0,0x80,0x00,0x04,0x0c,0x1c,0x3c,0x1c,0x0c,0x04,0x00,
		0x20,0x70,0xf8,0x20,0x20,0xf8,0x70,0x20,0x48,0x48,0x48,0x48,0x48,0x00,0x48,0x00,
		0x00,0x00,0x30,0x68,0x78,0x30,0x00,0x00,0x00,0x30,0x68,0x78,0x78,0x30,0x00,0x00,
		0x70,0xd8,0xe8,0xe8,0xf8,0xf8,0x70,0x00,0x1c,0x7c,0x74,0x44,0x44,0x4c,0xcc,0xc0,
		0x20,0x70,0xf8,0x70,0x70,0x70,0x70,0x00,0x70,0x70,0x70,0x70,0xf8,0x70,0x20,0x00,
		0x00,0x10,0xf8,0xfc,0xf8,0x10,0x00,0x00,0x00,0x20,0x7c,0xfc,0x7c,0x20,0x00,0x00,
		0xb0,0x54,0xb8,0xb8,0x54,0xb0,0x00,0x00,0x00,0x28,0x6c,0xfc,0x6c,0x28,0x00,0x00,
		0x00,0x30,0x30,0x78,0x78,0xfc,0x00,0x00,0xfc,0x78,0x78,0x30,0x30,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x20,0x20,0x20,0x20,0x20,0x00,0x20,0x00,
		0x50,0x50,0x50,0x00,0x00,0x00,0x00,0x00,0x00,0x50,0xf8,0x50,0xf8,0x50,0x00,0x00,
		0x20,0x70,0xc0,0x70,0x18,0xf0,0x20,0x00,0x40,0xa4,0x48,0x10,0x20,0x48,0x94,0x08,
		0x60,0x90,0xa0,0x40,0xa8,0x90,0x68,0x00,0x10,0x20,0x40,0x00,0x00,0x00,0x00,0x00,
		0x20,0x40,0x40,0x40,0x40,0x40,0x20,0x00,0x10,0x08,0x08,0x08,0x08,0x08,0x10,0x00,
		0x20,0xa8,0x70,0xf8,0x70,0xa8,0x20,0x00,0x00,0x20,0x20,0xf8,0x20,0x20,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x30,0x30,0x60,0x00,0x00,0x00,0xf8,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x30,0x30,0x00,0x00,0x08,0x10,0x20,0x40,0x80,0x00,0x00,
		0x70,0x88,0x88,0x88,0x88,0x88,0x70,0x00,0x10,0x30,0x10,0x10,0x10,0x10,0x10,0x00,
		0x70,0x88,0x08,0x10,0x20,0x40,0xf8,0x00,0x70,0x88,0x08,0x30,0x08,0x88,0x70,0x00,
		0x10,0x30,0x50,0x90,0xf8,0x10,0x10,0x00,0xf8,0x80,0xf0,0x08,0x08,0x88,0x70,0x00,
		0x70,0x80,0xf0,0x88,0x88,0x88,0x70,0x00,0xf8,0x08,0x08,0x10,0x20,0x20,0x20,0x00,
		0x70,0x88,0x88,0x70,0x88,0x88,0x70,0x00,0x70,0x88,0x88,0x88,0x78,0x08,0x70,0x00,
		0x00,0x00,0x30,0x30,0x00,0x30,0x30,0x00,0x00,0x00,0x30,0x30,0x00,0x30,0x30,0x60,
		0x10,0x20,0x40,0x80,0x40,0x20,0x10,0x00,0x00,0x00,0xf8,0x00,0xf8,0x00,0x00,0x00,
		0x40,0x20,0x10,0x08,0x10,0x20,0x40,0x00,0x70,0x88,0x08,0x10,0x20,0x00,0x20,0x00,
		0x30,0x48,0x94,0xa4,0xa4,0x94,0x48,0x30,0x70,0x88,0x88,0xf8,0x88,0x88,0x88,0x00,
		0xf0,0x88,0x88,0xf0,0x88,0x88,0xf0,0x00,0x70,0x88,0x80,0x80,0x80,0x88,0x70,0x00,
		0xf0,0x88,0x88,0x88,0x88,0x88,0xf0,0x00,0xf8,0x80,0x80,0xf0,0x80,0x80,0xf8,0x00,
		0xf8,0x80,0x80,0xf0,0x80,0x80,0x80,0x00,0x70,0x88,0x80,0x98,0x88,0x88,0x70,0x00,
		0x88,0x88,0x88,0xf8,0x88,0x88,0x88,0x00,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x00,
		0x08,0x08,0x08,0x08,0x88,0x88,0x70,0x00,0x88,0x90,0xa0,0xc0,0xa0,0x90,0x88,0x00,
		0x80,0x80,0x80,0x80,0x80,0x80,0xf8,0x00,0x88,0xd8,0xa8,0x88,0x88,0x88,0x88,0x00,
		0x88,0xc8,0xa8,0x98,0x88,0x88,0x88,0x00,0x70,0x88,0x88,0x88,0x88,0x88,0x70,0x00,
		0xf0,0x88,0x88,0xf0,0x80,0x80,0x80,0x00,0x70,0x88,0x88,0x88,0x88,0x88,0x70,0x08,
		0xf0,0x88,0x88,0xf0,0x88,0x88,0x88,0x00,0x70,0x88,0x80,0x70,0x08,0x88,0x70,0x00,
		0xf8,0x20,0x20,0x20,0x20,0x20,0x20,0x00,0x88,0x88,0x88,0x88,0x88,0x88,0x70,0x00,
		0x88,0x88,0x88,0x88,0x88,0x50,0x20,0x00,0x88,0x88,0x88,0x88,0xa8,0xd8,0x88,0x00,
		0x88,0x50,0x20,0x20,0x20,0x50,0x88,0x00,0x88,0x88,0x88,0x50,0x20,0x20,0x20,0x00,
		0xf8,0x08,0x10,0x20,0x40,0x80,0xf8,0x00,0x30,0x20,0x20,0x20,0x20,0x20,0x30,0x00,
		0x40,0x40,0x20,0x20,0x10,0x10,0x08,0x08,0x30,0x10,0x10,0x10,0x10,0x10,0x30,0x00,
		0x20,0x50,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xfc,
		0x40,0x20,0x10,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x70,0x08,0x78,0x88,0x78,0x00,
		0x80,0x80,0xf0,0x88,0x88,0x88,0xf0,0x00,0x00,0x00,0x70,0x88,0x80,0x80,0x78,0x00,
		0x08,0x08,0x78,0x88,0x88,0x88,0x78,0x00,0x00,0x00,0x70,0x88,0xf8,0x80,0x78,0x00,
		0x18,0x20,0x70,0x20,0x20,0x20,0x20,0x00,0x00,0x00,0x78,0x88,0x88,0x78,0x08,0x70,
		0x80,0x80,0xf0,0x88,0x88,0x88,0x88,0x00,0x20,0x00,0x20,0x20,0x20,0x20,0x20,0x00,
		0x20,0x00,0x20,0x20,0x20,0x20,0x20,0xc0,0x80,0x80,0x90,0xa0,0xe0,0x90,0x88,0x00,
		0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x00,0x00,0x00,0xf0,0xa8,0xa8,0xa8,0xa8,0x00,
		0x00,0x00,0xb0,0xc8,0x88,0x88,0x88,0x00,0x00,0x00,0x70,0x88,0x88,0x88,0x70,0x00,
		0x00,0x00,0xf0,0x88,0x88,0xf0,0x80,0x80,0x00,0x00,0x78,0x88,0x88,0x78,0x08,0x08,
		0x00,0x00,0xb0,0xc8,0x80,0x80,0x80,0x00,0x00,0x00,0x78,0x80,0x70,0x08,0xf0,0x00,
		0x20,0x20,0x70,0x20,0x20,0x20,0x18,0x00,0x00,0x00,0x88,0x88,0x88,0x98,0x68,0x00,
		0x00,0x00,0x88,0x88,0x88,0x50,0x20,0x00,0x00,0x00,0xa8,0xa8,0xa8,0xa8,0x50,0x00,
		0x00,0x00,0x88,0x50,0x20,0x50,0x88,0x00,0x00,0x00,0x88,0x88,0x88,0x78,0x08,0x70,
		0x00,0x00,0xf8,0x10,0x20,0x40,0xf8,0x00,0x08,0x10,0x10,0x20,0x10,0x10,0x08,0x00,
		0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x40,0x20,0x20,0x10,0x20,0x20,0x40,0x00,
		0x00,0x68,0xb0,0x00,0x00,0x00,0x00,0x00,0x20,0x50,0x20,0x50,0xa8,0x50,0x00,0x00,
};

#define DRAW_COLOR_BLACK draw_color(0,0,0)
#define DRAW_COLOR_GRAY draw_color(196,196,196)
#define DRAW_COLOR_WHITE draw_color(255,255,255)
#define DRAW_COLOR_RED draw_color(255,0,0)
#define DRAW_COLOR_GREEN draw_color(0,255,0)
#define DRAW_COLOR_BLUE draw_color(0,0,255)

static __inline__ unsigned draw_color(unsigned r, unsigned g, unsigned b) {
	if (video_is_graphics()) {
		video_rgb c;
		video_rgb_make(&c,r,g,b);
		return c;
	} else if (video_is_text()) {
		return (r >> 7) | ((g >> 7) << 1) | ((b >> 7) << 2);
	}

	return 0;
}

void draw_graphics_char(int x, int y, char c, unsigned color) {
	if (video_is_graphics()) {
		int i,j;
		unsigned char* p = draw_font + 8*(unsigned)c;
		x *= FONTX;
		y *= FONTY;
			for(i=0;i<FONTY;++i) {
			for(j=0;j<FONTX;++j) {
				if (p[i] & (1 << (7-j)))
					video_put_pixel(x+j,y+i,color);
				else
					video_put_pixel(x+j,y+i,DRAW_COLOR_BLACK);
			}
		}
	} else if (video_is_text()) {
		video_put_char(x,y,c,color);
	}
}

void draw_graphics_string(int x, int y, char* s, unsigned color) {
	unsigned len = strlen(s);
	while (len) {
		draw_graphics_char(x,y,*s, color);
		++s;
		--len;
		++x;
	}
}

/***************************************************************************/
/* Draw generic text */

void draw_text_fill(int x, int y, char c, int dx, unsigned color) {
	while (dx>0) {
		screen_put(x,y,c,color);
		--dx;
		++x;
	}
}

void draw_text_fillrect(int x, int y, char c, int dx, int dy, unsigned color) {
	while (dy-->0) {
		draw_text_fill(x,y++,c,dx,color);
	}
}

void draw_text_rect(int x, int y, int dx, int dy, unsigned color) {
	int i,j;
	screen_put(x,y,'',color);
	screen_put(x+dx-1,y,'',color);
	screen_put(x,y+dy-1,'',color);
	screen_put(x+dx-1,y+dy-1,'',color);
	for(i=1;i+1<dx;++i) {
		screen_put(x+i,y,'',color);
		screen_put(x+i,y+dy-1,'',color);
	}
	for(j=1;j+1<dy;++j) {
		screen_put(x,y+j,'',color);
		screen_put(x+dx-1,y+j,'',color);
	}
}

void draw_text_left(int x, int y, int dx, const char* s, unsigned color) {
	int len = strlen(s);
	int i;
	for(i=0;i<dx;++i) {
		if (i<len)
			screen_put(x+i,y,s[i], color);
		else
			screen_put(x+i,y,' ', color);
	}
}

void draw_text_center(int x, int y, int dx, const char* s, unsigned color) {
	int len = strlen(s);
	int pre;
	int post;
	if (len > dx)
		len = dx;

	pre = (dx - len + 1) / 2;
	post = dx - pre - len;

	draw_text_fill(x,y,' ',pre,color);
	draw_text_left(x+pre,y,len,s,color);
	draw_text_fill(x+pre+len,y,' ',post,color);
}

unsigned draw_text_string(int x, int y, const char* s, unsigned color) {
	int len = strlen(s);
	if (x + len > screen_size_x())
		len = screen_size_x() - x;
	draw_text_left(x,y,len,s,color);
	return len;
}

int draw_text_para(int x, int y, int dx, int dy, const char* s, unsigned color) {
	int iy = 0;
	int ix = 0;
	int needspace = 0;
	while (*s && iy<=dy) {
		unsigned l = strcspn(s," \n");
		if (needspace) {
		if (ix<dx) {
				draw_text_fill(x+ix,y+iy,' ',1,color);
				++ix;
			} else {
				++iy;
				ix = 0;
			}
			needspace = 0;
		}
		if (ix+l>dx) {
			draw_text_fill(x+ix,y+iy,' ',dx-ix,color);
			++iy;
			ix = 0;
			needspace = 0;
		}
		if (iy<=dy) {
			draw_text_left(x+ix,y+iy,l,s,color);
			ix += l;
			s += l;
			if (*s==' ') {
				++s;
				needspace = 1;
			} else if (*s=='\n') {
				draw_text_fill(x+ix,y+iy,' ',dx-ix,color);
				++s;
				++iy;
				ix = 0;
				needspace = 0;
			}
		}
	}
	if (iy<=dy && ix<dx) {
		draw_text_fill(x+ix,y+iy,' ',dx-ix,color);
		++iy;
	}
	return y+iy;
}

int draw_text_read(int x, int y, char* s, int dx, unsigned color) {
        int maxlen = dx - 1;
        if (strlen(s) > maxlen) {
		s[maxlen] = 0;
        }

        while (1) {
		int userkey;
		int len = strlen(s);
                draw_text_string(x,y,s,color);
		draw_text_fill(x+len,y,' ',dx-len,color);
		userkey = screen_getkey();
		switch (userkey) {
			case SCREEN_KEY_BACKSPACE :
			case SCREEN_KEY_DEL :
                        	if (len) {
					s[len-1] = 0;
                                }
                                break;
                        case SCREEN_KEY_ENTER :
			case SCREEN_KEY_ESC :
                                return userkey;
			default:
                        	if (len < maxlen) {
	                        	if (userkey>32 && userkey<=126 && userkey!=',') {
	                                	s[len] = userkey;
	                                	s[len+1] = 0;
	                                }
                                }
                }
        }
}

/***************************************************************************/
/* Draw various graphics */

/* Set the default color palette */
void draw_graphics_palette(void)
{
	if (video_is_graphics()) {
		if (video_index() == VIDEO_FLAGS_INDEX_PACKED) {
			video_palette_to_rgb(0);
		}
	}
}

/* Draw a test screen */
int draw_graphics_default_test(int x, int y) {
	int i,j;
	char buffer[256];

	int TEST_DELTA1 = 10;
	int TEST_DELTA2 = 2;

	draw_graphics_palette();

	if (video_is_graphics()) {
		video_rgb c;

		/* draw out of screen data */
		video_rgb_make(&c,255,0,0);
		for(i=0;i<video_size_x();++i)
			video_put_pixel(i,video_size_y(),c);
		for(j=0;j<video_size_y();++j)
			video_put_pixel(video_size_x(),j,c);

		/* draw background */
		video_rgb_make(&c,0,0,0);
		video_clear(0,0,video_size_x(),video_size_y(),c);

		/* draw border */
		video_rgb_make(&c,255,255,255);
		for(i=0;i<video_size_x();++i) {
			video_put_pixel(i,0,c);
			video_put_pixel(i,video_size_y()-1,c);
		}
		for(j=0;j<video_size_y();++j) {
			video_put_pixel(0,j,c);
			video_put_pixel(video_size_x()-1,j,c);
		}

		/* draw grid */
		video_rgb_make(&c,128,128,128);
		for(i=TEST_DELTA1;i<video_size_x()-1;i += TEST_DELTA1)
			for(j=TEST_DELTA2;j<video_size_y()-1;j += TEST_DELTA2)
				video_put_pixel(i,j,c);
		for(j=TEST_DELTA1;j<video_size_y()-1;j += TEST_DELTA1)
			for(i=TEST_DELTA2;i<video_size_x()-1;i += TEST_DELTA2)
				video_put_pixel(i,j,c);

		/* draw line */
		video_rgb_make(&c,196,196,196);
		for(i=1;i<video_size_x()/2 && i<video_size_y()/2;++i) {
			video_put_pixel(i,i,c);
			video_put_pixel(video_size_x()-i-1,i,c);
			video_put_pixel(i,video_size_y()-i-1,c);
			video_put_pixel(video_size_x()-i-1,video_size_y()-i-1,c);
		}

	} else if (video_is_text()) {

		/* draw out of screen data */
		for(i=0;i<screen_size_x();++i) {
			video_put_char(i,screen_size_y(),'',DRAW_COLOR_RED);
		}
		for(j=0;j<screen_size_y();++j) {
			video_put_char(screen_size_x(),j,'',DRAW_COLOR_RED);
		}

		/* draw border */
		for(i=0;i<screen_size_x();++i) {
			video_put_char(i,0,'',DRAW_COLOR_GRAY);
			video_put_char(i,screen_size_y()-1,'',DRAW_COLOR_GRAY);
		}
		for(j=0;j<screen_size_y();++j) {
			video_put_char(0,j,'',DRAW_COLOR_GRAY);
			video_put_char(screen_size_x()-1,j,'',DRAW_COLOR_GRAY);
		}

		/* draw grid */
		for(i=1;i+1<screen_size_x();++i) {
			for(j=1;j+1<screen_size_y();++j) {
				video_put_char(i,j,'',DRAW_COLOR_GRAY);
			}
		}

	}

	/* draw info */
	if (video_is_graphics()) {
		sprintf(buffer,"%s, graph %dx%d [%dx%d] %d bits",video_name(),video_size_x(), video_size_y(), video_virtual_x(), video_virtual_y(), video_bits_per_pixel());
	} else if (video_is_text()) {
		sprintf(buffer,"%s, text %dx%d [%dx%d]",video_name(),video_size_x(), video_size_y(), screen_size_x(), screen_size_y());
	} else {
		*buffer = 0;
	}

	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	return y;
}

void draw_graphics_ellipse(int xc, int yc, int a0, int b0, unsigned color) {
	int x = 0;
	int y = b0;
	int a = a0;
	int b = b0;
	int as = a*a;
	int as2 = 2*as;
	int bs = b*b;
	int bs2 = 2*bs;
	int d = bs - as*b + as/4;
	int dx = 0;
	int dy = as2 * b;

	while (dx<dy) {
		video_put_pixel(xc+x,yc+y,color);
		video_put_pixel(xc-x,yc+y,color);
		video_put_pixel(xc+x,yc-y,color);
		video_put_pixel(xc-x,yc-y,color);
		if (d>0) {
			--y;
			dy -= as2;
			d -= dy;
		}
		++x;
		dx += bs2;
		d += bs + dx;
	}

	d += (3*(as-bs)/2 - (dx+dy)) / 2;

	while (y>=0) {
		video_put_pixel(xc+x,yc+y,color);
		video_put_pixel(xc-x,yc+y,color);
		video_put_pixel(xc+x,yc-y,color);
		video_put_pixel(xc-x,yc-y,color);
		if (d<0) {
			++x;
			dx += bs2;
			d += dx;
		}
		--y;
		dy -= as2;
		d += as - dy;
	}
}

/* Draw the animation screen */
void draw_graphics_animate(int s_x, int s_y, int s_dx, int s_dy, unsigned counter, int clear) {
	if (video_is_graphics()) {
		video_rgb c;

		int dx = s_dx / (4*4);
		int dy = s_dy /	(3*4);

		int x = counter % ((s_dx - dx) * 2);
		int y = counter % ((s_dy - dy) * 2);

		if (x >= (s_dx - dx))
			x = (s_dx - dx) * 2 - x;
		if (y >= (s_dy - dy))
			y = (s_dy - dy) * 2 - y;

		assert( x + dx <= s_dx );
		assert( y + dy <= s_dy );

		/* draw bar */
		if (clear)
			video_rgb_make(&c,0,0,0);
		else
			video_rgb_make(&c,255,255,255);
		video_clear(s_x + x,s_y,dx,s_dy,c);
		video_clear(s_x,s_y+y,s_dx,dy,c);
	}
}

/* Draw the out of screen data */
void draw_graphics_out_of_screen(int clear) {
	if (video_is_graphics()) {
		video_rgb c;

		/* draw out of screen data */
		if (clear)
			video_rgb_make(&c,0,0,0);
		else
			video_rgb_make(&c,255,0,0);
		video_clear(0,0,video_virtual_x(),video_virtual_y(),c);
	}
}

/* Draw a test calibration screen */
void draw_graphics_calib(int s_x, int s_y, int s_dx, int s_dy) {
	int i,j,k,l;

	if (video_is_graphics()) {
		video_rgb c;

		unsigned dx = s_dx*3/5;
		unsigned dy = s_dy*4/5;
		unsigned x = (s_dx-dx)/2;
		unsigned y = (s_dy-dy)/2;
		unsigned dyB = dy/3;
		unsigned dyM = dyB/2;
		unsigned dyS = dyB/4;

		/* draw background */
		video_rgb_make(&c,128,128,128);
		video_clear(s_x,s_y,s_dx,s_dy,c);

		/* draw rows */
		video_rgb_make(&c,255,255,255);
		for(i=0;i<15;++i)
			video_clear(s_x,s_y + s_dy*i/15+s_dy/30,s_dx,1,c);

		/* draw columns */
		for(i=0;i<20;++i) {
			unsigned rx = s_dx*i/20+s_dx/40;
			for(j=0;j<video_size_y();++j)
				video_put_pixel(s_x+rx,s_y+j,c);
		}

		/* draw ramp */
		for(i=0;i<dx;++i) {
			switch (i*8/dx) {
				case 0 : video_rgb_make(&c,255,255,255); break;
				case 1 : video_rgb_make(&c,255,255,0); break;
				case 2 : video_rgb_make(&c,0,255,255); break;
				case 3 : video_rgb_make(&c,0,255,0); break;
				case 4 : video_rgb_make(&c,255,0,255); break;
				case 5 : video_rgb_make(&c,255,0,0); break;
				case 6 : video_rgb_make(&c,0,0,255); break;
				case 7 : video_rgb_make(&c,0,0,0); break;
			}
			for(j=0;j<dyB;++j)
				video_put_pixel(s_x+x+i,s_y+y+j,c);
		}
		y += dyB;

		for(i=0;i<dx;++i) {
			video_rgb_make(&c,i*255/dx,i*255/dx,i*255/dx);
			for(j=0;j<dyS;++j)
				video_put_pixel(s_x+x+i,s_y+y+j,c);
		}
		y += dyS;


		k = 0;
		l = 0;
		for(i=0;i<dx;++i) {
			if (k > i / (dx / 5)) {
				l = !l;
				k = 0;
			}
			++k;
			if  (l)
				video_rgb_make(&c,0,0,0);
			else
				video_rgb_make(&c,255,255,255);
			for(j=0;j<dyM;++j)
				video_put_pixel(s_x+x+i,s_y+y+j,c);
		}
		y += dyM;

		for(i=0;i<dx;++i) {
			if (i<dx/2)
				video_rgb_make(&c,i*255/(dx/2),0,0);
			else
				video_rgb_make(&c,255,(i-dx/2)*255/(dx/2),(i-dx/2)*255/(dx/2));
			for(j=0;j<dyS;++j)
				video_put_pixel(s_x+x+i,s_y+y+j,c);
		}
		y += dyS;

		for(i=0;i<dx;++i) {
			if (i<dx/2)
				video_rgb_make(&c,0,i*255/(dx/2),0);
			else
				video_rgb_make(&c,(i-dx/2)*255/(dx/2),255,(i-dx/2)*255/(dx/2));
			for(j=0;j<dyS;++j)
				video_put_pixel(s_x+x+i,s_y+y+j,c);
		}
		y += dyS;

		for(i=0;i<dx;++i) {
			if (i<dx/2)
				video_rgb_make(&c,0,0,i*255/(dx/2));
			else
				video_rgb_make(&c,(i-dx/2)*255/(dx/2),(i-dx/2)*255/(dx/2),255);
			for(j=0;j<dyS;++j)
				video_put_pixel(s_x+x+i,s_y+y+j,c);
		}
		y += dyS;

		for(i=0;i<dx;++i) {
			switch (i*5/dx) {
				case 0 : video_rgb_make(&c,0,0,0); break;
				case 1 : video_rgb_make(&c,64,64,64); break;
				case 2 : video_rgb_make(&c,128,128,128); break;
				case 3 : video_rgb_make(&c,196,196,196); break;
				case 4 : video_rgb_make(&c,255,255,255); break;
			}
			for(j=0;j<dyM;++j)
				video_put_pixel(s_x+x+i,s_y+y+j,c);
		}
		y += dyM;


		video_rgb_make(&c,255,255,255);
		draw_graphics_ellipse(s_dx / 2,  s_dy / 2, s_dx * 3 / 9, s_dy * 4 / 9, c);
     	}
}

int draw_graphics_default_command(int x, int y) {
	draw_graphics_string(x,y,"ENTER  Save & Exit",DRAW_COLOR_WHITE);
	++y;
	draw_graphics_string(x,y,"ESC    Exit",DRAW_COLOR_WHITE);
	++y;

	return y;
}

int draw_graphics_vgatweak_test(int x, int y, video_mode* vm) {
	char buffer[256];
	struct tweak_crtc info;
	unsigned memlimit;
	double horzclock;
	double vertclock;

	tweak_crtc_get(&info,tweak_register_get,&vm->driver.vgatweak);

	horzclock = (double)info.horz_clock / 1E3;
	vertclock = (double)info.vert_clock;

	y = draw_graphics_default_test(x,y);

	if (info.is_linear)
		memlimit = 0x10000;
	else
		memlimit = 0x10000*4;

	if (info.size_x * info.size_y > memlimit && info.is_linear) {
		sprintf(buffer,"Memory size %d Kbytes too big",info.size_x * info.size_y / 1024);
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

	++y;

	sprintf(buffer,"Horz  Vert");
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	if (horzclock!=0 || vertclock!=0) {
		sprintf(buffer,"%4.1f %5.1f      Clock [kHz Hz]", horzclock, vertclock);
		draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
		++y;
	}

	sprintf(buffer,"%4d  %4d [qa] Display End",info.hde,info.vde);
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	sprintf(buffer,"%4d  %4d [ws] Blank Start",info.hbs,info.vbs);
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	if (!(info.hde<=info.hbs)) {
		sprintf(buffer,"HDE<=HBS");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(info.vde<=info.vbs)) {
		sprintf(buffer,"VDE<=VBS");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

	sprintf(buffer,"%4d  %4d [ed] Retrace Start",info.hrs,info.vrs);
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	if (!(info.hde<=info.hrs)) {
		sprintf(buffer,"HDE<=HRS");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(info.vde<=info.vrs)) {
		sprintf(buffer,"VDE<=VRS");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(info.hbs<=info.hrs)) {
		sprintf(buffer,"HBS<=HRS");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(info.vbs<=info.vrs)) {
		sprintf(buffer,"VBS<=VRS");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

	sprintf(buffer,"%4d  %4d [rf] Retrace End",info.hre,info.vre);
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	if (!(info.hrs<info.hre)) {
		sprintf(buffer,"HRS<HRE");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
			++y;
	}
	if (!(info.vrs<info.vre)) {
		sprintf(buffer,"VRS<VRE");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

	sprintf(buffer,"%4d  %4d [tg] Blank End",info.hbe,info.vbe);
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	if (!(info.hre<=info.hbe)) {
		sprintf(buffer,"HRE<=HBE");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(info.vre<=info.vbe)) {
		sprintf(buffer,"VRE<=VBE");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

	sprintf(buffer,"%4d  %4d [yh] Total",info.ht,info.vt);
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	if (!(info.hre<=info.ht)) {
		sprintf(buffer,"HRE<=HT");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(info.vre<=info.vt)) {
		sprintf(buffer,"VRE<=VT");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

	if (!(info.hbe<=info.ht)) {
		sprintf(buffer,"HBE<=HT");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(info.vbe<=info.vt)) {
		sprintf(buffer,"VBE<=VT");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

	sprintf(buffer,"   %c     %c [uj] Polarization", info.hsync ? '-' : '+', info.vsync ? '-' : '+');
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	sprintf(buffer,"%4d  %4d [ik] Char Size", info.char_size_x,info.scanlines_per_line);
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	sprintf(buffer,"      %4d [o]  Scanline size", info.bytes_per_scanline);
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	sprintf(buffer,"      %4s [x]  Doublescan", info.is_doublescan ? "on" : "off");
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	sprintf(buffer,"      %4s [c]  Middle clock", info.dotclock_middle ? "on" : "off");
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	sprintf(buffer,"%4s       [v]  Master clock [MHz]", info.dotclock_master ? "28" : "25");
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	++y;
	draw_graphics_string(x,y,"Q...I  Inc horz (SHIFT dec)",DRAW_COLOR_WHITE);
	++y;
	draw_graphics_string(x,y,"A...K  Inc vert (SHIFT dec)",DRAW_COLOR_WHITE);
	++y;
	draw_graphics_string(x,y,"XCV    Flip flag",DRAW_COLOR_WHITE);
	++y;
	draw_graphics_string(x,y,"Z      Compute blank",DRAW_COLOR_WHITE);
	++y;

	y = draw_graphics_default_command(x,y);

	return y;
}

int draw_graphics_crtc_test(int x, int y, video_crtc* crtc, int print_clock, int print_measured_clock, int print_key) {
	char buffer[256];

	sprintf(buffer,"Horz  Vert");
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	if (print_clock && crtc->pixelclock!=0) {
		double horzclock = (double)crtc->pixelclock / (crtc->ht * 1E3);
		double vertclock = (double)crtc->pixelclock / (crtc->ht * crtc->vt);
		if (crtc->doublescan)
			vertclock /= 2;
		if (crtc->interlaced)
			vertclock *= 2;
		sprintf(buffer,"%4.1f %5.1f %sClock Requested [kHz Hz]", horzclock, vertclock, print_key ? "     " : "");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
		++y;
	}

	if (print_measured_clock && crtc->pixelclock!=0) {
		double vertclock = video_measure_vert_freq();
		double horzclock = vertclock * crtc->vt / 1E3;
		if (crtc->doublescan)
			horzclock *= 2;
		if (crtc->interlaced)
			horzclock /= 2;
		sprintf(buffer,"%4.1f %5.1f %sClock Measured [kHz Hz]", horzclock, vertclock, print_key ? "     " : "");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
		++y;
	}

	sprintf(buffer,"%4d  %4d %sDisplay End",crtc->hde,crtc->vde, print_key ? "[qa] " : "");
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

/*
	sprintf(buffer,"%4d  %4d %sBlank Start",crtc->hbs,crtc->vbs, print_key ? "     " : "");
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;
*/
/*	if (!(crtc->hde<=crtc->hbs)) {
		sprintf(buffer,"HDE<=HBS");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
*/
/*	if (!(crtc->vde<=crtc->vbs)) {
		sprintf(buffer,"VDE<=VBS");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
*/

	sprintf(buffer,"%4d  %4d %sRetrace Start",crtc->hrs,crtc->vrs, print_key ? "[ed] " : "");
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;
	if (!(crtc->hde<=crtc->hrs)) {
		sprintf(buffer,"HDE<=HRS");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(crtc->vde<=crtc->vrs)) {
		sprintf(buffer,"VDE<=VRS");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

	sprintf(buffer,"%4d  %4d %sRetrace End",crtc->hre,crtc->vre, print_key ? "[rf] " : "");
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;
	if (!(crtc->hrs<crtc->hre)) {
		sprintf(buffer,"HRS<HRE");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(crtc->vrs<crtc->vre)) {
		sprintf(buffer,"VRE<VRE");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

/*
	sprintf(buffer,"%4d  %4d %sBlank End",crtc->hbe,crtc->vbe, print_key ? "     " : "");
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;
*/
/*
	if (!(crtc->hre<=crtc->hbe)) {
		sprintf(buffer,"HRE<=HBE");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
*/
/*
	if (!(crtc->vre<=crtc->vbe)) {
		sprintf(buffer,"VRE<=VBE");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
*/
	sprintf(buffer,"%4d  %4d %sTotal",crtc->ht,crtc->vt, print_key ? "[yh] " : "");
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;
	if (!(crtc->hre<=crtc->ht)) {
		sprintf(buffer,"HRE<=HT");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(crtc->vre<=crtc->vt)) {
		sprintf(buffer,"VRE<=VT");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

	sprintf(buffer,"   %c     %c %sPolarization", crtc->hsync ? '-' : '+', crtc->vsync ? '-' : '+', print_key ? "[uj] " : "");
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	sprintf(buffer,"      %4s %sDoublescan", crtc->doublescan ? "on" : "off", print_key ? "[x]  " : "");
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	sprintf(buffer,"      %4s %sInterlaced", crtc->interlaced ? "on" : "off", print_key ? "[c]  " : "");
	draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	if (print_clock) {
		sprintf(buffer,"%4.2f      %sPixelclock [MHz]", (double)crtc->pixelclock / 1E6, print_key ? "[v]  " : "");
		draw_graphics_string(x,y,buffer,DRAW_COLOR_WHITE);
		++y;
	}

	if (print_key) {
		++y;
		draw_graphics_string(x,y,"Q...U  Inc horz (SHIFT dec)",DRAW_COLOR_WHITE);
		++y;
		draw_graphics_string(x,y,"A...J  Inc vert (SHIFT dec)",DRAW_COLOR_WHITE);
		++y;
		draw_graphics_string(x,y,"XCV    Flip flag",DRAW_COLOR_WHITE);
		++y;
		draw_graphics_string(x,y,"ARROWS Center",DRAW_COLOR_WHITE);
		++y;
		draw_graphics_string(x,y,"CTRL+ARROWS Expand/Shrink",DRAW_COLOR_WHITE);
		++y;
	}

	return y;
}

int draw_graphics_vgaline_test(int x, int y, video_mode* mode) {
	y = draw_graphics_default_test(x,y);
	++y;

	y = draw_graphics_crtc_test(x,y,&mode->driver.vgaline.crtc,1,1,1);
 	y = draw_graphics_default_command(x,y);

	return y;
}

int draw_graphics_vbeline_test(int x, int y, video_mode* mode) {
	y = draw_graphics_default_test(x,y);

	++y;

	y = draw_graphics_crtc_test(x,y,&mode->driver.vbeline.crtc,1,1,1);
 	y = draw_graphics_default_command(x,y);
	return y;
}

/***************************************************************************/
/* Draw information screens */

int draw_text_warn(void) {
	int x = 0;
	int y = 0;
	int dx = screen_size_x();
	int dy = screen_size_y();
	static int userkey = 0;

	if (userkey != SCREEN_KEY_SPACE) {

	screen_clear();

	draw_text_center(x,y,dx,
"WARNING! Playing with video parameter can DAMAGE YOUR MONITOR!"
	,SCREEN_COLOR_REVERSE);
	++y;

	draw_text_center(x,y,dx,
"IF YOU HEAR STRANGE SOUNDS PRESS ESC IMMEDITIALLY!"
	,SCREEN_COLOR_REVERSE);
	++y;

	y = draw_text_para(x,y,dx,dy-y,
"You can modify two series of parameters, horizontal and vertical. These "
"parameters control the course of the video beam."
"\n\n"
"The display region is the part used for drawing. Start always from 0. "
"The retrace region is the part used by the beam to return backward. "
"The blank region is the part in which the beam is turned off. "
"The total end is the total number of columns and lines."
"\n\n"
"The starting and ending values of these regions are shown in the correct order, from the "
"smallest to the greatest, you must mantain this condition."
"\n\n"
"Others important values are the horizontal and vertical clock frequencies, these "
"values MUST BE COMPATIBLE with your monitor. You MUST check your monitor's manual "
"for getting correct values. "
"Generic old SVGA monitors accept fixed HClock of 31.5 and 35.5 kHz, new multisync monitors "
"accept values in the range 30-70 kHz. The generic VClock's acceptable range is 50-120 Hz."
"\n\n"
"For centering the screen modify the retrace start or use the ARROWS. "
"For changing the resolution modify the display end. "
"For resizing the screen modify the total end and the polarization or use the CTRL+ARROWS. "
"\n\n"
"Now press SPACE to continue or ESC to exit."
	,SCREEN_COLOR_NORMAL);

	do {
		userkey = screen_getkey();
	} while (userkey!=SCREEN_KEY_SPACE && userkey!=SCREEN_KEY_ESC);

	}

	return userkey != SCREEN_KEY_SPACE;
}

int draw_text_warn_bios(void) {
	int x = 0;
	int y = 0;
	int dx = screen_size_x();
	int dy = screen_size_y();
	static int userkey = 0;

	if (userkey != SCREEN_KEY_SPACE) {

	screen_clear();

	draw_text_center(x,y,dx,
"WARNING! Using wrong video mode can DAMAGE YOUR MONITOR!"
	,SCREEN_COLOR_REVERSE);
	++y;

	draw_text_center(x,y,dx,
"IF YOU HEAR STRANGE SOUNDS PRESS ESC IMMEDITIALLY!"
	,SCREEN_COLOR_REVERSE);
	++y;

	++y;
	y = draw_text_para(x,y,dx,dy-y,
"You are going to test a BIOS vga or vbe mode with a low frequency monitor.\n"
"Usually these modes work only with normal high frequency monitors. "
"So proceed only if you know that you are doing."
"\n\n"
"Now press SPACE to continue or ESC to exit."
	,SCREEN_COLOR_NORMAL);

	do {
		userkey = screen_getkey();
	} while (userkey!=SCREEN_KEY_SPACE &&  userkey!=SCREEN_KEY_ESC);

	}

	return userkey != SCREEN_KEY_SPACE;
}

int draw_text_help(void) {
	int x = 0;
	int y = 0;
	int dx = screen_size_x();
	int dy = screen_size_y();
	int userkey = 0;

	screen_clear();

	y = draw_text_para(x,y,dx,dy,
" HELP"
	,SCREEN_COLOR_REVERSE);

	y = draw_text_para(x,y,dx,dy-y,
"F2    Save selected modes in mame.cfg\n"
"F5    Create a new video mode based on the current using the GTF formula\n"
"F6    Duplicate the current video mode grabbing hardware parameteres\n"
"F7    Change the mode number of vgaline/vbeline modes\n"
"F8    Export a C registers structure of the current video mode to modes.c\n"
"F9    Show a static test screen for the current video mode\n"
"F10   Show a dynamic test screen for the current video mode\n"
"SPACE Select/Unselect the current video mode\n"
"ENTER Test the current video mode\n"
"TAB   Rename the current video mode\n"
"ESC   Exit\n"
"\n"
"q/a   Increase the x/y Display End register (SHIFT for decreasing)\n"
"w/s   Increase the x/y Blank Start register (SHIFT for decreasing)\n"
"e/d   Increase the x/y Retrace Start register (SHIFT for decreasing)\n"
"r/f   Increase the x/y Retrace End register (SHIFT for decreasing)\n"
"t/g   Increase the x/y Blank End register (SHIFT for decreasing)\n"
"y/h   Increase the x/y Total End register (SHIFT for decreasing)\n"
"v     Increase the pixel clock (SHIFT for decreasing)\n"
"u/j   Switch the x/y polarization\n"
"x     On/Off the doublescan mode\n"
"c     On/Off the interlace mode\n"
"\n"
"Now press ESC"
	,SCREEN_COLOR_NORMAL);

	do {
		userkey = screen_getkey();
	} while (userkey!=SCREEN_KEY_ESC);

	return userkey;
}

int draw_text_error(void) {
	int x = 0;
	int y = 0;
	int dx = screen_size_x();
	int dy = screen_size_y();
	static int userkey = 0;

	screen_clear();

	draw_text_center(x,y,dx,
"ERROR! An error occour in your last action!"
	,SCREEN_COLOR_REVERSE);
	++y;

	++y;
	y = draw_text_para(x,y,dx,dy,
"Your last action generated an error. Probably you have requested an "
"unsupported feature by your hardware or software."
	,SCREEN_COLOR_NORMAL);

	if (*video_error_description_get()) {
		y = draw_text_para(x,y,dx,dy-y,"\nThe video software report this error:",SCREEN_COLOR_NORMAL);
		y = draw_text_para(x,y,dx,dy-y,video_error_description_get(),SCREEN_COLOR_ERROR);
	}

	y = draw_text_para(x,y,dx,dy-y,"\nNow press ESC",SCREEN_COLOR_NORMAL);

	do {
		userkey = screen_getkey();
	} while (userkey!=SCREEN_KEY_ESC);

	return userkey;
}

void draw_text_menu_item(int x, int y, int dx, int pos, int selected) {
	char buffer[256];
	video_mode* vm = video_mode_container_pos(&vm_all,pos);

	if (vm) {
		const char* driver;
		float aspect = (float)vm->aspect_x / vm->aspect_y;
		char tag;
		unsigned color;
		char type;
		char vfreq[8];
		char hfreq[8];
		unsigned mode;

		switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
			case VIDEO_FLAGS_DRIVER_VGA :
				driver = "vga";
				mode = vm->driver.vga.mode;
				break;
			case VIDEO_FLAGS_DRIVER_VGATWEAK :
				driver = "vgatweak";
				mode = vm->driver.vgatweak.mode;
				break;
			case VIDEO_FLAGS_DRIVER_VGALINE :
				driver = "vgaline";
				mode = vm->driver.vgaline.mode;
				break;
			case VIDEO_FLAGS_DRIVER_VBE :
				driver = "vbe";
				mode = vm->driver.vbe.mode;
				break;
			case VIDEO_FLAGS_DRIVER_VBELINE :
				driver = "vbeline";
				mode = vm->driver.vbeline.mode;
				break;
			default:
				driver = "unknow";
				mode = 0;
		}

		switch (vm->flags & VIDEO_FLAGS_TYPE_MASK) {
			case VIDEO_FLAGS_TYPE_TEXT : type = 'T'; break;
			case VIDEO_FLAGS_TYPE_GRAPHICS : type = 'G'; break;
			default: type = '?';
		}

		if (selected) {
			if (vm->flags & VIDEO_FLAGS_USER_BIT0) {
				tag = '';
	 			color = SCREEN_COLOR_SELECTED_MARK;
			} else {
				tag = ' ';
				color = SCREEN_COLOR_SELECTED;
			}
		} else {
			if (vm->flags & VIDEO_FLAGS_USER_BIT0) {
				tag = '';
				if (monitor_check(vm)) {
	 				color = SCREEN_COLOR_MARK;
				} else {
					color = SCREEN_COLOR_MARK_BAD;
				}
			} else {
				tag = ' ';
				if (monitor_check(vm)) {
					color = SCREEN_COLOR_NORMAL;
				} else {
					color = SCREEN_COLOR_BAD;
				}
			}
		}

		if (vm->vert_freq) {
			sprintf(vfreq,"%5.1f",(double)vm->vert_freq);
		} else {
			sprintf(vfreq,"     ");
		}

		if (vm->horz_freq) {
			sprintf(hfreq,"%5.1f",(double)vm->horz_freq / 1E3);
		} else {
			sprintf(hfreq,"     ");
		}

		sprintf(buffer," %c %8s(%4x) %c %4dx%4d %2d %s %s %4.2f %s",
			tag,
			driver,
			mode,
			type,
			vm->size_x,
			vm->size_y,
			vm->bits,
			vfreq,
			hfreq,
			(double)aspect,
			vm->name
		);

		draw_text_left(x,y,dx,buffer,color);

	} else {
		draw_text_fill(x,y,' ',dx,SCREEN_COLOR_NORMAL);
	}
}

void draw_text_menu(int x, int y, int dx, int dy) {
	unsigned i;
	for(i=0;i<dy;++i) {
		if (menu_base + i < menu_max) {
			draw_text_menu_item(x,y+i,dx, menu_base + i, i == menu_rel);
		} else
			draw_text_fill(x,y+i,' ',dx,SCREEN_COLOR_NORMAL);
	}
}

/***************************************************************************/
/* Draw information bars */

void format_info_vgatweak(char* buffer0, char* buffer1, char* buffer2, vgatweak_video_mode* mode) {
	struct tweak_crtc info;
	unsigned xdup;
	unsigned ydup;

	tweak_crtc_get(&info,tweak_register_get,mode);

	xdup = info.pixels_per_clock;
	if (!info.is_graphics_mode) {
		ydup = 1;
	} else {
		ydup = info.scanlines_per_line;
	}
	sprintf(buffer0,"plz  freq dsen blst rtst rten blen totl ch type pclock scan");
	sprintf(buffer1,"h%c %6.2f%5d%5d%5d%5d%5d%5d%3d %s %6.2f %4d",info.hsync ? '-' : '+',(double)info.horz_clock/1E3,info.hde*xdup,info.hbs*xdup,info.hrs*xdup,info.hre*xdup,info.hbe*xdup,info.ht*xdup, info.char_size_x, !info.is_graphics_mode ? "text" : "grph" ,(double)(info.pixel_clock/1E6),info.bytes_per_scanline);
	sprintf(buffer2,"v%c %6.2f%5d%5d%5d%5d%5d%5d%3d%s%s",info.vsync ? '-' : '+',(double)info.vert_clock,info.vde/ydup,info.vbs/ydup,info.vrs/ydup,info.vre/ydup,info.vbe/ydup,info.vt/ydup, info.scanlines_per_line, info.is_doublescan ? " doublescan" : "", info.vbe != info.vbe7 ? " blank8bit" : "");
}

void format_info_crtc(char* buffer0, char* buffer1, char* buffer2,video_crtc* crtc) {
	double horzclock = (double)crtc->pixelclock / (crtc->ht * 1E3);
	double vertclock = (double)crtc->pixelclock / (crtc->ht * crtc->vt);

	if (crtc->doublescan)
		vertclock /= 2;
	if (crtc->interlaced)
		vertclock *= 2;

	sprintf(buffer0,"plz  freq dsen blst rtst rten blen totl ch type pclock scan");
	sprintf(buffer1,"h%c %6.2f%5d     %5d%5d     %5d         %6.2f",crtc->hsync ? '-' : '+',horzclock,crtc->hde,crtc->hrs,crtc->hre,crtc->ht,(double)crtc->pixelclock / 1E6);
	sprintf(buffer2,"v%c %6.2f%5d     %5d%5d     %5d   %s%s",crtc->vsync ? '-' : '+',vertclock,crtc->vde,crtc->vrs,crtc->vre,crtc->vt,crtc->doublescan ? " doublescan" : "", crtc->interlaced ? " interlaced" : "");
}

void format_info(char* buffer0, char* buffer1, char* buffer2, video_mode* vm) {
	*buffer0 = 0;
	*buffer1 = 0;
	*buffer2 = 0;

	if (vm) {
		switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
			case VIDEO_FLAGS_DRIVER_VGATWEAK :
				format_info_vgatweak(buffer0,buffer1,buffer2,&vm->driver.vgatweak);
				break;
			case VIDEO_FLAGS_DRIVER_VGALINE :
				format_info_crtc(buffer0,buffer1,buffer2,&vm->driver.vgaline.crtc);
				break;
			case VIDEO_FLAGS_DRIVER_VBELINE :
				format_info_crtc(buffer0,buffer1,buffer2,&vm->driver.vbeline.crtc);
				break;
		}
	}
}

void draw_text_info(int x, int y, int dx, int dy, int pos) {
	char buffer[3][256];
	video_mode* vm = video_mode_container_pos(&vm_all,pos);
	format_info(buffer[0],buffer[1],buffer[2],vm);

	draw_text_left(x,y+0,dx,buffer[0],SCREEN_COLOR_INFO_TITLE);
	draw_text_left(x,y+1,dx,buffer[1],SCREEN_COLOR_INFO_NORMAL);
	draw_text_left(x,y+2,dx,buffer[2],SCREEN_COLOR_INFO_NORMAL);

	if (vm) {
		if (!monitor_check(vm)) {
			draw_text_left(x+dx-14,y,14," OUT OF RANGE ", SCREEN_COLOR_ERROR);
		}
	} else {
		draw_text_fillrect(x,y,' ',dx,dy,SCREEN_COLOR_INFO_NORMAL);
	}
}

void draw_text_bar(int x, int y1, int y2, int dx) {
	char buffer[256];
	sprintf(buffer," Select a subset of video modes");
	draw_text_left(x,y1,dx,buffer,SCREEN_COLOR_BAR);

	if (card_active()) {
		sprintf(buffer,"[%s] ",card_driver_name());
	} else {
		sprintf(buffer,"[no] ");
	}
	draw_text_left(x + dx - strlen(buffer),y1,strlen(buffer),buffer,SCREEN_COLOR_BAR);

	sprintf(buffer,
" #   driver(mode) m  x    y    c vfreq hfreq x/y  name");
	draw_text_left(x,y1+1,dx,buffer,SCREEN_COLOR_TITLE);

	sprintf(buffer," F1 Help  F2 Save  SPACE Select  TAB Rename  DEL Del  ENTER Test  ESC Exit");
	draw_text_left(x,y2,dx,buffer,SCREEN_COLOR_BAR);
}

/***************************************************************************/
/* Execute userkey action */

void tweak_vert_scale(unsigned _mul, unsigned _div, vgatweak_video_mode* mode, struct tweak_crtc* info) {
	tweak_vde_set(info->vde * _mul / _div, tweak_register_set, tweak_register_get, mode);
	tweak_vbs_set(info->vbs * _mul / _div, tweak_register_set, tweak_register_get, mode);
	tweak_vrs_set(info->vrs * _mul / _div, tweak_register_set, tweak_register_get, mode);
	tweak_vre_set(info->vre * _mul / _div, tweak_register_set, tweak_register_get, mode);
	tweak_vbe_set(info->vbe * _mul / _div, tweak_register_set, tweak_register_get, mode);
	tweak_vtt_set(info->vt * _mul / _div, tweak_register_set, tweak_register_get, mode);
}

int exe_test_vgatweak(int userkey, vgatweak_video_mode* mode, struct tweak_crtc* info) {
	int modify = 0;
	switch (userkey) {
		case 'o' :
			modify = 1;
			tweak_offset_set(info->clocks_per_scanline/(2*(info->is_dword_mode+1)) + 1, tweak_register_set, tweak_register_get, mode);
			break;
		case 'O' :
			modify = 1;
			tweak_offset_set(info->clocks_per_scanline/(2*(info->is_dword_mode+1)) - 1, tweak_register_set, tweak_register_get, mode);
			break;
		case 'i' :
			modify = 1;
			tweak_char_size_x_set(info->char_size_x + 1, tweak_register_set, tweak_register_get, mode);
			break;
		case 'I' :
			modify = 1;
			tweak_char_size_x_set(info->char_size_x - 1, tweak_register_set, tweak_register_get, mode);
			break;
		case 'k' :
			modify = 1;
			tweak_vert_scale(info->scanlines_per_line+1,info->scanlines_per_line,mode,info);
			tweak_char_size_y_set(info->scanlines_per_line + 1 + info->is_doublescan, tweak_register_set, tweak_register_get, mode);
			break;
		case 'K' :
			modify = 1;
			tweak_vert_scale(info->scanlines_per_line-1,info->scanlines_per_line,mode,info);
			tweak_char_size_y_set(info->scanlines_per_line - 1 - info->is_doublescan, tweak_register_set, tweak_register_get, mode);
			break;
		case 'x' :
		case 'X' :
			modify = 1;
			if (info->is_doublescan)
				tweak_vert_scale(1,2,mode,info);
			else
				tweak_vert_scale(2,1,mode,info);
			tweak_doublescan_set(!info->is_doublescan, tweak_register_set, tweak_register_get, mode);
			break;
		case 'u' :
		case 'U' :
			modify = 1;
			tweak_hsync_set(!info->hsync, tweak_register_set, tweak_register_get, mode);
			break;
		case 'j' :
		case 'J' :
			modify = 1;
			tweak_vsync_set(!info->vsync, tweak_register_set, tweak_register_get, mode);
			break;
		case 'c' :
		case 'C' :
			modify = 1;
			tweak_dotclock_middle_set(!info->dotclock_middle, tweak_register_set, tweak_register_get, mode);
			break;
		case 'v' :
		case 'V' :
			modify = 1;
			tweak_masterclock_input_set(!info->dotclock_master, tweak_register_set, tweak_register_get, mode);
			break;
		case 'z' :
		case 'Z' :
			modify = 1;
			info->hbs = (info->hrs + info->hde) / 2;
			info->hbe = (info->hre + info->ht) / 2;
			tweak_hbs_set(info->hbs, tweak_register_set, tweak_register_get, mode);
			tweak_hbe_set(info->hbe, tweak_register_set, tweak_register_get, mode);
			info->vbs = (info->vrs + info->vde) / 2;
			info->vbe = (info->vre + info->vt) / 2;
			tweak_vbs_set(info->vbs, tweak_register_set, tweak_register_get, mode);
			tweak_vbe_set(info->vbe, tweak_register_set, tweak_register_get, mode);
			break;
		case 'q' :
			tweak_hde_set(info->hde + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'Q' :
			tweak_hde_set(info->hde - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'w' :
			tweak_hbs_set(info->hbs + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'W' :
			tweak_hbs_set(info->hbs - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'e' :
			tweak_hrs_set(info->hrs + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'E' :
			tweak_hrs_set(info->hrs - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'r' :
			tweak_hre_set(info->hre + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'R' :
			tweak_hre_set(info->hre - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 't' :
			tweak_hbe_set(info->hbe + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'T' :
			tweak_hbe_set(info->hbe - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'y' :
			tweak_htt_set(info->ht + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'Y' :
			tweak_htt_set(info->ht - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'a' :
			tweak_vde_set(info->vde + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'A' :
			tweak_vde_set(info->vde - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 's' :
			tweak_vbs_set(info->vbs + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'S' :
			tweak_vbs_set(info->vbs - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'd' :
			tweak_vrs_set(info->vrs + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'D' :
			tweak_vrs_set(info->vrs - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'f' :
			tweak_vre_set(info->vre + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'F' :
			tweak_vre_set(info->vre - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'g' :
			tweak_vbe_set(info->vbe + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'G' :
			tweak_vbe_set(info->vbe - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'h' :
			tweak_vtt_set(info->vt + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'H' :
			tweak_vtt_set(info->vt - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
	}
	return modify;
}

int exe_test_crtc(int userkey, video_crtc* crtc, int xdelta) {
	int modify = 0;
	unsigned pred_t;
	switch (userkey) {
		case SCREEN_KEY_CTRLLEFT :
			pred_t = crtc->ht;
			crtc->ht &= ~(xdelta - 1);
			crtc->ht -= xdelta;
			crtc->pixelclock = (double)crtc->pixelclock * crtc->ht / pred_t;
			modify = 1;
			break;
		case SCREEN_KEY_CTRLRIGHT :
			pred_t = crtc->ht;
			crtc->ht &= ~(xdelta - 1);
			crtc->ht += xdelta;
			crtc->pixelclock = (double)crtc->pixelclock * crtc->ht / pred_t;
			modify = 1;
			break;
		case SCREEN_KEY_CTRLUP :
			crtc->vt -= 1;
			modify = 1;
			break;
		case SCREEN_KEY_CTRLDOWN :
			crtc->vt += 1;
			modify = 1;
			break;
		case SCREEN_KEY_LEFT :
			crtc->hrs &= ~(xdelta - 1);
			crtc->hrs += xdelta;
			crtc->hre &= ~(xdelta - 1);
			crtc->hre += xdelta;
			modify = 1;
			break;
		case SCREEN_KEY_RIGHT :
			crtc->hrs &= ~(xdelta - 1);
			crtc->hrs -= xdelta;
			crtc->hre &= ~(xdelta - 1);
			crtc->hre -= xdelta;
			modify = 1;
			break;
		case SCREEN_KEY_DOWN :
			crtc->vrs -= 1;
			crtc->vre -= 1;
			modify = 1;
			break;
		case SCREEN_KEY_UP :
			crtc->vrs += 1;
			crtc->vre += 1;
			modify = 1;
			break;
		case 'x' :
		case 'X' :
			modify = 1;
			crtc->doublescan = !crtc->doublescan;
			break;
		case 'u' :
		case 'U' :
			modify = 1;
			crtc->hsync = !crtc->hsync;
			break;
		case 'j' :
		case 'J' :
			modify = 1;
			crtc->vsync = !crtc->vsync;
			break;
		case 'e' :
			crtc->hrs &= ~(xdelta - 1);
			crtc->hrs += xdelta;
			modify = 1;
			break;
		case 'E' :
			crtc->hrs &= ~(xdelta - 1);
			crtc->hrs -= xdelta;
			modify = 1;
			break;
		case 'r' :
			crtc->hre &= ~(xdelta - 1);
			crtc->hre += xdelta;
			modify = 1;
			break;
		case 'R' :
			crtc->hre &= ~(xdelta - 1);
			crtc->hre -= xdelta;
			modify = 1;
			break;
		case 'y' :
			crtc->ht &= ~(xdelta - 1);
			crtc->ht += xdelta;
			modify = 1;
			break;
		case 'Y' :
			crtc->ht &= ~(xdelta - 1);
			crtc->ht -= xdelta;
			modify = 1;
			break;
		case 'q' :
			crtc->hde &= ~(xdelta - 1);
			crtc->hde += xdelta;
			modify = 1;
			break;
		case 'Q' :
			crtc->hde &= ~(xdelta - 1);
			crtc->hde -= xdelta;
			modify = 1;
			break;
		case 'a' :
			++crtc->vde;
			modify = 1;
			break;
		case 'A' :
			--crtc->vde;
			modify = 1;
			break;
		case 'd' :
			++crtc->vrs;
			modify = 1;
			break;
		case 'D' :
			--crtc->vrs;
			modify = 1;
			break;
		case 'f' :
			++crtc->vre;
			modify = 1;
			break;
		case 'F' :
			--crtc->vre;
			modify = 1;
			break;
		case 'h' :
			++crtc->vt;
			modify = 1;
			break;
		case 'H' :
			--crtc->vt;
			modify = 1;
			break;
	}
	return modify;
}

int exe_test_vgaline(int userkey, vgaline_video_mode* mode) {
	video_crtc* crtc = &mode->crtc;
	unsigned oldpixelclock = crtc->pixelclock;
	int modify = exe_test_crtc(userkey,crtc,4);
	switch (userkey) {
		case 'v' :
			vga_pixelclock_getnext(&crtc->pixelclock,mode->mode);
			modify = oldpixelclock != crtc->pixelclock;
			break;
		case 'V' :
			vga_pixelclock_getpred(&crtc->pixelclock,mode->mode);
			modify = oldpixelclock != crtc->pixelclock;
			break;
	}

	return modify;
}

int exe_test_vbeline(int userkey, vbeline_video_mode* mode) {
	video_crtc* crtc = &mode->crtc;
	unsigned oldpixelclock = crtc->pixelclock;
	int modify = exe_test_crtc(userkey,crtc,4);
	switch (userkey) {
		case 'c' :
		case 'C' :
			modify = 1;
			crtc->interlaced = !crtc->interlaced;
			break;
		case 'v' :
			card_pixelclock_getnext(&crtc->pixelclock,mode->mode);
			modify = oldpixelclock != crtc->pixelclock;
			break;
		case 'V' :
			card_pixelclock_getpred(&crtc->pixelclock,mode->mode);
			modify = oldpixelclock != crtc->pixelclock;
			break;
	}
	return modify;
}

void exe_vsync(video_mode* vm) {
	uclock_t start,stop;

	/* inizialize uclock (needed to reprogram the PIC) */
	uclock();

	video_wait_vsync();
	start = uclock();
	video_wait_vsync();
	stop = uclock();

	if (stop!=start)
		vm->vert_freq = (double)UCLOCKS_PER_SEC/(stop-start);
	else
		vm->vert_freq = 0;

	/* check valid value */
	if (vm->vert_freq<25 || vm->vert_freq>200)
		vm->vert_freq = 0;
}

/***************************************************************************/
/* Menu command */

void cmd_select(int i) {
	video_mode* vm;

	if (i >= menu_max)
		return;

	vm = video_mode_container_pos(&vm_all,i);
	if (vm->flags & VIDEO_FLAGS_USER_BIT0)
		vm->flags &= ~VIDEO_FLAGS_USER_BIT0;
	else
		vm->flags |= VIDEO_FLAGS_USER_BIT0;

	vm_all_modified = 1;
}

int cmd_offvideo_test(int i, int userkey) {
	video_mode* vm;
	struct tweak_crtc info;
	video_mode vm_save;
	int vm_save_modified;
	int modify = 0;

	if (i >= menu_max)
		return 0;

	vm = video_mode_container_pos(&vm_all,i);

	vm_save = *vm;
	vm_save_modified = vm_all_modified;

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			tweak_crtc_get(&info, tweak_register_get, &vm->driver.vgatweak);
			modify = exe_test_vgatweak(userkey,&vm->driver.vgatweak,&info);
			break;
		case VIDEO_FLAGS_DRIVER_VGALINE :
			modify = exe_test_vgaline(userkey,&vm->driver.vgaline);
			break;
		case VIDEO_FLAGS_DRIVER_VBELINE :
			modify = exe_test_vbeline(userkey,&vm->driver.vbeline);
			break;
		default:
			modify = 0;
	}

	vm_all_modified =  vm_all_modified || modify;

	if (modify) {
		if (video_mode_import(vm)!=0) {
			/* abort */
			*vm = vm_save;
			vm_all_modified = vm_save_modified;
			return -1;
		}

		if (strcmp(vm->name,DEFAULT_TEXT_MODE)==0) {
			if (!monitor_check(vm) && monitor_check(&vm_save)) {
				*vm = vm_save;
				vm_all_modified = vm_save_modified;

				sound_error();

				return 0;
			}
			screen_reset();
		}
	} else {
		sound_warn();
	}

	return 0;
}

int cmd_onvideo_test(int i) {
	video_mode* vm;
	int done ;
	video_mode vm_save;
	int vm_save_modified;
	int dirty = 1;

	if (i >= menu_max)
		return 0;

	vm = video_mode_container_pos(&vm_all,i);

	if (!monitor_check(vm))
		return -1;

	if (!monitor_check_safe(vm)) {
		if (draw_text_warn_bios() != 0)
			return -1;
	}

 	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
		case VIDEO_FLAGS_DRIVER_VBELINE :
			if (draw_text_warn()!=0)
				return 0;
		break;
	}

	if (video_mode_set(vm)!=0) {
		return -1;
	}

	screen_set_font();

	vm_save = *vm;
	vm_save_modified = vm_all_modified;

	done = 0;
	while (!done) {
		int userkey;
		int modify = 0;

		int x = 1; /* default border */
		int y = 1;
		struct tweak_crtc info;
		video_crtc crtc;

		video_mode vm_last = *vm;
		int vm_last_modified = vm_all_modified;

		if (dirty) {
			double vert_freq;
			switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
				case VIDEO_FLAGS_DRIVER_VGATWEAK :
					/* import specific vgatweak information */
					draw_graphics_vgatweak_test(x,y,vm);
					break;
				case VIDEO_FLAGS_DRIVER_VGALINE :
					draw_graphics_vgaline_test(x,y,vm);
					break;
				case VIDEO_FLAGS_DRIVER_VBELINE :
					draw_graphics_vbeline_test(x,y,vm);
					break;
				case VIDEO_FLAGS_DRIVER_VGA :
				case VIDEO_FLAGS_DRIVER_VBE :
					y = draw_graphics_default_test(x,y);
					tweak_crtc_get(&info,tweak_reg_read,0);
					vert_freq = video_measure_vert_freq();
					if (video_crtc_import(&crtc,&info,vm,vert_freq)==0) {
						++y;
						y = draw_graphics_crtc_test(x,y,&crtc,0,1,0);
					}
					++y;
					y = draw_graphics_default_command(x,y);
					break;
			}
			dirty = 0;
		}

		userkey = screen_getkey();
		switch (userkey) {
			case SCREEN_KEY_ESC :
				done = 1;
				/* restore */
				*vm = vm_save;
				vm_all_modified = vm_save_modified;
				break;
			case SCREEN_KEY_ENTER :
				done = 1;
				break;
			case 'm' :
			case 'M' :
				exe_vsync(vm);
				break;
		}
		switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
			case VIDEO_FLAGS_DRIVER_VGATWEAK :
				modify = exe_test_vgatweak(userkey,&vm->driver.vgatweak,&info);
				break;
			case VIDEO_FLAGS_DRIVER_VGALINE :
				modify = exe_test_vgaline(userkey,&vm->driver.vgaline);
				break;
			case VIDEO_FLAGS_DRIVER_VBELINE :
				modify = exe_test_vbeline(userkey,&vm->driver.vbeline);
				break;
			default:
				modify = 0;
		}

		vm_all_modified =  vm_all_modified || modify;

		if (!done) {
			if (modify) {
				dirty = 1;

				if (video_mode_import(vm)==0) {
					if (!monitor_check(vm) && monitor_check(&vm_last)) {
						/* restore */
						*vm = vm_last;
						vm_all_modified = vm_last_modified;

						dirty = 0;
						sound_error();
					} else {
						if (video_mode_set(vm) != 0) {
							/* abort */
							*vm = vm_save;
							vm_all_modified = vm_save_modified;
							return -1;
						}
						screen_set_font();
					}
				} else {
					/* abort */
					*vm = vm_save;
					vm_all_modified = vm_save_modified;
					return -1;
				}
			} else {
				sound_warn();
			}
		}
	}

	return 0;
}

int cmd_onvideo_calib(int i) {
	video_mode* vm;

	if (i >= menu_max)
		return 0;

	vm = video_mode_container_pos(&vm_all,i);

	if (!video_mode_is_graphics(vm)) {
		video_error_description_set("Calibration screen not supported in text modes");
		return -1;
	}

	if (!monitor_check(vm))
		return -1;

	if (!monitor_check_safe(vm)) {
		if (draw_text_warn_bios() != 0)
			return -1;
	}

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
		case VIDEO_FLAGS_DRIVER_VBELINE :
			if (draw_text_warn() != 0)
				return 0;
		break;
	}

	if (video_mode_set(vm)!=0) {
		return -1;
	}

	draw_graphics_palette();
	draw_graphics_out_of_screen(0);
	draw_graphics_calib(0,0,video_size_x(),video_size_y());

	while (screen_kbhit())
		screen_getkey();

	screen_getkey();

	return 0;
}

int cmd_onvideo_animate(int i) {
	video_mode* vm;
	unsigned counter = 0;

	if (i >= menu_max)
		return 0;

	vm = video_mode_container_pos(&vm_all,i);

	if (!video_mode_is_graphics(vm)) {
		video_error_description_set("Calibration screen not supported in text modes");
		return -1;
	}

	if (!monitor_check(vm))
		return -1;

	if (!monitor_check_safe(vm)) {
		if (draw_text_warn_bios() != 0)
			return -1;
	}

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
		case VIDEO_FLAGS_DRIVER_VBELINE :
			if (draw_text_warn() != 0)
				return 0;
		break;
	}

	if (video_mode_set(vm)!=0) {
		return -1;
	}

	video_update_init(3);

	draw_graphics_palette();
	draw_graphics_out_of_screen(1);

	while (screen_kbhit())
		screen_getkey();

	while (!screen_kbhit()) {
		video_update_start();
		draw_graphics_animate(video_update_x(),video_update_y(),video_size_x(),video_size_y(), counter - video_update_page_max() + 1, 1);
		++counter;
		draw_graphics_animate(video_update_x(),video_update_y(),video_size_x(),video_size_y(), counter, 0);
		video_update_stop(1);
	}

	video_update_done();

	screen_getkey();

	return 0;
}

void cmd_gotopos(int i) {
	if (i >= menu_max)
		i = menu_max - 1;
	if (i<0)
		i = 0;
	if (menu_base <= i && i < menu_base + menu_rel_max) {
		menu_rel = i - menu_base;
	} else if (i<menu_base) {
		menu_base = i;
		menu_rel = 0;
	} else {
		menu_base = i - menu_rel_max + 1;
		if (menu_base<0)
			menu_base = 0;
		menu_rel = i - menu_base;
	}
}

int cmd_input_key(const char* tag, const char* keys) {
	draw_text_fill(0,screen_size_y()-1,' ',screen_size_x(),SCREEN_COLOR_REVERSE);
	draw_text_string(2,screen_size_y()-1,tag,SCREEN_COLOR_REVERSE);
	draw_text_string(2+strlen(tag),screen_size_y()-1,keys,SCREEN_COLOR_INPUT);

	while (1) {
		int i;
		int key = screen_getkey();
		if (key == SCREEN_KEY_ESC)
			return -1;

		for(i=0;keys[i];++i)
			if (toupper(key)==toupper(keys[i]))
				return i;
	}
}

int cmd_input_string(const char* tag, char* buffer, unsigned length) {
	draw_text_fill(0,screen_size_y()-1,' ',screen_size_x(),SCREEN_COLOR_REVERSE);
	draw_text_string(2,screen_size_y()-1,tag,SCREEN_COLOR_REVERSE);

	if (draw_text_read(2 + strlen(tag),screen_size_y()-1,buffer,length,SCREEN_COLOR_INPUT) == SCREEN_KEY_ENTER) {
		return 0;
	}

	return -1;
}

void cmd_rename(void) {
	video_mode* vm;
	char buffer[128];

	int i = menu_base + menu_rel;
	if (i >= menu_max)
		return;

	vm = video_mode_container_pos(&vm_all,i);

	strcpy(buffer,"");
	if (cmd_input_string(" Name : ",buffer,40)!=0)
 		return;

	strcpy(vm->name,buffer);
	vm_all_modified = 1;
}

void cmd_mode_change(void) {
	video_mode* vm;
	char buffer[128];
	unsigned mode;

	int i = menu_base + menu_rel;
	if (i >= menu_max)
		return;

	vm = video_mode_container_pos(&vm_all,i);

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VBE :
			sound_error();
			return;
	}

	strcpy(buffer,"");
	if (cmd_input_string(" Mode number [hex] : ",buffer,8)!=0)
		return;

	mode = strtol(buffer,0,16);
	if (!mode)
		return;

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			vm->driver.vgatweak.mode = mode;
		break;
		case VIDEO_FLAGS_DRIVER_VGALINE :
			vm->driver.vgaline.mode = mode;
		break;
		case VIDEO_FLAGS_DRIVER_VBELINE :
			vm->driver.vbeline.mode = mode;
		break;
	}

	vm_all_modified = 1;
}

void video_crtc_compute(video_crtc* crtc, unsigned xres, unsigned yres, int want_interlaced, double freq, int freq_type) {
	GTF_timings gtf;
	unsigned doublescan;
	unsigned ydup;

	if (xres < 640 && yres < 400) {
		doublescan = 1;
		ydup = 2;
	} else {
		doublescan = 0;
		ydup = 1;
	}

	yres *= ydup;


	switch (freq_type) {
		case GTF_lockVF : break;
		case GTF_lockHF : freq /= 1E3; break;
		case GTF_lockPF : freq /= 1E6; break;
	}
	GTF_calcTimings(xres,yres,freq,freq_type,1,want_interlaced,&gtf);

	crtc->hde = xres;
	crtc->ht = gtf.h.hTotal;
	crtc->hrs = gtf.h.hSyncStart;
	crtc->hre = gtf.h.hSyncEnd;
	crtc->vde = yres / ydup;
	crtc->vt = gtf.v.vTotal / ydup;
	crtc->vrs = gtf.v.vSyncStart / ydup;
	crtc->vre = gtf.v.vSyncEnd / ydup;
	crtc->pixelclock = gtf.dotClock * 1E6;
	crtc->interlaced = want_interlaced;
	crtc->doublescan = doublescan;

	if (xres <= 640 && yres <= 400) {
		if (yres == 350) {
			crtc->hsync = 0;
			crtc->vsync = 1;
		} else if (yres == 200 || yres == 384 || yres == 400) {
			crtc->hsync = 1;
			crtc->vsync = 0;
		} else {
			crtc->hsync = 1;
			crtc->vsync = 1;
		}
	} else {
		crtc->hsync = gtf.hSyncPol == '-';
		crtc->vsync = gtf.vSyncPol == '-';
	}
}

int cmd_mode_create(void) {
	video_mode* vm;
	video_mode copy;
	char buffer[80];
	double freq = 0;
	int freq_type;
	unsigned x;
	unsigned y;

	int i = menu_base + menu_rel;
	if (i >= menu_max)
		return -1;

	vm = video_mode_container_pos(&vm_all,i);
	copy = *vm;
	strcpy(copy.name,"computed_mode");

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VGALINE :
			freq_type = GTF_lockPF;
			break;
		case VIDEO_FLAGS_DRIVER_VBELINE :
		case VIDEO_FLAGS_DRIVER_VBE :
			switch (cmd_input_key(" Set Vertical/Horizontal/Pixel frequency ? ","vhp")) {
				case 0 : freq_type = GTF_lockVF; break;
				case 1 : freq_type = GTF_lockHF; break;
				case 2 : freq_type = GTF_lockPF; break;
				default: return 0;
			}
			break;
		default:
			return -1;
	}

	strcpy(buffer,"");

	switch (freq_type) {
		case GTF_lockVF :
			if (cmd_input_string(" Vertical clock [Hz] (example 60.0) : ",buffer,10)!=0)
				return 0;
			freq = strtod(buffer,0);
			if (freq < 10 || freq > 200) {
				video_error_description_set("Invalid vertical clock value, usually in the range 10 - 200.0 [Hz]");
				return -1;
			}
			break;
		case GTF_lockHF :
			if (cmd_input_string(" Horizontal clock [kHz] (example 31.5) : ",buffer,10)!=0)
				return 0;
			freq = strtod(buffer,0) * 1E3;
			if (freq < 10*1E3 || freq > 100*1E3) {
				video_error_description_set("Invalid horizontal clock value, usually in the range 15.0 - 80.0 [kHz]");
				return -1;
			}
			break;
		case GTF_lockPF :
			if (cmd_input_string(" Pixel clock [MHz] (example 14.0) : ",buffer,10)!=0)
				return 0;
			freq = strtod(buffer,0) * 1E6;
			if (freq < 1*1E6 || freq > 300*1E6) {
				video_error_description_set("Invalid pixel clock value, usually in the range 1.0 - 300.0 [MHz]");
				return -1;
			}
			break;
	}

	strcpy(buffer,"");
	if (cmd_input_string(" X resolution [pixel] : ",buffer,10)!=0)
		return 0;
	x = atoi(buffer);
	if (x < 64 || x > 2048) {
		video_error_description_set("Invalid x resolution value");
		return -1;
	}

	strcpy(buffer,"");
	if (cmd_input_string(" Y resolution [pixel] : ",buffer,10)!=0)
		return 0;
	y = atoi(buffer);
	if (y < 64 || y > 2048) {
		video_error_description_set("Invalid y resolution value");
		return -1;
	}

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGA :
			copy.flags = VIDEO_FLAGS_DRIVER_VGALINE;
			copy.driver.vgaline.mode = vm->driver.vga.mode;
			video_crtc_compute(&copy.driver.vgaline.crtc, x, y, 0, freq, freq_type);
			if (video_mode_import(&copy)!=0)
				return -1;
		break;
		case VIDEO_FLAGS_DRIVER_VBE :
			copy.flags = VIDEO_FLAGS_DRIVER_VBELINE;
			copy.driver.vbeline.mode = vm->driver.vbe.mode;
			video_crtc_compute(&copy.driver.vbeline.crtc, x, y, 0, freq, freq_type);
			if (video_mode_import(&copy)!=0)
				return -1;
		break;
		case VIDEO_FLAGS_DRIVER_VGALINE :
			copy.flags = VIDEO_FLAGS_DRIVER_VGALINE;
			copy.driver.vgaline.mode = vm->driver.vgaline.mode;
			video_crtc_compute(&copy.driver.vgaline.crtc, x, y, 0, freq, freq_type);
			if (video_mode_import(&copy)!=0)
				return -1;
		break;
		case VIDEO_FLAGS_DRIVER_VBELINE :
			copy.flags = VIDEO_FLAGS_DRIVER_VBELINE;
			copy.driver.vbeline.mode = vm->driver.vbeline.mode;
			video_crtc_compute(&copy.driver.vbeline.crtc, x, y, 0, freq, freq_type);
			if (video_mode_import(&copy)!=0)
				return -1;
		break;
		default:
			return -1;
		break;
	}

	copy.flags |= VIDEO_FLAGS_USER_BIT0;

	vm_all_modified = 1;
	video_mode_container_insert_sort(&vm_all,&copy,mode_cmp);

	menu_max = video_mode_container_max(&vm_all);

	return 0;
}

int cmd_export_current(const char* name, vgatweak_video_mode* regs) {
	FILE* f;
	char buffer[3][256];
	int j;

	f = fopen("modes.h","at");
	if (!f)
		return -1;

	fprintf(f,"/*\n");

	format_info_vgatweak(buffer[0],buffer[1],buffer[2], regs);

	fprintf(f," * %s\n",buffer[0]);
	fprintf(f," * %s\n",buffer[1]);
	fprintf(f," * %s\n",buffer[2]);
	fprintf(f," */\n\n");

	fprintf(f,"Register %s[] = {\n",name);
	for(j=0;j<regs->regs_num;++j) {
		fprintf(f,"\t{ 0x%x, 0x%x, 0x%x }", regs->regs[j].port, regs->regs[j].index, regs->regs[j].value);
		if (j+1==regs->regs_num)
			fprintf(f,"\n");
		else
			fprintf(f,",\n");
	}
	fprintf(f,"};\n\n");

	fclose(f);

	return 0;
}

int cmd_export(void) {
	video_mode* vm;
	vgatweak_video_mode regs;

	int i = menu_base + menu_rel;
	if (i >= menu_max)
		return -1;

	vm = video_mode_container_pos(&vm_all,i);

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGA :
			if (video_mode_set(vm)!=0) {
				return -1;
			}
			vgatweak_mode_grab(&regs);
		break;
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			regs = vm->driver.vgatweak;
		break;
		case VIDEO_FLAGS_DRIVER_VGALINE :
			if (vgaline_mode_realize(&regs,&vm->driver.vgaline) != 0) {
				return -1;
			}
		break;
		default:
			sound_error();
			return 0;
	}

	return cmd_export_current(video_mode_name(vm),&regs);
}

int cmd_mode_grab(void) {
	video_mode* vm;
	video_mode copy;
	struct tweak_crtc info;
	video_crtc crtc;
	double vert_freq;

	int i = menu_base + menu_rel;
	if (i >= menu_max)
		return -1;

	vm = video_mode_container_pos(&vm_all,i);

	if (!monitor_check(vm))
		return 0;

	copy = *vm;
	strcpy(copy.name,"grabbed_mode");

	if (video_mode_set(vm)!=0) {
		return -1;
	}

	tweak_crtc_get(&info,tweak_reg_read,0);
	vert_freq = video_measure_vert_freq();

	if (video_crtc_import(&crtc,&info,vm,vert_freq)!=0)
		return -1;

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGA :
			copy.flags = VIDEO_FLAGS_DRIVER_VGALINE;
			copy.driver.vgaline.mode = vm->driver.vga.mode;
			copy.driver.vgaline.crtc = crtc;
			break;
		case VIDEO_FLAGS_DRIVER_VGALINE :
			copy.flags = VIDEO_FLAGS_DRIVER_VGALINE;
			copy.driver.vgaline.mode = vm->driver.vgaline.mode;
			copy.driver.vgaline.crtc = crtc;
			break;
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			copy.flags = VIDEO_FLAGS_DRIVER_VGALINE;
			copy.driver.vgaline.mode = vm->driver.vgatweak.mode;
			copy.driver.vgaline.crtc = crtc;
			break;
		case VIDEO_FLAGS_DRIVER_VBE :
			copy.flags = VIDEO_FLAGS_DRIVER_VBELINE;
			copy.driver.vbeline.mode = vm->driver.vbe.mode;
			copy.driver.vbeline.crtc = crtc;
			break;
		case VIDEO_FLAGS_DRIVER_VBELINE:
			copy.flags = VIDEO_FLAGS_DRIVER_VBELINE;
			copy.driver.vbeline.mode = vm->driver.vbeline.mode;
			copy.driver.vbeline.crtc = crtc;
			break;
	}

	if (video_mode_import(&copy)!=0)
		return -1;

	copy.flags |= VIDEO_FLAGS_USER_BIT0;

	vm_all_modified = 1;
	video_mode_container_insert_sort(&vm_all,&copy,mode_cmp);

	menu_max = video_mode_container_max(&vm_all);

	return 0;
}

int video_mode_compare_by_addr(const video_mode* a, const video_mode* b) {
	if (a==b)
		return 0;
	return 1;
}

void cmd_del(void) {
	video_mode* vm;

	int i = menu_base + menu_rel;
	if (i >= menu_max)
		return;

	vm = video_mode_container_pos(&vm_all,i);

	vm_all_modified = 1;
	video_mode_container_remove(&vm_all,vm,video_mode_compare_by_addr);

	menu_max = video_mode_container_max(&vm_all);
}

void cmd_save(void) {
	video_mode_container vm_sel;
	video_mode_container_iterator i;
	video_mode_container_init(&vm_sel);
	for(video_mode_container_iterator_begin(&i,&vm_all);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
		const video_mode* vm = video_mode_container_iterator_get(&i);
		if (vm->flags &	VIDEO_FLAGS_USER_BIT0)
			video_mode_container_insert(&vm_sel,vm);
	}
	video_mode_container_save(&vm_sel);
	video_mode_container_done(&vm_sel);

	vm_all_modified = 0;
}

int cmd_exit(void) {
	if (vm_all_modified) {
		int res;

		sound_warn();

		res = cmd_input_key(" Save before exiting : ","yn");
		if (res < 0)
			return 0;

		if (res == 0)
			cmd_save();
	}

	return 1;
}

/* Window coordinate */
#define MENU_X 0
#define MENU_Y 2
#define MENU_DX (screen_size_x())
#define MENU_DY (screen_size_y()-6)

#define INFO_X 0
#define INFO_Y (screen_size_y()-4)
#define INFO_DX	(screen_size_x())
#define INFO_DY 3

#define BAR_X 0
#define BAR_Y1 0
#define BAR_Y2 (screen_size_y()-1)
#define BAR_DX (screen_size_x())

/* Menu */
int cmd_run(void) {
	int done;
	int userkey;

	menu_base = 0;
	menu_rel = 0;
	menu_rel_max = MENU_DY;
	menu_max = video_mode_container_max(&vm_all);
	menu_base_max =	menu_max -  menu_rel_max;
	if (menu_base_max < 0)
		menu_base_max = 0;

	done = 0;
	while (!done) {
		draw_text_bar(BAR_X,BAR_Y1,BAR_Y2,BAR_DX);
		draw_text_info(INFO_X,INFO_Y,INFO_DX,INFO_DY,menu_base + menu_rel);
		draw_text_menu(MENU_X,MENU_Y,MENU_DX,MENU_DY);

		userkey = screen_getkey();

		switch (userkey) {
			case SCREEN_KEY_UP:
				cmd_gotopos( menu_base + menu_rel - 1);
				break;
			case SCREEN_KEY_DOWN:
				cmd_gotopos( menu_base + menu_rel + 1 );
				break;
			case SCREEN_KEY_HOME: {
				int i =	menu_base + menu_rel - 1;
				if (i<0)
					i = 0;
				while (i>0 && !(video_mode_container_pos(&vm_all,i)->flags & VIDEO_FLAGS_USER_BIT0))
					--i;
				cmd_gotopos( i );
				break;
			}
			case SCREEN_KEY_END: {
				int i =	menu_base + menu_rel + 1;
				if (i >= menu_max)
					i = menu_max - 1;
				while (i < menu_max - 1 && !(video_mode_container_pos(&vm_all,i)->flags & VIDEO_FLAGS_USER_BIT0))
					++i;
				cmd_gotopos( i );
				break;
			}
			case SCREEN_KEY_PGDN:
				cmd_gotopos( menu_base + menu_rel + menu_rel_max );
				break;
			case SCREEN_KEY_PGUP:
				cmd_gotopos( menu_base + menu_rel - menu_rel_max );
				break;
			case SCREEN_KEY_F2:
				cmd_save();
				break;
			case SCREEN_KEY_ESC:
				done = cmd_exit();
				break;
			case SCREEN_KEY_SPACE:
				cmd_select( menu_base + menu_rel );
				cmd_gotopos( menu_base + menu_rel + 1 );
				break;
			case SCREEN_KEY_ENTER:
				if (cmd_onvideo_test( menu_base + menu_rel ) != 0) {
					screen_reset();
					draw_text_error();
				} else {
					screen_reset();
				}
				break;
			case SCREEN_KEY_F9:
				if (cmd_onvideo_calib( menu_base + menu_rel ) != 0) {
					screen_reset();
					draw_text_error();
				} else {
					screen_reset();
				}
				break;
			case SCREEN_KEY_F10:
				if (cmd_onvideo_animate( menu_base + menu_rel ) != 0) {
					screen_reset();
					draw_text_error();
				} else {
					screen_reset();
				}
				break;
			case SCREEN_KEY_TAB :
				cmd_rename();
				break;
			case SCREEN_KEY_F5 :
				if (cmd_mode_create() !=0) {
					screen_reset();
					draw_text_error();
				} else {
					screen_reset();
				}
				break;
			case SCREEN_KEY_F6 :
				if (cmd_mode_grab() !=0 ) {
					screen_reset();
					draw_text_error();
				} else {
					screen_reset();
				}
				break;
			case SCREEN_KEY_F7 :
				cmd_mode_change();
				break;
			case SCREEN_KEY_DEL :
				cmd_del();
				cmd_gotopos( menu_base + menu_rel );
				break;
			case SCREEN_KEY_F1:
				draw_text_help();
				break;
			case SCREEN_KEY_F8:
				if (cmd_export() != 0) {
					screen_reset();
					draw_text_error();
				} else {
					screen_reset();
				}
				break;
			default:
				if (cmd_offvideo_test(menu_base + menu_rel, userkey) != 0) {
					draw_text_error();
				}
				break;
		}
	}
	return userkey;
}

/***************************************************************************/
/* log */

FILE* video_log_file;

void video_log(const char *text,...)
{
	va_list arg;
	va_start(arg,text);
	if (video_log_file)
		vfprintf(video_log_file,text,arg);
	va_end(arg);
}

int main(int argc, char* argv[]) {
	video_mode_container vm_sel;
	video_mode_container_iterator i;

	int j;
	char* cfg = "mame.cfg";
	char* log = 0;

	for (j = 1;j < argc;j++)
	{
	        if (stricmp(argv[j],"-c") == 0 && j+1<argc) {
			cfg = argv[++j];
		} else if (stricmp(argv[j],"-log") == 0) {
			log = "mv.log";
		} else if (stricmp(argv[j],"-nosound") == 0) {
			sound_flag = 0;
		}
	}

	if (log)
		video_log_file = fopen(log,"wt");
	else
		video_log_file = 0;

	allegro_init();
	set_config_file(cfg);

	if (video_options_load() != 0) {
		allegro_exit();
		printf("Error loading the video options\n%s", video_error_description_get());
		exit(1);
	}

	video_init();

	if (video_monitor_load(&monitor) != 0) {
		video_done();
		allegro_exit();
		printf("Error loading the monitor options\n%s", video_error_description_get());
		exit(1);
	}

	/* all mode */
	video_mode_container_init(&vm_sel);
	video_mode_container_insert_vga(&vm_sel);
	video_mode_container_insert_vbe(&vm_sel);
	video_mode_container_insert_vgaline(&vm_sel);
	video_mode_container_insert_vbeline(&vm_sel);

	/* sort */
	video_mode_container_init(&vm_all);
	for(video_mode_container_iterator_begin(&i,&vm_sel);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
		video_mode* vm = video_mode_container_iterator_get(&i);
		video_mode_container_insert_sort(&vm_all,vm,mode_cmp);
	}
	video_mode_container_done(&vm_sel);

	/* load selected */
	video_mode_container_init(&vm_sel);
	video_mode_container_load(&vm_sel);

	/* union set */
	for(video_mode_container_iterator_begin(&i,&vm_sel);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
		video_mode* vm = video_mode_container_iterator_get(&i);
		int has = video_mode_container_has(&vm_all,vm,video_mode_compare) != 0;
		int insert = 1;
		if (!has) {
			/* if a vesa mode is not autodetected don't insert */
 			switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
				case VIDEO_FLAGS_DRIVER_VBE :
					insert = 0;
				break;
			}
		}
		if (insert) {
			video_mode_container_remove(&vm_all,vm,video_mode_compare);
			vm->flags |= VIDEO_FLAGS_USER_BIT0;
			video_mode_container_insert_sort(&vm_all,vm,mode_cmp);
		}
	}
	video_mode_container_done(&vm_sel);
	vm_all_modified = 0;

	screen_init();

	sound_signal();

	cmd_run();

	screen_done();

	video_mode_container_done(&vm_all);

        video_done();
	allegro_exit();

	if (video_log_file) {
		printf("Log saved in %s\n",log);
		fclose(video_log_file);
	}

	return 0;
}


