/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 *
 * This video driver is STRONGLY derived from the video drivers of the
 * VSyncMAME project by S.Sakamaki (Saka)
 */

#include <pc.h>
#include <conio.h>
#include <sys/farptr.h>
#include <go32.h>
#include <time.h>
#include <math.h>
#include <dpmi.h>
#include "g400.h"
#include "mode15k.h"
#include <stdio.h>

#define USE_ZOOM
//#define ZOOMLIMITATION
//#define ADJUST_VRATE

static int g_endofline = 0;
static int g_interlaceflag = 0;
static int g_valid = 0;

static void SetG400ExtRegister(CRTCPARAM *cp);
static void SetG400Interlace(int flag, int HSyncStart, int HSyncEnd, int HTotal, int ratio);
static long GetG400VClockParam(unsigned long vclk, int *N, int *M, int *P, int *S);
static void SetG400VClock(int N, int M, int P, int S);
static void SetG400HZoom(int d0);
static int GetHorizontalZoomRatio(CRTCPARAM *cp);
static void AdjustCrtcParam(CRTCPARAM *cp, int hzoom);
static void AdjustSyncParam(CRTCPARAM *cp, long dotclock);
static int CheckHorizontalTotal(int HTotal, int hzoom);
static void SetG400DoubleScan(int flag);

static DWORD GetG400BytePerLine(void);
static void SetG400BytePerLine(DWORD d0);
static BYTE G400IdxReg_inb(BYTE _index);
static BYTE G400CrtcExtReg_inb(BYTE _index);
static void G400IdxReg_outb(BYTE _index, BYTE _data);
static void G400CrtcExtReg_outb(BYTE _index, BYTE _data);

static unsigned long G400PhysicalAddress = 0;
static unsigned long G400LinearAddress = 0;
static int G400LinearSelector = 0;

int detectG400(void)
{
	__dpmi_regs 	r;

	r.d.eax = 0x0000b101;						// PCI BIOS - INSTALLATION CHECK
	r.d.edi = 0x00000000;
	__dpmi_int(0x1a, &r);
	if( r.d.edx != 0x20494350 ){				// ' ICP'
		logerror("PCI BIOS not Installed.\n");
		return 0;
	}

	r.d.eax = 0x0000b102;						// PCI BIOS - FIND PCI DEVICE
	r.d.ecx = 0x00000525;						// device ID
	r.d.edx = 0x0000102B;						// vendor ID
	r.d.esi = 0x00000000;						// device index
	__dpmi_int(0x1a, &r);
	if( r.h.ah != 0 ){
		logerror( "g400: not found\n");
		return 0;
	}

	r.d.eax = 0x0000b10a;						// READ CONFIGURATION DWORD
												// BH = bus number
												// BL = device/function number
	r.d.edi = 0x00000014;						// register number
	__dpmi_int(0x1a, &r);
	if( r.h.ah != 0 ){
		logerror( "g400: Read Configuration DWORD error.\n");
		return 0;
	}

	logerror("G400 found\n");

	G400PhysicalAddress = r.d.ecx & 0xFFFFC000;
	logerror( "g400: MMIO Address %08lx\n", G400PhysicalAddress);

	if (_create_linear_mapping(&G400LinearAddress, G400PhysicalAddress, 0x10000) != 0){
		logerror( "g400: _create_linear_mapping error\n");
		return 0;
	}
	logerror( "g400: Linear Address %08lx\n", G400LinearAddress);

	if (_create_selector(&G400LinearSelector, G400LinearAddress, 0x10000) != 0) {
		_remove_linear_mapping(&G400LinearAddress);
		logerror( "g400: _create_selector error\n");
		return 0;
	}

#ifdef MAME
	DisableStretch();
#endif

	return 1;
}

void resetG400(void)
{
	_remove_linear_mapping(&G400LinearAddress);
	_remove_selector(&G400LinearSelector);
}

int setG400(void)
{
	CRTCPARAM cp;
	int hzoom;
	int N, M, P, S;
	long vclk;
#ifdef ADJUST_VRATE
	int tempdt;
	int dtHTotal;
	long temp;
	long dt;
#endif

	if(GetMode15KHz(&cp) == 0) return 0;

#if 1
	if (cp.doublescan) {
		/* Richiesto dal test di Corrado sulla G400 */
		cp.dotclockHz /= 2;
	}
#endif

	cp.HTotal += 8;
	hzoom = GetHorizontalZoomRatio(&cp);
	AdjustCrtcParam(&cp, hzoom);

#ifdef ADJUST_VRATE
	logerror( "desrate %d\n", (int)cp.vert001Hz);
	dtHTotal = 0;
	dt		 = 10000;	   // 100.00Hz
	tempdt	 = 0;
	for( tempdt = 0 ; tempdt < (hzoom * 8) ; tempdt += 8 ){
		if( CheckHorizontalTotal(cp.HTotal + tempdt, hzoom) == 0 )
			continue;
		vclk = CalcDotclock(cp.vert001Hz, cp.HTotal + tempdt,
								cp.VTotal, cp.interlace);
		vclk = GetG400VClockParam(vclk, &N, &M, &P, &S);
		temp = CalcVert001Hz(vclk, cp.HTotal + tempdt, cp.VTotal, cp.interlace);
		logerror( "refreshrate %d\n", (int)temp);
		temp = temp - cp.vert001Hz;
		logerror( "dt %d\n\n", (int)temp);
		if(temp < 0) temp *= -1;
		if(temp < dt){
			dt = temp;
			dtHTotal = tempdt;
		}
	}
	logerror( "dt %dHz\n", (int)dt);
	logerror( "dtHTotal %d\n", dtHTotal);
	cp.HTotal += dtHTotal;
	cp.HBEnd   = cp.HTotal - 8;
	vclk = CalcDotclock(cp.vert001Hz, cp.HTotal, cp.VTotal, cp.interlace);
	vclk = GetG400VClockParam(vclk, &N, &M, &P, &S);

#else

	vclk = GetG400VClockParam(cp.dotclockHz, &N, &M, &P, &S);

#endif

	AdjustSyncParam(&cp, vclk);
	DispCrtcParam(&cp);

	DisableGenerateSignal();
	SetGenericVGARegister(&cp);
	SetG400ExtRegister(&cp);
	SetG400VClock(N, M, P, S);
	SetG400HZoom(hzoom);
	SetG400DoubleScan(cp.doublescan);
	SetHVPolarity(cp.hpolarity, cp.vpolarity);
	EnableGenerateSignal();

	return 1;
}

static int GetHorizontalZoomRatio(CRTCPARAM *cp)
{
	int zoom = 1;

#ifdef USE_ZOOM

	if(cp->HTotal < 512){
		zoom = 4;
	}else if(cp->HTotal < 1024){
		zoom = 2;
	}else{
		zoom = 1;
	}

#endif

	return zoom;
}


static void AdjustSyncParam(CRTCPARAM *cp, long dotclock)
{
	int d0;

	if(dotclock == 0){
		cp->dotclockHz = CalcDotclock(cp->vert001Hz, cp->HTotal,
											cp->VTotal, cp->interlace);
	}else{
		cp->dotclockHz = dotclock;
	}
	cp->horzHz	   = cp->dotclockHz / cp->HTotal;
	d0 = cp->VTotal * 2;
	if(cp->interlace != 0){
		d0 += 1;
	}
	cp->vert001Hz  = (cp->horzHz * 100 * 2) / d0;
}


static void AdjustCrtcParam(CRTCPARAM *cp, int hzoom)
{
	cp->HTotal	*= hzoom;
	cp->HDisp	*= hzoom;
	cp->HBStart *= hzoom;
	cp->HBEnd	*= hzoom;
	cp->HSStart *= hzoom;
	cp->HSEnd	*= hzoom;

	cp->HBStart = cp->HDisp - 8;
	cp->HBEnd	= cp->HTotal - 8;
	cp->VBStart = cp->VDisp  - 1;
	cp->VBEnd	= cp->VTotal - 1;


	if( (cp->HBEnd - cp->HBStart) > (0x7F << 3) ){
		cp->HBEnd = cp->HBStart + (0x7F << 3);
	}
	if( (cp->HSEnd - cp->HSStart) > (0x1F << 3) ){
		cp->HSEnd = cp->HSStart + (0x1F << 3);
	}

	if(cp->interlace != 0){
		if( (cp->VTotal % 2) != 0 ){
			cp->VTotal += 1;
			cp->VBEnd	= cp->VTotal - 1;
		}
	}

	if( CheckHorizontalTotal(cp->HTotal, hzoom) == 0 ){
		cp->HTotal += 8;
	}

	AdjustSyncParam(cp, 0);
}



static int CheckHorizontalTotal(int HTotal, int hzoom)
{
	int result = 1;
#ifdef ZOOMLIMITATION
	int d0;

	d0 = 8 * hzoom;
	if( ((HTotal >> 3) % d0)  ==  (d0 - 1) ){
		result = 0;
	}
#endif
	return result;
}

static void SetG400Interlace(int flag, int HSyncStart, int HSyncEnd, int HTotal, int ratio)
{
	int d1;

	if(flag != 0){
		d1 = G400CrtcExtReg_inb(0x00);
		if((d1 & 0x80) == 0){
			d1 = (HSyncStart/8) + (HSyncEnd/8) - (HTotal/8);
			d1 = (d1 * ratio) / 100;
			d1 = (d1 / 2) - 1;
			if(d1 < 0)	  d1 = 0;
			if(d1 > 0xFF) d1 = 0xFF;
			G400CrtcExtReg_outb(0x05, d1 & 0xFF);
			G400CrtcExtReg_outb(0x00, BitMov(G400CrtcExtReg_inb(0x00), 7, 1, 0));
			SetG400BytePerLine(GetG400BytePerLine() * 2);
		}
	}else{
		d1 = G400CrtcExtReg_inb(0x00);
		if((d1 & 0x80) != 0){
			G400CrtcExtReg_outb(0x00, BitMov(G400CrtcExtReg_inb(0x00), 7, 0, 0));
			G400CrtcExtReg_outb(0x05, 0);
			SetG400BytePerLine(GetG400BytePerLine() / 2);
		}
	}
}

static DWORD GetG400BytePerLine(void)
{
	DWORD d0;

	d0 = GetCrtcReg(0x13);
	d0 = BitMov(d0, 8, G400CrtcExtReg_inb(0x00), 4);
	d0 = BitMov(d0, 9, G400CrtcExtReg_inb(0x00), 5);
	return d0;
}

static void SetG400BytePerLine(DWORD d0)
{
	SetCrtcReg(0x13, d0 & 0xFF);
	G400CrtcExtReg_outb(0x00, BitMov(G400CrtcExtReg_inb(0x00), 4, d0, 8));
	G400CrtcExtReg_outb(0x00, BitMov(G400CrtcExtReg_inb(0x00), 5, d0, 9));
}

void SetG400ExtRegister(CRTCPARAM *cp)
{
	int d0;

	d0 = 0;
	d0 = BitMov(d0, 0, (cp->HTotal	>> 3) - 5, 8);
	d0 = BitMov(d0, 1, (cp->HBStart >> 3)	 , 8);
	d0 = BitMov(d0, 2, (cp->HSStart >> 3)	 , 8);
	d0 = BitMov(d0, 6, (cp->HBEnd	>> 3)	 , 6);
	G400CrtcExtReg_outb(0x01, d0);

	d0 = 0;
	d0 = BitMov(d0, 0, cp->VTotal - 2, 10);
	d0 = BitMov(d0, 1, cp->VTotal - 2, 11);
	d0 = BitMov(d0, 2, cp->VDisp  - 1, 10);
	d0 = BitMov(d0, 3, cp->VBStart	 , 10);
	d0 = BitMov(d0, 4, cp->VBStart	 , 11);
	d0 = BitMov(d0, 5, cp->VSStart	 , 10);
	d0 = BitMov(d0, 6, cp->VSStart	 , 11);
	G400CrtcExtReg_outb(0x02, d0);

	SetG400Interlace(cp->interlace, cp->HSStart, cp->HSEnd, cp->HTotal, cp->interlaceratio);

	g_interlaceflag = cp->interlace;
	g_endofline = cp->VTotal;
	g_valid = 1;
}

#define BASE_FREQ 27000.0

static long GetG400VClockParam(unsigned long vclk, int *N, int *M, int *P, int *S)
{
	int d0;
	int d1;
	double fd;
	double dt;
	double tempfd;
	double tempdt;
	double ref = BASE_FREQ;
	double Fvco;

	vclk = vclk / 1000;
	*P = 0;
	if(vclk > 240000) vclk = 240000;
	if( 	 vclk <   30000){
		*P = 8;
	}else if(vclk <   60000){
		*P = 4;
	}else if(vclk <  120000){
		*P = 2;
	}else if(vclk <= 240000){
		*P = 1;
	}

	*N = 0;
	*M = 0;
	tempfd = 0;
	tempdt = 0;
	fd = vclk * (*P);
	dt = 240000;

	d0 = 8;
	while(d0 <= 128){
		d1 = 2;
		while(d1 <= 32){
			tempfd = ref * d0 / d1;
			tempdt = fd - tempfd;
			if(tempdt < 0) tempdt *= -1;
			if(dt > tempdt){
				*N = d0;
				*M = d1;
				dt = tempdt;
			}
			d1 += 1;
		}
		d0 += 1;
	}

	Fvco = ref * (*N) / (*M);
	tempfd = Fvco / (*P);

	*S = 0;
	if( 	 Fvco < 110000){
		*S = 0;
	}else if(Fvco < 170000){
		*S = 1;
	}else if(Fvco < 240000){
		*S = 2;
	}else if(Fvco < 310000){
		*S = 3;
	}

#if 1
	logerror( "N: %d\n", *N);
	logerror( "M: %d\n", *M);
	logerror( "P: %d\n", *P);
	logerror( "S: %d\n", *S);

	logerror( "Fd	%f\n", fd / (*P));
	logerror( "Fout %f\n", tempfd);
	logerror( "Fvco %f\n", Fvco);
#endif

	return (long)(tempfd * 1000);
}

static void SetG400VClock(int N, int M, int P, int S)
{
	int PLLn;

	G400IdxReg_outb(0x1A, BitMov(G400IdxReg_inb(0x1A), 2, 1, 0));

	PLLn = _in(0x3CC);
	PLLn = PLLn | 0x0C;
	_out(0x3C2, PLLn);

	M -= 1;
	N -= 1;
	P -= 1;

	M = M & 0x1F;
	G400IdxReg_outb(0x4C, M);
	N = N & 0x7F;
	G400IdxReg_outb(0x4D, N);
	P = P | (S << 3);
	P = P & 0x1F;
	G400IdxReg_outb(0x4E, P);

	while(1){
		if( (G400IdxReg_inb(0x4F) & 0x40) != 0 ) break;
	}

	G400IdxReg_outb(0x1A, BitMov(G400IdxReg_inb(0x1A), 2, 0, 0));

}

static void SetG400HZoom(int d0)
{
	if(d0 == 1){
		d0 = 0;
	}else if(d0 == 2){
		d0 = 1;
	}else if(d0 == 4){
		d0 = 3;
	}else{
		d0 = 0;
	}
	G400IdxReg_outb(0x38, d0);
}

static void SetG400DoubleScan(int flag)
{
	int d0;

	d0 = GetCrtcReg(0x09) & ~0x1F;
	if(flag != 0){
		d0 = d0 | 0x01;
	}
	SetCrtcReg(0x09, d0);
}

static BYTE G400IdxReg_inb(BYTE _index)
{
	_farpokeb(G400LinearSelector, 0x3C00, _index);
	return _farpeekb(G400LinearSelector, 0x3C0A);
}

static void G400IdxReg_outb(BYTE _index, BYTE _data)
{
	_farpokeb(G400LinearSelector, 0x3C00, _index);
	_farpokeb(G400LinearSelector, 0x3C0A, _data);
}

static BYTE G400CrtcExtReg_inb(BYTE _index)
{
	_farpokeb(G400LinearSelector, 0x1FDE, _index);
	return _farpeekb(G400LinearSelector, 0x1FDF);
}

static void G400CrtcExtReg_outb(BYTE _index, BYTE _data)
{
	_farpokeb(G400LinearSelector, 0x1FDE, _index);
	_farpokeb(G400LinearSelector, 0x1FDF, _data);
}


