/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 *
 * This video driver is STRONGLY derived from the video drivers of the
 * VSyncMAME project by S.Sakamaki (Saka)
 */

#include <pc.h>
#include <conio.h>
#include <sys/farptr.h>
#include <go32.h>
#include <time.h>
#include <math.h>
#include <dpmi.h>
#include "s6326.h"
#include "mode15k.h"

#define SiS6326    1
#define SiS620	   2
#define SiS530	   3
#define SiS300	   4
#define SiS630	   5
#define SiS540	   6

static void SetSiS6326ExtRegister(CRTCPARAM *cp);
static int GetSiS6326BytePerLine(void);
static void SetSiS6326BytePerLine(int d0);
static void SetSiS6326Interlace(int flag);
static int SetSiS6326VClock(BYTE Numerator, BYTE Divider, BYTE DeNumerator, BYTE PostScalar, BYTE SeqDivider);
static void SetSiS300ExtRegister(CRTCPARAM *cp);
static int GetSiS300BytePerLine(void);
static void SetSiS300BytePerLine(int d0);
static void SetSiS300Interlace(int flag);
static int SetSiS300VClock(BYTE Numerator, BYTE Divider, BYTE DeNumerator, BYTE PostScalar, BYTE SeqDivider);
static void SetSis6326DoubleScan(int flag);
static void SetSis6326ClockRegisterInternal(void);
static int ExtSis6326RegisterUnLock(void);
static long GetSiS6326VclkParam(long vclk, int *Num, int *DeN, int *Div, int *PS, int *SeqDiv);

static DWORD DeviceType = 0;

int detectSiS6326(void)
{
//
//	DPMI & PCI BIOS calling from ifsega2p.c. Thanks to SUZ.
//
	__dpmi_regs 	r;

	r.d.eax = 0x0000b101;						// PCI BIOS - INSTALLATION CHECK
	r.d.edi = 0x00000000;
	__dpmi_int(0x1a, &r);
	if( r.d.edx != 0x20494350 ){				// ' ICP'
		logerror("PCI BIOS not Installed.\n");
		return 0;
	}

	while(1){
		r.d.eax = 0x0000b102;						// PCI BIOS - FIND PCI DEVICE
		r.d.ecx = 0x00000620;						// device ID
		r.d.edx = 0x00001039;						// vendor ID
		r.d.esi = 0x00000000;						// device index
		__dpmi_int(0x1a, &r);
		if( r.h.ah == 0 ){
			logerror("SiS620 found\n");
			DeviceType = SiS620;
			break;
		}
		r.d.eax = 0x0000b102;						// PCI BIOS - FIND PCI DEVICE
		r.d.ecx = 0x00000530;						// device ID
		r.d.edx = 0x00001039;						// vendor ID
		r.d.esi = 0x00000000;						// device index
		__dpmi_int(0x1a, &r);
		if( r.h.ah == 0 ){
			logerror("SiS530 found\n");
			DeviceType = SiS530;
			break;
		}
		r.d.eax = 0x0000b102;						// PCI BIOS - FIND PCI DEVICE
		r.d.ecx = 0x00006326;						// device ID
		r.d.edx = 0x00001039;						// vendor ID
		r.d.esi = 0x00000000;						// device index
		__dpmi_int(0x1a, &r);
		if( r.h.ah == 0 ){
			logerror("SiS6326 found\n");
			DeviceType = SiS6326;
			break;
		}
		r.d.eax = 0x0000b102;						// PCI BIOS - FIND PCI DEVICE
		r.d.ecx = 0x00000300;						// device ID
		r.d.edx = 0x00001039;						// vendor ID
		r.d.esi = 0x00000000;						// device index
		__dpmi_int(0x1a, &r);
		if( r.h.ah == 0 ){
			logerror("SiS300 found\n");
			DeviceType = SiS300;
			break;
		}
		r.d.eax = 0x0000b102;						// PCI BIOS - FIND PCI DEVICE
		r.d.ecx = 0x00000630;						// device ID
		r.d.edx = 0x00001039;						// vendor ID
		r.d.esi = 0x00000000;						// device index
		__dpmi_int(0x1a, &r);
		if( r.h.ah == 0 ){
			logerror("SiS630 found\n");
			DeviceType = SiS630;
			break;
		}
		r.d.eax = 0x0000b102;						// PCI BIOS - FIND PCI DEVICE
		r.d.ecx = 0x00000540;						// device ID
		r.d.edx = 0x00001039;						// vendor ID
		r.d.esi = 0x00000000;						// device index
		__dpmi_int(0x1a, &r);
		if( r.h.ah == 0 ){
			logerror("SiS540 found\n");
			DeviceType = SiS540;
			break;
		}

		logerror("sis: not found\n");
		return 0;
	}


	if(ExtSis6326RegisterUnLock() == 0){
		logerror("15.75KHz: Failed unlock extention registers\n");
		return 0;
	}

#ifdef MAME
	DisableStretch();
#endif

	return 1;
}

static int ExtSis6326RegisterUnLock(void)
{
	BYTE d0;

	d0 = GetSeqReg(0x05);
	if(d0 == 0xA1)
		return 1;
	SetSeqReg(0x05, 0x86);
	if(GetSeqReg(0x05) == 0xA1){
		logerror("15.75KHz: SiS extention registers unlocked\n");
		return 1;
	}
	SetSeqReg(0x05, d0);
	return 0;
}


void resetSiS6326(void)
{
}

int setSiS6326(void)
{
	CRTCPARAM cp;
	int Num;
	int DeN;
	int Div;
	int PS;
	int SeqDiv;

	if(GetMode15KHz(&cp) == 0) return 0;
	cp.dotclockHz = GetSiS6326VclkParam(cp.dotclockHz, &Num, &Div, &DeN, &PS, &SeqDiv);
	cp.vert001Hz = CalcVert001Hz(cp.dotclockHz, cp.HTotal, cp.VTotal, cp.interlace);
	cp.horzHz = CalcHorizontalHz(cp.dotclockHz, cp.HTotal);
	cp.HBStart = cp.HDisp;
	cp.HBEnd   = cp.HTotal - 8;
	cp.VBStart = cp.VDisp;
	cp.VBEnd   = cp.VTotal - 1;

	DispCrtcParam(&cp);

	DisableGenerateSignal();
	switch(DeviceType){
		case SiS6326:
		case SiS620:
		case SiS530:
			SetSiS6326VClock(Num, Div, DeN, PS, SeqDiv);
			SetGenericVGARegister(&cp);
			SetSiS6326ExtRegister(&cp);
			break;

		case SiS300:
		case SiS630:
		case SiS540:
			SetSiS300VClock(Num, Div, DeN, PS, SeqDiv);
			SetGenericVGARegister(&cp);
			SetSiS300ExtRegister(&cp);
			break;
	}

	EnableGenerateSignal();
	return 1;
}

static void SetSis6326ClockRegisterInternal(void)
{
	BYTE d0;

	d0 = _in(0x3CC);
	d0 = d0 | 0x0C;
	_out(0x3C2, d0);
}


static BYTE PostScalarTable[]={ 0, 0x00, 0x01, 0x02, 0x03, 0, 0x02, 0, 0x03};

static int SetSiS6326VClock(BYTE Numerator, BYTE Divider, BYTE DeNumerator, BYTE PostScalar, BYTE SeqDivider)
{
	BYTE _2B_DeN;
	BYTE _2B_PS;
	BYTE _2A_Div;
	BYTE _2A_Num;
	BYTE _13_PS;
	BYTE d0;
	int d1;

	SetSis6326ClockRegisterInternal();
	SetSeqReg(0x38, GetSeqReg(0x38) & 0xFC);

	d0 = 0x01;
	if(SeqDivider == 2)  d0 = d0 | 0x08;
	SetSeqReg(0x01, d0);

	_2A_Div = ((Divider == 1) ? 0 : 1) << 7;
	_2A_Num = (Numerator - 1) & 0x7F;
	_13_PS	= ((PostScalar == 6 || PostScalar == 8) ? 1 : 0) << 6;
	_2B_PS	= PostScalarTable[PostScalar] << 5;
	_2B_DeN = (DeNumerator - 1) & 0x1F;

	d0 = _2A_Div | _2A_Num;
	SetSeqReg(0x2A, d0);
	d0 = GetSeqReg(0x13);
	d0 = d0 & ~0x40;
	d0 = d0 | _13_PS;
	SetSeqReg(0x13, d0);
	d0 = _2B_PS | _2B_DeN;
//	if( frequency >= high ) d0 |= 0x80;
	SetSeqReg(0x2B, d0);			   // trigger ?

	d1 = 65536;
	while(d1 > 0){
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		d1 -= 1;
	}

	return 0;
}

static int SetSiS300VClock(BYTE Numerator, BYTE Divider, BYTE DeNumerator, BYTE PostScalar, BYTE SeqDivider)
{
	BYTE d0;
	int d1;
	BYTE xr2b;
	BYTE xr2c;
	int sbit;

	SetSis6326ClockRegisterInternal();

	d0 = 0x01;
	if(SeqDivider == 2)  d0 = d0 | 0x08;
	SetSeqReg(0x01, d0);

	sbit = (PostScalar == 6 || PostScalar == 8) ? 1 : 0;
	PostScalar = PostScalarTable[PostScalar];

	xr2b = (Numerator - 1) & 0x7f;
	if (Divider == 2)
		xr2b |= 0x80;
	xr2c = (DeNumerator - 1) & 0x1f;
	xr2c |= PostScalar << 5;
	if (sbit)
		xr2c |= 0x80;
	SetSeqReg(0x2B, xr2b);
	SetSeqReg(0x2C, xr2c);
	SetSeqReg(0x2D, 0x80);

	d1 = 65536;
	while(d1 > 0){
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		d1 -= 1;
	}

	return 0;
}

static void SetSiS6326ExtRegister(CRTCPARAM *cp)
{
	int d0;

	d0 = GetSeqReg(0x0A);
	d0 = BitMov(d0, 0, cp->VTotal  - 2,  10);
	d0 = BitMov(d0, 1, cp->VDisp   - 1 , 10);
	d0 = BitMov(d0, 2, cp->VBStart	   , 10);
	d0 = BitMov(d0, 3, cp->VSStart	   , 10);
	SetSeqReg(0x0A, d0);


	d0 = BitMov(d0, 0, (cp->HTotal	>> 3) - 5,	8);
	d0 = BitMov(d0, 1, (cp->HDisp	>> 3) - 1,	8);
	d0 = BitMov(d0, 2, (cp->HBStart >> 3)	 ,	8);
	d0 = BitMov(d0, 3, (cp->HSStart >> 3)	 ,	8);
	d0 = BitMov(d0, 4, (cp->HBEnd	>> 3)	 ,	6);
#if 0
	d0 = BitMov(d0, 5, cp->HSStart, 0);
	d0 = BitMov(d0, 6, cp->HSStart, 1);
	d0 = BitMov(d0, 7, cp->HSStart, 2);
#else
	d0 = d0 & 0x1F;
#endif
	SetSeqReg(0x12, d0);


	SetSis6326DoubleScan(cp->doublescan);
	SetSiS6326Interlace(cp->interlace);
	SetHVPolarity(cp->hpolarity, cp->vpolarity);
}

static void SetSiS300ExtRegister(CRTCPARAM *cp)
{
	int d0;

	d0 = (
		(((cp->VSEnd)	 & 0x10 ) <<  1) |		 /* D5 */
		(((cp->VBEnd)	 & 0x100) >>  4) |		 /* D4 */
		(((cp->VSStart)  & 0x400) >>  7) |		 /* D3 */
		(((cp->VBStart)  & 0x400) >>  8) |		 /* D2 */
		(((cp->VDisp-1)  & 0x400) >>  9) |		 /* D1 */
		(((cp->VTotal-2) & 0x400) >> 10));		 /* D0 */
	SetSeqReg(0x0A, d0);

	d0 = (
		(( (cp->HSStart >> 3)	   & 0x300) >> 2)  |  /* D76 */
		(( (cp->HBStart >> 3)	   & 0x300) >> 4)  |  /* D54 */
		((((cp->HDisp	>> 3) - 1) & 0x300) >> 6)  |  /* D32 */
		((((cp->HTotal	>> 3) - 5) & 0x300) >> 8));   /* D10 */
	SetSeqReg(0x0B, d0);

	d0	= GetSeqReg(0x0C) & 0xF8;
	d0 |= (((cp->HSEnd >> 3) & 0x20) >> 3) |	  /* D2 */
		  (((cp->HBEnd >> 3) & 0xC0) >> 6); 	  /* D10 */
	SetSeqReg(0x0C, d0);

	SetSis6326DoubleScan(cp->doublescan);
	SetSiS300Interlace(cp->interlace);
	SetHVPolarity(cp->hpolarity, cp->vpolarity);
}

static int GetSiS6326BytePerLine(void)
{
	DWORD d0;

	d0	= (DWORD)GetCrtcReg(0x13);
	d0 |= (DWORD)((GetSeqReg(0x0A) << 4) & 0x0F00);
	d0 = d0 << 3;
	return d0;
}

static void SetSiS6326BytePerLine(int d0)
{
	BYTE d1;

	d0 = d0 >> 3;
	d1 = (BYTE)(d0 & 0xFF);
	SetCrtcReg(0x13, d1);
	d1 = GetSeqReg(0x0A) & 0x0F;
	d1 = d1 | (BYTE)((d0 >> 4) & 0xF0);
	SetSeqReg(0x0A, d1);
}

static void SetSis6326DoubleScan(int flag)
{
	if(flag == 0){
		SetCrtcReg(0x09, BitMov(GetCrtcReg(0x09), 7, 0, 0));
	}else{
		SetCrtcReg(0x09, BitMov(GetCrtcReg(0x09), 7, 1, 0));
	}
}

static void SetSiS6326Interlace(int flag)
{
	BYTE d1;

	if(flag == 0){
		d1 = GetSeqReg(0x06);
		if( (d1 & (1 << 5)) != 0){
			d1 = d1 & ~(1 << 5);					// Interlace Disable
			SetSeqReg(0x06, d1);
			SetSiS6326BytePerLine( GetSiS6326BytePerLine() / 2);
		}
	}else{
		d1 = GetSeqReg(0x06);
		if( (d1 & (1 << 5)) == 0){
			d1 = d1 | (1 << 5); 					// Interlace Enable
			SetSeqReg(0x06, d1);
			SetSiS6326BytePerLine( GetSiS6326BytePerLine() * 2);
		}
	}
}

static int GetSiS300BytePerLine(void)
{
	DWORD d0;

	d0	= (DWORD)GetCrtcReg(0x13) & 0xFF;
	d0 |= (DWORD)(GetSeqReg(0x0E) << 8) & 0x0F00;
	d0 = d0 << 3;
	return d0;
}

static void SetSiS300BytePerLine(int d0)
{
	BYTE d1;

	d0 = d0 >> 3;
	d1 = (BYTE)(d0 & 0xFF);
	SetCrtcReg(0x13, d1);
	d1 = GetSeqReg(0x0E) & 0xF0;
	d1 = d1 | ((d0 >> 8) & 0x0F);
	SetSeqReg(0x0E, d1);
}

static void SetSiS300Interlace(int flag)
{
	BYTE d1;

	if(flag == 0){
		d1 = GetSeqReg(0x06);
		if( (d1 & (1 << 5)) != 0){
			d1 = d1 & ~(1 << 5);					// Interlace Disable
			SetSeqReg(0x06, d1);
			SetSiS300BytePerLine( GetSiS300BytePerLine() / 2);
		}
	}else{
		d1 = GetSeqReg(0x06);
		if( (d1 & (1 << 5)) == 0){
			d1 = d1 | (1 << 5); 					// Interlace Enable
			SetSeqReg(0x06, d1);
			SetSiS300BytePerLine( GetSiS300BytePerLine() * 2);
		}
	}
}

static long GetSiS6326VclkParam(long vclk, int *Num, int *Div, int *DeN, int *PS, int *SeqDiv)
{
	double desfreq;
	double temp;
	double tempdt;
	double dt;
	int d0;
	int d1;

	if(vclk > 175000000) vclk = 175000000;
	*SeqDiv = 1;
	*Div	= 1;
	*PS 	= 1;
	if( 	 vclk >= 150000000){
		*PS  = 1;
	}else if(vclk >= 125000000){
		*Div = 2;
		*PS  = 3;
	}else if(vclk >=  75000000){
		*PS  = 2;
	}else if(vclk >=  62500000){
		*PS  = 3;
	}else if(vclk >=  37500000){
		*PS  = 4;
	}else if(vclk >=  31250000){
		*PS  = 6;
	}else if(vclk >=  13500000){
		*PS  = 8;
	}else{
		*PS  = 8;
		*SeqDiv = 2;
	}

	desfreq = vclk * (*PS / (double)(*Div)) * (*SeqDiv);
	dt = 250000000.0;
	*Num = 0;
	*DeN = 0;
	d0 = 2;
	while(d0 < 128){
		d1 = 3;
		while(d1 < 32){
			temp   = 14318000 * (double)d0 / d1;
			tempdt = desfreq - temp;
			if(tempdt < 0) tempdt *= -1;
			if(tempdt < dt){
				*Num = d0;
				*DeN = d1;
				dt = tempdt;
			}
			d1 += 1;
		}
		d0 += 1;
	}
	logerror("Num %3d, DeN %3d, Div %1d, PS %1d, SeqDiv %1d\n",
				*Num, *DeN, *Div, *PS, *SeqDiv);
	logerror("14.318 * (Num/DeN) : %lf\n", 14.318*(*Num)/(*DeN));
	vclk = 14318000 * (*Num * *Div)/(*DeN * *PS * *SeqDiv);
	logerror("vclk : %lf\n", (double)vclk / 1000000.0);

	return vclk;
}
