/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "scrcard.h"
#include "mode15k.h"
#include "vbe3.h"
#include "laguna.h"
#include "s6326.h"
#include "savage.h"
#include "tdfx.h"
#include "ati.h"
#include "s3.h"
#include "cirrus.h"
#include "matrox.h"
/* #include "r128.h" */

#include <stdio.h>
#include <dpmi.h>

/***************************************************************************/
/* Interface */

int GetMode15KHz(CRTCPARAM *cp) {
	unsigned de,bs,rs,re,be,tt;

	cp->bits = card_state.vbe_info_request.BitsPerPixel;

	video_crtc_realize_h(&card_state.crtc_request,&de,&bs,&rs,&re,&be,&tt,1);
	cp->HDisp = de;
	cp->HBStart = bs;
	cp->HSStart = rs;
	cp->HSEnd = re;
	cp->HBEnd = be;
	cp->HTotal = tt;

	video_crtc_realize_v(&card_state.crtc_request,&de,&bs,&rs,&re,&be,&tt,1);
	cp->VDisp = de;
	cp->VBStart = bs;
	cp->VSStart = rs;
	cp->VSEnd = re;
	cp->VBEnd = be;
	cp->VTotal = tt;

	/* VSyncMAME driver requires effective vertical values */
	if (card_state.crtc_request.doublescan) {
		cp->VDisp *= 2;
		cp->VBStart *= 2;
		cp->VSStart *= 2;
		cp->VSEnd *= 2;
		cp->VBEnd *= 2;
		cp->VTotal *= 2;
	}
	if (card_state.crtc_request.interlaced) {
		cp->VDisp /= 2;
		cp->VBStart /= 2;
		cp->VSStart /= 2;
		cp->VSEnd /= 2;
		cp->VBEnd /= 2;
		cp->VTotal /= 2;
	}

	cp->width = card_state.crtc_request.hde;
	cp->height = card_state.crtc_request.vde;
	cp->hpolarity = card_state.crtc_request.hsync;
	cp->vpolarity = card_state.crtc_request.vsync;
	cp->doublescan = card_state.crtc_request.doublescan;
	cp->interlace = card_state.crtc_request.interlaced;
	cp->interlaceratio = 50;

	cp->dotclockHz = card_state.crtc_request.pixelclock;
	cp->horzHz = cp->dotclockHz / cp->HTotal;
	cp->vert001Hz =	100.0 * (double)cp->dotclockHz / (cp->VTotal * cp->HTotal); /* double is required to prevent overflow */

	video_log_modeline_b("card: mode_set", cp->dotclockHz, cp->HDisp, cp->HBStart, cp->HSStart, cp->HSEnd, cp->HBEnd, cp->HTotal,cp->VDisp, cp->VBStart, cp->VSStart, cp->VSEnd, cp->VBEnd, cp->VTotal, cp->hpolarity, cp->vpolarity, cp->doublescan, cp->interlace);
	video_log("card: expected vert clock %.2f\n", (double)cp->vert001Hz / 100);

	return 1;
}

/* Dummy value */
#define horizontal_sync_delay 0
#define horizontal_disp_delay 0

void SetGenericVGARegister(CRTCPARAM *cp)
{
	int d0;
	int d1;
	int HDispSkew;
	int HSyncDelay;

	HSyncDelay = horizontal_sync_delay & 0x03;
	HDispSkew  = horizontal_disp_delay & 0x03;

	/* Unlock */
	SetCrtcReg(0x11, BitMov(GetCrtcReg(0x11), 7, 0, 0));

	/* HTotal */
	d0 = cp->HTotal;
	d0 = (d0 / 8) - 5;
	SetCrtcReg(0x00, d0 & 0xFF);

	/* HDisp */
	d0 = cp->HDisp;
	d0 = (d0 / 8) - 1;
	SetCrtcReg(0x01, d0 & 0xFF);

	/* HBlankStart */
	d0 = cp->HBStart;
	d0 = d0 / 8;
	SetCrtcReg(0x02, d0 & 0xFF);

	/* HBlankEnd */
	d0 = cp->HBEnd;
	d0 = d0 / 8;
	d1 = GetCrtcReg(0x03);
	d1 = BitMov(d1, 0, d0, 0);
	d1 = BitMov(d1, 1, d0, 1);
	d1 = BitMov(d1, 2, d0, 2);
	d1 = BitMov(d1, 3, d0, 3);
	d1 = BitMov(d1, 4, d0, 4);
	d1 = BitMov(d1, 5, HDispSkew, 0);	/* horizontal disp skew */
	d1 = BitMov(d1, 6, HDispSkew, 1);	/* horizontal disp skew */
	d1 = BitMov(d1, 7, 1, 0);			/* compatible read */
	SetCrtcReg(0x03, d1);
	SetCrtcReg(0x05, BitMov(GetCrtcReg(0x05), 7, d0, 5));

	/* HSyncStart */
	d0 = cp->HSStart;
	d0 = d0 / 8;
	SetCrtcReg(0x04, d0 & 0xFF);

	/* HSyncEnd */
	d0 = cp->HSEnd;
	d0 = d0 / 8;
	d1 = GetCrtcReg(0x05);
	d1 = BitMov(d1, 0, d0, 0);
	d1 = BitMov(d1, 1, d0, 1);
	d1 = BitMov(d1, 2, d0, 2);
	d1 = BitMov(d1, 3, d0, 3);
	d1 = BitMov(d1, 4, d0, 4);
	d1 = BitMov(d1, 5, HSyncDelay, 0);
	d1 = BitMov(d1, 6, HSyncDelay, 1);
	SetCrtcReg(0x05, d1);

	/* VTotal */
	d0 = cp->VTotal;
	d0 = d0 - 2;
	SetCrtcReg(0x06, d0 & 0xFF);
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 0, d0, 8));
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 5, d0, 9));

	/* VDisp */
	d0 = cp->VDisp;
	d0 = d0 - 1;
	SetCrtcReg(0x12, d0 & 0xFF);
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 1, d0, 8));
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 6, d0, 9));


	/* VBlankStart */
	d0 = cp->VBStart;
	SetCrtcReg(0x15, d0 & 0xFF);
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 3, d0, 8));
	SetCrtcReg(0x09, BitMov(GetCrtcReg(0x09), 5, d0, 9));

	/* VBlankEnd */
	d0 = cp->VBEnd;
	SetCrtcReg(0x16, d0 & 0xFF);

	/* VSyncStart */
	d0 = cp->VSStart;
	SetCrtcReg(0x10, d0 & 0xFF);
	d1 = GetCrtcReg(0x07);
	d1 = BitMov(d1, 2, d0,	8);  /* bit 8 */
	d1 = BitMov(d1, 7, d0,	9);  /* bit 9 */
	SetCrtcReg(0x07, d1);

	/* VSyncEnd */
	d0 = cp->VSEnd;
	d0 = d0 & 0x0F;
	d1 = GetCrtcReg(0x11) & 0xF0;
	d1 = d1 | d0;
	SetCrtcReg(0x11, d1);

	/* Double Scan Dis. */
	SetCrtcReg(0x09, BitMov(GetCrtcReg(0x09), 7, 0, 0));

}

long CalcDotclock(long Vert001Hz, int HTotal, int VTotal, int interlace)
{
	double f0;

	f0 = (double)VTotal;
	if(interlace != 0){
		f0 += 0.5;
	}
	f0 = f0 * HTotal;
	f0 = f0 * Vert001Hz;
	f0 = f0 / 100.0;

	return (long)f0;
}

long CalcHorizontalHz(long dotclockHz, int HTotal)
{
	return dotclockHz / HTotal;
}

long CalcVert001Hz(long dotclockHz, int HTotal, int VTotal, int interlace)
{
	double f0;

	f0 = (double)VTotal;
	if(interlace != 0){
		f0 += 0.5;
	}
	f0 = f0 * HTotal;
	f0 = dotclockHz / f0;
	f0 = f0 * 100.0;

	return (long)f0;
}

void DispCrtcParam(CRTCPARAM *cp)
{
	(void)cp;
	/* nothing */
}

void SetHVPolarity(int HPolar, int VPolar)
{
	BYTE d0;

	if(HPolar != 0) HPolar = 1;
	if(VPolar != 0) VPolar = 1;

	d0 = _in(0x3CC);
	d0 = d0 & 0x3F;
	d0 = d0 | (VPolar << 7) | (HPolar << 6);
	_out(0x3C2, d0);
}

DWORD BitMov(DWORD dstdata, int dstbitpos, DWORD srcdata, int srcbitpos)
{
	DWORD d0;
	DWORD d1;

	d1 = srcdata >> srcbitpos;
	d1 = d1 & 1;
	d0 = dstdata & ~(1 << dstbitpos);
	d1 = d1 << dstbitpos;
	d0 = d0 | d1;
	return d0;
}

void SetSeqReg(BYTE __idx, BYTE data)
{
	_out(0x3c4, __idx);
	_out(0x3c5, data);
}

BYTE GetSeqReg(BYTE __idx)
{
	BYTE d0;

	_out(0x3c4, __idx);
	d0 = _in(0x3c5);
	return d0;
}

BYTE GetAtrbData(BYTE __idx)
{
	_in(0x3da);
	_out(0x3c0, __idx | (1 << 5));
	return _in(0x3c1);
}

void SetAtrbData(BYTE __idx, BYTE data)
{
	_in(0x3da);
	_out(0x3c0, __idx | 0x20);
	_out(0x3c0, data);
}

int GetCrtcAddress(void)
{
	if((_in(0x3CC) & 0x01) == 0)
		return 0x3B4;
	else
		return 0x3D4;
}

void SetCrtcReg(BYTE __idx, BYTE data)
{
	_out(GetCrtcAddress(),    __idx);
	_out(GetCrtcAddress() + 1, data);
}

BYTE GetCrtcReg(BYTE __idx)
{
	BYTE d0;
	_out(GetCrtcAddress(), __idx);
	d0 = _in(GetCrtcAddress() + 1);
 	return d0;
}

void SetGraphReg(BYTE __idx, BYTE data)
{
	_out(0x3CE,  __idx);
	_out(0x3CE + 1, data);
}

BYTE GetGraphReg(BYTE __idx)
{
	BYTE d0;
	_out(0x3CE, __idx);
	d0 = _in(0x3CE + 1);
 	return d0;
}

int GetSeqDivider(void)
{
	int d0;

	d0	= (GetSeqReg(0x01) & 0x08) >>3;
	d0	= d0 + 1;
	return d0;
}

void SetSeqDivider(BYTE d0) {
	assert( d0 == 1 || d0 == 2 );
	d0 -= 1;
	SetSeqReg(0x01, (GetSeqReg(0x01) & ~0x08) | d0 << 3);
}

void _out(int edx, BYTE al)
{
	outportb(edx, al);
}

BYTE _in(int edx)
{
	return inportb(edx);
}

void _outw(int edx, WORD ax)
{
	outportw(edx, ax);
}

WORD _inw(int edx)
{
	return inportw(edx);
}

void _outl(int edx, DWORD eax)
{
	outportl(edx, eax);
}

DWORD _inl(int edx)
{
	return inportl(edx);
}

void DisableGenerateSignal(void)
{
	BYTE d0;

	d0 = GetCrtcReg(0x17);
	d0 = d0 & 0x7F;
	SetCrtcReg(0x17, d0);
}

void EnableGenerateSignal(void)
{
	BYTE d0;

	d0 = GetCrtcReg(0x17);
	d0 = d0 | 0x80;
	SetCrtcReg(0x17, d0);
}

/* _create_linear_mapping:
 *  Maps a physical address range into linear memory.
 */
int card_create_linear_mapping(unsigned long *linear, unsigned long physaddr, int size)
{
   __dpmi_meminfo meminfo;

   if (physaddr >= 0x100000) {
      /* map into linear memory */
      meminfo.address = physaddr;
      meminfo.size = size;
      if (__dpmi_physical_address_mapping(&meminfo) != 0)
	 return -1;

      *linear = meminfo.address;

      /* lock the linear memory range */
      __dpmi_lock_linear_region(&meminfo);
   }
   else
      /* exploit 1 -> 1 physical to linear mapping in low megabyte */
      *linear = physaddr;

   return 0;
}

/* _remove_linear_mapping:
 *  Frees the DPMI resources being used to map a linear address range.
 */
void card_remove_linear_mapping(unsigned long *linear)
{
   __dpmi_meminfo meminfo;

   if (*linear) {
      if (*linear >= 0x100000) {
	 meminfo.address = *linear;
	 __dpmi_free_physical_address_mapping(&meminfo);
      }

      *linear = 0;
   }
}

/* _create_selector:
 *  Allocates a selector to access a region of linear memory.
 */
int card_create_selector(int *segment, unsigned long linear, int size)
{
   /* allocate an ldt descriptor */
   *segment = __dpmi_allocate_ldt_descriptors(1);
   if (*segment < 0) {
      *segment = 0;
      return -1;
   }

   /* set the descriptor base and limit */
   __dpmi_set_segment_base_address(*segment, linear);

   --size;
   if (size < 0xFFFF)
	size = 0xFFFF;
   __dpmi_set_segment_limit(*segment, size);

   return 0;
}

/* _remove_selector:
 *  Frees a DPMI segment selector.
 */
void card_remove_selector(int *segment)
{
   if (*segment) {
      __dpmi_free_ldt_descriptor(*segment);
      *segment = 0;
   }
}

/***************************************************************************/

card_driver cards[]=
{
	{"laguna", 1, VIDEO_OPTION_VBELINE_DRIVER_LAGUNA, detectLaguna, setLaguna, resetLaguna}, /* Laguna driver */
	{"3dfx", 1, VIDEO_OPTION_VBELINE_DRIVER_3DFX, detect3dfx, set3dfx, reset3dfx}, /* 3dfx driver */
	{"savage", 1, VIDEO_OPTION_VBELINE_DRIVER_SAVAGE, detectSavage, setSavage, resetSavage}, /* Savage driver */
	{"sis6326", 1, VIDEO_OPTION_VBELINE_DRIVER_SIS6326, detectSiS6326, setSiS6326, resetSiS6326}, /* SiS6326 driver */
	{"ati", 1, VIDEO_OPTION_VBELINE_DRIVER_ATI, detectATI, setATI, resetATI} , /* ATI driver */
	{"cirrus", 1, VIDEO_OPTION_VBELINE_DRIVER_CIRRUS, detectCirrus, setCirrus, resetCirrus}, /* Cirrus driver */
	{"matrox", 1, VIDEO_OPTION_VBELINE_DRIVER_MATROX, detectMatrox, setMatrox, resetMatrox}, /* Matrox driver */
	{"s3", 1, VIDEO_OPTION_VBELINE_DRIVER_S3, detectS3, setS3, resetS3}, /* S3 driver */
/*	{"r128", 1, VIDEO_OPTION_VBELINE_DRIVER_R128, detectR128, setR128, resetR128}, */ /* R128 driver */
	{"vbe3", 0, VIDEO_OPTION_VBELINE_DRIVER_VBE3, detectVBE3, setVBE3, resetVBE3}, /* VBE3 driver */
	{0,0,0,0,0,0}
};

void card_init(void) {
	unsigned i;

	assert(	!card_state.active );

	/* VBE (any) is required */
	if (!vbe_active())
		return;

	if (video_option_vbeline_driver_get() == VIDEO_OPTION_VBELINE_DRIVER_NONE)
		return;

	card_state.driver = 0;
	for(i=0;cards[i].name;++i) {
		if (video_option_vbeline_driver_get() == VIDEO_OPTION_VBELINE_DRIVER_AUTO
			|| video_option_vbeline_driver_get() == cards[i].type) {
			if (cards[i].detect && cards[i].detect() > 0) {
				video_log("card: found %s\n",cards[i].name);
				card_state.driver = cards + i;
				break;
			}
		}
	}

	card_state.active = card_state.driver != 0;
}

void card_done(void) {
	assert(	card_active() );
	assert(	!card_mode_active() );

	card_state.driver->reset();

	card_state.active = 0;
}

int card_mode_set(unsigned mode, const video_crtc* crtc) {
	assert(	card_active() );
	assert(	!card_mode_active() );

	card_state.mode_request = mode;
 	card_state.crtc_request = *crtc;

	if (vbe_mode_info_get(&card_state.vbe_info_request, card_state.mode_request)!=0)
		return -1;

	if (video_option_vbeline_clock_multiplier_get()!=1.0) {
		video_log("card: clock multiplier %.2f\n",video_option_vbeline_clock_multiplier_get());
		card_state.crtc_request.pixelclock *= video_option_vbeline_clock_multiplier_get();
	}

	if (card_state.driver->require_vbe) {
		if (vbe_mode_set(card_state.mode_request,0)!=0) {
			return -1;
		}
		if (vbe_scanline_pixel_request(crtc->hde) != 0) {
			vbe_mode_done();
			return -1;
		}
	}

	if (card_state.driver->set() <= 0) {
		video_log("card: error in the video driver mode set\n");
		if (card_state.driver->require_vbe) {
			vbe_mode_done();
			video_error_description_set("Error in the video driver mode set");
		} else {
			/* video_state.error already set */
		}
		return -1;
	}

	if (card_state.driver->require_vbe) {
		vbe_state.mode_info.XResolution = crtc->hde;
		vbe_state.mode_info.YResolution = crtc->vde;
	}

	card_state.mode_active = 1;

	return 0;
}

void card_mode_done(void) {
	assert(	card_active() );
	assert(	card_mode_active() );

	vbe_mode_done();

	card_state.mode_active = 0;
}

#define CARD_PIXELCLOCK_DELTA (50*1000)
#define CARD_PIXELCLOCK_MAX (250*1000*1000)
#define CARD_PIXELCLOCK_MIN (1*1000*1000)

int card_pixelclock_getnext(unsigned* pixelclock, unsigned mode) {
	assert(	card_active() );
	if (card_state.driver->type == VIDEO_OPTION_VBELINE_DRIVER_VBE3) {
		return vbe_pixelclock_getnext(pixelclock,mode);
	} else {
		/* assume all clocks avaliable */
		if (*pixelclock + CARD_PIXELCLOCK_DELTA < CARD_PIXELCLOCK_MAX)
			*pixelclock += CARD_PIXELCLOCK_DELTA;
		return 0;
	}
}

int card_pixelclock_getpred(unsigned* pixelclock, unsigned mode) {
	assert(	card_active() );
	if (card_state.driver->type == VIDEO_OPTION_VBELINE_DRIVER_VBE3) {
		return vbe_pixelclock_getpred(pixelclock,mode);
	} else {
		/* assume all clocks avaliable */
		if (*pixelclock > CARD_PIXELCLOCK_MIN - CARD_PIXELCLOCK_DELTA)
			*pixelclock -= CARD_PIXELCLOCK_DELTA;
		return 0;
	}
}

/* Import information for one video mode
	mode->vbeline video mode
	Return 0 if successful
*/
int card_mode_import(video_mode* mode) {
	vbe_ModeInfoBlock info;

	if (!card_active()) {
		video_error_description_set("VBE CRTC programming isn't avaliable on VBE version prior of 3 or without the dedicated driver");
		return -1;
	}

	if (vbe_mode_info_get(&info, mode->driver.vbeline.mode)!=0) {
		video_error_description_set("VBE report that mode %x isn't supported by your hardware", mode->driver.vbeline.mode);
		return -1;
	}

	if ((mode->driver.vbeline.mode & vbeLinearBuffer) != 0) {
		if (vbe_state.info.VESAVersion < 0x200) {
			video_error_description_set("Linear frame buffer isn't available on VBE version prior 2");
			return -1;
		}
		if ((info.ModeAttributes & vbeMdLinear) == 0) {
			video_error_description_set("Linear frame buffer isn't available in mode %x", mode->driver.vbeline.mode);
			return -1;
		}
	} else {
		if ((info.ModeAttributes & vbeMdNonBanked) != 0) {
			video_error_description_set("Banked frame buffer isn't available in mode %x",mode->driver.vbeline.mode);
			return -1;
		}
	}
	/* 8,15,16 bits per pixel */
	if (info.BitsPerPixel!=8 && info.BitsPerPixel!=15 && info.BitsPerPixel!=16) {
		video_error_description_set("Unsupported bits per pixel size");
		return -1;
	}
	/* Packed or RGB memory model */
	if (info.MemoryModel!=vbeMemPK && info.MemoryModel!=vbeMemRGB) {
		video_error_description_set("Unsupported memory model");
		return -1;
	}
	/* Non planar mode */
	if (info.NumberOfPlanes!=1) {
		video_error_description_set("Unsupported number of planes");
		return -1;
	}

	if (card_state.driver->type == VIDEO_OPTION_VBELINE_DRIVER_VBE3) {
		if (vbe_pixelclock_get(&mode->driver.vbeline.crtc.pixelclock,mode->driver.vbeline.mode)!=0) {
			return -1;
		}
	} else {
		/* assume all clocks avaliable */
	}

	mode->flags = VIDEO_FLAGS_DRIVER_VBELINE |
		VIDEO_FLAGS_ASYNC_SETPAGE |
		VIDEO_FLAGS_TYPE_GRAPHICS |
		(mode->flags & VIDEO_FLAGS_USER_MASK);
	if ((info.ModeAttributes & vbeMdTripleBuffer) != 0)
		mode->flags |= VIDEO_FLAGS_SYNC_SETPAGE;
	if (info.MemoryModel==vbeMemPK)
		mode->flags |= VIDEO_FLAGS_INDEX_PACKED;
	else
		mode->flags |= VIDEO_FLAGS_INDEX_RGB;
	if (mode->driver.vbe.mode & vbeLinearBuffer)
		mode->flags |= VIDEO_FLAGS_LINEAR;

	mode->size_x = mode->driver.vbeline.crtc.hde;
	mode->size_y = mode->driver.vbeline.crtc.vde;
	mode->virtual_x = 0;
	mode->virtual_y = 0;
	mode->vert_freq = (double)mode->driver.vbeline.crtc.pixelclock / (mode->driver.vbeline.crtc.ht * mode->driver.vbeline.crtc.vt);
	if (mode->driver.vbeline.crtc.doublescan)
		mode->vert_freq /= 2;
	if (mode->driver.vbeline.crtc.interlaced)
		mode->vert_freq *= 2;
	mode->horz_freq = (double)mode->driver.vbeline.crtc.pixelclock / mode->driver.vbeline.crtc.ht;
	mode->bits = info.BitsPerPixel;
	mode->aspect_x = 4 * mode->size_y;
	mode->aspect_y = 3 * mode->size_x;
	return 0;
}

