/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#ifndef __SCRCARD_H
#define	__SCRCARD_H

#include "screen.h"

typedef struct card_driver_struct {
	const char* name;
	int require_vbe; /* require a vbe mode set */
	int type; /* VIDEO_VBELINE_DRIVER_* */
	int (*detect)(void);
	int (*set)(void);
	void (*reset)(void);
} card_driver;

typedef struct card_internal_struct {
	int active; /* !=0 if present */
	card_driver* driver;

	int mode_active; /* !=0 if mode set */

	/* requested mode */
	unsigned mode_request;
	video_crtc crtc_request;
	vbe_ModeInfoBlock vbe_info_request;
} card_internal;

card_internal card_state;

static __inline__ int card_active(void) {
	return card_state.active != 0;
}

static __inline__ int card_mode_active(void) {
	return card_state.mode_active != 0;
}

static __inline__ const char* card_driver_name(void) {
	return card_state.driver->name;
}

int card_pixelclock_getnext(unsigned* pixelclock, unsigned mode);
int card_pixelclock_getpred(unsigned* pixelclock, unsigned mode);

int card_mode_import(video_mode* mode);

void card_init(void);
void card_done(void);
int card_mode_set(unsigned mode, const video_crtc* crtc);
void card_mode_done(void);

#endif
