/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "screen.h"

#include <assert.h>

/* Position for drawing */
static unsigned video_update_offset;

/* Screen in drawing */
static unsigned _video_update_page;

/* Draw possible (boolean) */
static int video_update_draw_allowed;

/* Pages used */
static unsigned _video_update_page_max;

/* Inizialize */
void video_update_init(int max_buffer) {
	unsigned pages = (video_virtual_y() * video_bytes_per_scanline()) / video_bytes_per_page();
	if (pages > max_buffer)
		pages = max_buffer;
	_video_update_page = 0;
	if (pages >= 3 && (video_flags() & VIDEO_FLAGS_SYNC_SETPAGE)) {
		_video_update_page_max = 3;
	} else if (pages >= 2 && (video_flags() & VIDEO_FLAGS_ASYNC_SETPAGE)) {
		_video_update_page_max = 2;
	} else {
		_video_update_page_max = 1;
	}
	video_update_draw_allowed = 0;
}

/* Return the number of pages used */
unsigned video_update_page_max(void) {
	return _video_update_page_max;
}

/* Deinizialize */
void video_update_done(void) {
	assert( video_update_draw_allowed == 0 );

	/* display the first page */
	if (_video_update_page_max!=1) {
		video_display_set_async(0,0);
	}
}

/* Relative position of the page in drawing */
unsigned video_update_x(void) {
	unsigned x;
	assert( video_update_draw_allowed );
	x = (video_update_offset % video_bytes_per_scanline()) / video_bytes_per_pixel();
	if (video_unchained_active())
		x *= 4;
	return x;
}

unsigned video_update_y(void) {
	assert( video_update_draw_allowed );
	return video_update_offset / video_bytes_per_scanline();
}

/* Page in drawing */
unsigned video_update_page(void) {
	assert( video_update_draw_allowed );
	return _video_update_page;
}

/* Start drawing to page */
void video_update_start(void) {
	assert(	video_update_draw_allowed == 0 );
	video_update_draw_allowed = 1;
	/* compute coordinate for drawing */
	if (_video_update_page_max!=1) {
		video_update_offset = _video_update_page * video_bytes_per_page();
	} else {
		video_update_offset = 0;
	}
}

/* End drawing to page */
void video_update_stop(int wait_retrace) {
	assert(	video_update_draw_allowed );
	video_update_draw_allowed = 0;

	if (_video_update_page_max!=1) {
		unsigned offset;
		offset = _video_update_page * video_bytes_per_page();
		if (_video_update_page_max == 3)
			video_display_set_sync(offset);
		else {
			video_display_set_async(offset,wait_retrace);
		}

		/* next page to draw */
		++_video_update_page;
		if (_video_update_page == _video_update_page_max)
			_video_update_page = 0;
	} else {
		if (wait_retrace)
			video_wait_vsync();
	}
}


