/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "scrvga.h"

#include <stdlib.h>
#include <assert.h>
#include <dpmi.h>
#include <go32.h>
#include <dos.h>
#include <pc.h>
#include <time.h>
#include <sys/movedata.h>
#include <sys/farptr.h>

vga_internal vga_state;

#define VGA_REGISTER_MAX 25

/***************************************************************************/
/* Mode */

/* Reprogram the vga card for support the font selection
 * note:
 *   This is important beacause many vga2tv TSR reprogram
 *   also the vga in the int 10 call for the font selection
 */
static void vga_font_realize(unsigned rows, const tweak_register* regs, unsigned regs_num) {
	assert( vga_active() && vga_mode_active() && !vga_state.mode_info.is_graphics_mode);

	tweak_reg_array_out(regs,regs_num);

	/* set font rows */
	tweak_char_size_y_set(rows,tweak_reg_write,tweak_reg_read,0);

	/* update info */
	tweak_crtc_get(&vga_state.mode_info,tweak_reg_read,0);
}

int vga_font_set_default8(void) {
	tweak_register regs[VGA_REGISTER_MAX];
	unsigned regs_num;
	unsigned mode;
	__dpmi_regs r;

	if (vga_mode_grab(&mode,regs,&regs_num,VGA_REGISTER_MAX) != 0)
		return -1;

	r.x.ax = 0x1102;
	r.x.bx = 0;
	__dpmi_int(0x10, &r);

	vga_font_realize(8,regs,regs_num);

	return 0;
}

int vga_font_set_default14(void) {
	tweak_register regs[VGA_REGISTER_MAX];
	unsigned regs_num;
	unsigned mode;
	__dpmi_regs r;

	if (vga_mode_grab(&mode,regs,&regs_num,VGA_REGISTER_MAX) != 0)
		return -1;

	r.x.ax = 0x1101;
	r.x.bx = 0;
	__dpmi_int(0x10, &r);

	vga_font_realize(14,regs,regs_num);

	return 0;
}

int vga_font_set_default16(void) {
	tweak_register regs[VGA_REGISTER_MAX];
	unsigned regs_num;
	unsigned mode;
	__dpmi_regs r;

	if (vga_mode_grab(&mode,regs,&regs_num,VGA_REGISTER_MAX) != 0)
		return -1;

	r.x.ax = 0x1104;
	r.x.bx = 0;
	__dpmi_int(0x10, &r);

	vga_font_realize(16,regs,regs_num);

	return 0;
}

int vga_font_set(void* font, unsigned rows) {
	tweak_register regs[VGA_REGISTER_MAX];
	unsigned regs_num;
	unsigned mode;
	__dpmi_regs r;

	if (vga_mode_grab(&mode,regs,&regs_num,VGA_REGISTER_MAX) != 0)
		return -1;

	r.x.ax = 0x1100;
	r.x.bx = rows << 8;
	r.x.cx = 256;
	r.x.dx = 0;

	dosmemput(font, 256*rows, __tb);
	r.x.es = __tb / 16;
	r.x.bp = 0;

	__dpmi_int(0x10, &r);

	vga_font_realize(rows,regs,regs_num);

	return 0;
}

struct vga_register {
	unsigned port;
	unsigned index;
} VGA_REGISTER[] = {
	{ 0x3c2, 0x0 },
	{ 0x3d4, 0x0 },
	{ 0x3d4, 0x1 },
	{ 0x3d4, 0x2 },
	{ 0x3d4, 0x3 },
	{ 0x3d4, 0x4 },
	{ 0x3d4, 0x5 },
	{ 0x3d4, 0x6 },
	{ 0x3d4, 0x7 },
	{ 0x3d4, 0x8 },
	{ 0x3d4, 0x9 },
	{ 0x3d4, 0x10 },
	{ 0x3d4, 0x11 },
	{ 0x3d4, 0x12 },
	{ 0x3d4, 0x13 },
	{ 0x3d4, 0x14 },
	{ 0x3d4, 0x15 },
	{ 0x3d4, 0x16 },
	{ 0x3d4, 0x17 },
	{ 0x3c4, 0x1 },
	{ 0x3c4, 0x4 },
	{ 0x3ce, 0x5 },
	{ 0x3ce, 0x6 },
	{ 0x3c0, 0x10 },
	{ 0x3c0, 0x13 },
	{ 0,0 }
};

int vga_mode_grab(unsigned* mode, tweak_register* regs, unsigned* regs_num, unsigned regs_max) {
	unsigned i;

	/* get the video mode */
	if (vga_mode_get(mode) != 0)
		return -1;

	for(i=0;VGA_REGISTER[i].port;++i) {
		if (i>=regs_max)
			return -1;
		regs[i].port = VGA_REGISTER[i].port;
		regs[i].index = VGA_REGISTER[i].index;
		regs[i].value = tweak_reg_in(regs[i].port, regs[i].index);
	}

	*regs_num = i;

	return 0;
}

/* Set the current video mode */
int vga_mode_set(unsigned mode, const tweak_register* regs, unsigned regs_num) {
	__dpmi_regs r;

	assert( vga_active() );

	r.x.ax = mode & 0xFF;
	__dpmi_int(0x10, &r);

	vga_state.mode = mode & 0xFF;

	if (vga_state.mode == 0x6A) {
		/* convert the mode from 800x600 16 color to 400x300 256 color */
		int c;

		/* disable chain-4 */
		outportw(0x3C4, 0x0604);

		/* set 256 color mode, write mode 0 */
		outportb(0x3CE, 0x05);
		outportb(0x3CF, (inportb(0x3CF) & 0x60) | 0x40);

		/* reset the palette */
		outportb(0x3C0, 0x30);
		outportb(0x3C0, inportb(0x3C1)|0x40);
		for (c=0; c<16; c++) {
			outportb(0x3C0, c);
			outportb(0x3C0, c);
		}
		outportb(0x3C0, 0x20);

		/* set the char_y and clear the doublescan */
		outportb(0x3D4, 0x09);
		outportb(0x3D5, (inportb(0x3D5)&0x60) | 1);
	}

	/* set VGA register */
	if (regs && regs_num) {
		tweak_reg_array_out(regs,regs_num);
	}

	/* get info */
	tweak_crtc_get(&vga_state.mode_info,tweak_reg_read,0);

	vga_state.unchained_active = !vga_state.mode_info.is_linear;
	vga_state.unchained_plane_mask = 0;

	/* enable */
	vga_state.mode_active = 1;

	return 0;
}

/* Get the current video mode */
int vga_mode_get(unsigned* mode) {
	__dpmi_regs r;

	assert( vga_active() );

	/* read video mode */
	r.x.ax = 0x0F00;
	__dpmi_int(0x10, &r);

	*mode =	r.x.ax & 0x7F; /* clear unused bit */

	return 0;
}

void vga_mode_done(void) {
	assert( vga_active() && vga_mode_active() );

	/* disable */
	vga_state.mode_active = 0;
}

/***************************************************************************/
/* VGA */

static unsigned vga_wait_vsync_tune(void);
static unsigned wait_counter_limit = -1; /* counter used to limit the wait_vsync */

/* Inizialize the vga system
 * return:
 *	0 on successfull
 */
int vga_init(void) {
	assert( !vga_active() );

	/* inizialize with all zero */
	memset(&vga_state, 0, sizeof(vga_state) );

	/* enable */
	vga_state.active = 1;

	wait_counter_limit = 0xFFFFFFFFUL;
	vga_wait_vsync_tune();
	wait_counter_limit = vga_wait_vsync_tune() * 10;

	return 0;
}

/* Deinizialize the vga system */
void vga_done(void) {
	assert( vga_active() );

	if (vga_mode_active()) {
		vga_mode_set(0x3,0,0);
	}

	/* disable */
	vga_state.active = 0;
}

/* Return the selector for accessing in writing the video memory */
unsigned FAST vga_write_selector(void) {
	return _dos_ds;
}

/* Return the offset for accessing in writing the video memory */
unsigned FAST vga_write_line(unsigned y) {
	assert( vga_mode_active() );
	if (!vga_state.mode_info.is_graphics_mode)
		return 0xB8000 + y * vga_state.mode_info.bytes_per_scanline;
	else
		return 0xA0000 + y * vga_state.mode_info.bytes_per_scanline;
}

/* Wait a vsync */
#if 0
void vga_wait_vsync(void) {
	int enabled = disable();
	while ((inportb(0x03da) & 0x08) != 0);
	while ((inportb(0x03da) & 0x08) == 0);
	if (enabled)
		enable();
}
#else
static unsigned vga_wait_vsync_tune(void) {
	/* Prevent infinite loop, don't disable int */
	unsigned counter = 0;
	while ((inportb(0x03da) & 0x08) != 0 && counter < wait_counter_limit)
		++counter;
	while ((inportb(0x03da) & 0x08) == 0 && counter < wait_counter_limit)
		++counter;
	return counter;
}

void vga_wait_vsync(void) {
	vga_wait_vsync_tune();
}
#endif


int vga_scroll(unsigned offset, int waitvsync) {
	assert( vga_active() && vga_mode_active() );

	if (waitvsync)
		vga_wait_vsync();

  	tweak_reg_out(offset >> 8, TWEAK_CRTC_ADDR, 0xC);
	tweak_reg_out(offset & 0xFF, TWEAK_CRTC_ADDR, 0xD);

	return 0;
}

static unsigned vga_dotclock[] = {
	TWEAK_DOTCLOCK_LOW/2,
	TWEAK_DOTCLOCK_HIGH/2,
	TWEAK_DOTCLOCK_LOW,
	TWEAK_DOTCLOCK_HIGH,
	0
};

int vga_dotclock_get(unsigned* dotclock) {
	unsigned best;
	int best_diff;
	unsigned i;

	best = 0;
	best_diff = abs(vga_dotclock[0] - *dotclock);
	for(i=1;vga_dotclock[i];++i) {
		int diff = abs(vga_dotclock[i] - *dotclock);
		if (diff < best_diff) {
			best = i;
			best_diff = diff;
		}
	}

	*dotclock = vga_dotclock[best];

	return 0;
}

static int vga_clock_multiplier(unsigned* multiplier, unsigned mode) {
	switch (mode) {
		case 0 :
		case 1 :
		case 2 :
		case 3 :
			*multiplier = 1;
		break;
		case 0x13 :
			*multiplier = 2;
		break;
		default:
			video_error_description_set("Invalid vga mode %x in vga_clock_multiplier\n",mode);
			return -1;
	}
	return 0;
}

/* Fake pixel clock for mode 6A, 0 or (ht*vt*doublescan*vert_freq) */
/* #define VGA_MODE_6A_PIXELCLOCK (528*314*2*60) */
#define VGA_MODE_6A_PIXELCLOCK (0)

int vga_pixelclock_get(unsigned* pixelclock, unsigned mode) {
	unsigned multiplier;

	if (mode == 0x6A) {
		*pixelclock = VGA_MODE_6A_PIXELCLOCK; /* fake pixel clock */
		return 0;
	}

	if (vga_clock_multiplier(&multiplier,mode)!=0)
		return -1;
	*pixelclock *= multiplier;
	if (vga_dotclock_get(pixelclock) !=0 )
		return -1;

	*pixelclock /= multiplier;

	return 0;
}

/* Return the next avaliable pixel clock */
int vga_pixelclock_getnext(unsigned* pixelclock, unsigned mode) {
	unsigned i;
	unsigned multiplier;

	if (mode == 0x6A) {
		*pixelclock = VGA_MODE_6A_PIXELCLOCK; /* fake pixel clock */
		return 0;
	}

	if (vga_clock_multiplier(&multiplier,mode)!=0)
		return -1;
	*pixelclock *= multiplier;

	i = 0;
	while (vga_dotclock[i]!=0 && vga_dotclock[i]<=*pixelclock)
		++i;
	if (!vga_dotclock[i])
		--i;

	*pixelclock = vga_dotclock[i];
	*pixelclock /= multiplier;

	return 0;
}

/* Return the pred avaliable pixel clock */
int vga_pixelclock_getpred(unsigned* pixelclock, unsigned mode) {
	unsigned i;
	unsigned multiplier;

	if (mode == 0x6A) {
		*pixelclock = VGA_MODE_6A_PIXELCLOCK; /* fake pixel clock */
		return 0;
	}

	if (vga_clock_multiplier(&multiplier,mode)!=0)
		return -1;
	*pixelclock *= multiplier;

	i = 0;
	while (vga_dotclock[i]!=0 && vga_dotclock[i]<*pixelclock)
		++i;
	if (i)
		--i;

	*pixelclock = vga_dotclock[i];
	*pixelclock /= multiplier;

	return 0;
}

/* Set palette
 * in:
 *   palette palette to set in 6 bit format
 */
void vga_palette6_set(video_color* palette, unsigned start, unsigned count, int waitvsync) {
	if (waitvsync)
		vga_wait_vsync();

	outportb(0x3C8, start);
	__asm__ __volatile__ (
		"std ;" /* decrement esi */
		"addl $2,%0 ;"
		"0: ;"
		"outsb ;" /* r */
		"outsb ;" /* g */
		"outsb ;" /* b */
		"addl $7,%0 ;"
		"decl %1 ;"
		"jnz 0b ;" /* local label, b==backward */
		: /* no output */
		:
		"S" (palette),
		"r" (count),
		"d" (0x3C9)
		:
		"eax", "cc"
	);
}

/* Set palette
 * in:
 *   palette palette to set in 8 bit format
 */
void vga_palette8_set(video_color* palette, unsigned start, unsigned count, int waitvsync) {
 	if (waitvsync)
		vga_wait_vsync();
	
	outportb(0x3C8, start);
	while (count) {
		outportb(0x3C9, palette->red >> 2);
		outportb(0x3C9, palette->green >> 2);
		outportb(0x3C9, palette->blue >> 2);
		--count;
		++palette;
	}
}


