/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 *
 * This video driver is STRONGLY derived from the video drivers of the
 * VSyncMAME project by S.Sakamaki (Saka)
 */

#include <pc.h>
#include <conio.h>
#include <sys/farptr.h>
#include <go32.h>
#include <time.h>
#include <math.h>
#include <dpmi.h>
#include "tdfx.h"
#include "mode15k.h"

static void Set3dfxScreenSize(CRTCPARAM *cp);
static void Set3dfxDacMode(int flag);
static void Set3dfxExtRegister(CRTCPARAM *cp);
static int Get3dfxBytePerLine(void);
static void Set3dfxBytePerLine(int d0);
static void Set3dfxInterlace(int flag, CRTCPARAM* cp);
static void Set3dfxVClock(int Num, int DeN, int PS);
static long Get3dfxVclkParam(long vclk, int *Num, int *DeN, int *PS);
static void SetDoubleScan(int flag);
static void SelectPLL(void);

static int g_interlace_status = 0;
static int _3dfxiobase = 0;

int detect3dfx(void)
{
	__dpmi_regs 	r;

	r.d.eax = 0x0000b101;						// PCI BIOS - INSTALLATION CHECK
	r.d.edi = 0x00000000;
	__dpmi_int(0x1a, &r);
	if( r.d.edx != 0x20494350 ){				// ' ICP'
		logerror("15.75KHz: PCI BIOS not Installed.\n");
		return 0;
	}

	while(1){
		r.d.eax = 0x0000b102;						// PCI BIOS - FIND PCI DEVICE
		r.d.ecx = 0x00000005;						// device ID
		r.d.edx = 0x0000121A;						// vendor ID
		r.d.esi = 0x00000000;						// device index
		__dpmi_int(0x1a, &r);
		if( r.h.ah == 0 ){
			logerror("Voodoo3 ID5 found\n");
			break;
		}
		r.d.eax = 0x0000b102;						// PCI BIOS - FIND PCI DEVICE
		r.d.ecx = 0x00000004;						// device ID
		r.d.edx = 0x0000121A;						// vendor ID
		r.d.esi = 0x00000000;						// device index
		__dpmi_int(0x1a, &r);
		if( r.h.ah == 0 ){
			logerror("Voodoo3 ID4 found\n");
			break;
		}
		r.d.eax = 0x0000b102;						// PCI BIOS - FIND PCI DEVICE
		r.d.ecx = 0x00000003;						// device ID
		r.d.edx = 0x0000121A;						// vendor ID
		r.d.esi = 0x00000000;						// device index
		__dpmi_int(0x1a, &r);
		if( r.h.ah == 0 ){
			logerror("Banshee found\n");
			break;
		}
		logerror("3dfx: not found\n");
		return 0;
	}

	r.d.eax = 0x0000b10a;						// READ CONFIGURATION DWORD
												// BH = bus number
												// BL = device/function number
	r.d.edi = 0x00000018;						// register number
	__dpmi_int(0x1a, &r);
	if( r.h.ah != 0 ){
		logerror("3dfx: Read Configuration DWORD error.\n");
		return -1;
	}

	_3dfxiobase = r.d.ecx & 0xFFFFFF00;

	logerror("3dfx: iobase %X\n", _3dfxiobase);

#ifdef MAME
	DisableStretch();
#endif

	return 1;
}

void reset3dfx(void)
{
}

int set3dfx(void)
{
	CRTCPARAM cp;
	int Num;
	int DeN;
	int PS;
	DWORD d0;

	if(GetMode15KHz(&cp) == 0) return 0;
	cp.dotclockHz = Get3dfxVclkParam(cp.dotclockHz, &Num, &DeN, &PS);
	cp.vert001Hz  = CalcVert001Hz(cp.dotclockHz, cp.HTotal,
								  cp.VTotal, cp.interlace);
	cp.horzHz	  = CalcHorizontalHz(cp.dotclockHz, cp.HTotal);

	cp.HBStart = cp.HSStart;
	cp.HBEnd   = cp.HTotal - 8;
	cp.VBStart = cp.VDisp;
	cp.VBEnd   = cp.VTotal - 1;

	DispCrtcParam(&cp);
	d0 =  _inl(_3dfxiobase + 0x28);
	d0 = d0 | 0x40;
	_outl(_3dfxiobase + 0x28, d0);		// activate vga ext register
	d0 =  _inl(_3dfxiobase + 0x2C);
	d0 = d0 & ~0x1FE00000;
	_outl(_3dfxiobase + 0x2C, d0);		// unlock vga register

	DisableGenerateSignal();
	SetGenericVGARegister(&cp);
	Set3dfxExtRegister(&cp);
	Set3dfxDacMode(0);
	Set3dfxVClock(Num, DeN, PS);
	Set3dfxScreenSize(&cp);

	EnableGenerateSignal();

	return 1;
}

static void Set3dfxDacMode(int flag)
{
	DWORD d0;

	flag = (flag != 0) ? 1 : 0;
	d0 =  _inl(_3dfxiobase + 0x4C) | flag;
	_outl(_3dfxiobase + 0x4C, d0);
}

static void Set3dfxScreenSize(CRTCPARAM *cp)
{
	int width;
	int height;
	DWORD d0;
	int d1;

	width  = cp->width;
	height = cp->height;
	if(cp->doublescan != 0)
		height *= 2;

	d0 = width | (height << 12);
	_outl(_3dfxiobase + 0x98, d0);
	d0 =  _inl(_3dfxiobase + 0x98);

//	  logerror("3dfx: %dx%d\n", d0 & 0x0FFF, (d0 >> 12) & 0x0FFF);

	d0 =  _inl(_3dfxiobase + 0x5C);
	d0 = d0 & ~0x01;
	_outl(_3dfxiobase + 0x5C, d0);			// off
	d1 = 0;
	while(d1 < 2048){
		d0 =  _inl(_3dfxiobase + 0x5C);
		d1 += 1;
	}
	d0 = d0 | 1;
	_outl(_3dfxiobase + 0x5C, d0);			// on
}

static void SelectPLL(void)
{
	BYTE d0;

	d0 = _in(0x3CC);
	d0 = d0 | 0x0C;
	_out(0x3C2, d0);
}

void Set3dfxExtRegister(CRTCPARAM *cp)
{
	DWORD d0;

	d0 = 0;
	d0 = BitMov(d0, 0, (cp->HTotal	>> 3) - 5, 8);
	d0 = BitMov(d0, 2, (cp->HDisp	>> 3) - 1, 8);
	d0 = BitMov(d0, 4, (cp->HBStart >> 3)	 , 8);
	d0 = BitMov(d0, 5, (cp->HBEnd	>> 3)	 , 6);
	d0 = BitMov(d0, 6, (cp->HSStart >> 3)	 , 8);
	d0 = BitMov(d0, 7, (cp->HSEnd	>> 3)	 , 5);
	SetCrtcReg(0x1A, d0);						// ext horz reg

	d0 = 0;
	d0 = BitMov(d0, 0, cp->VTotal  - 2, 10);
	d0 = BitMov(d0, 2, cp->VDisp   - 1, 10);
	d0 = BitMov(d0, 4, cp->VBStart, 	10);
	d0 = BitMov(d0, 6, cp->VSStart, 	10);
	SetCrtcReg(0x1B, d0);						// ext vert reg


	SetDoubleScan(cp->doublescan);
	Set3dfxInterlace(cp->interlace, cp);
	SetHVPolarity(cp->hpolarity, cp->vpolarity);
}

static int Get3dfxBytePerLine(void)
{
	DWORD d0;

	d0 = _inl(_3dfxiobase + 0xE8) & 0x7FFF;

	return (int)d0;
}

static void Set3dfxBytePerLine(int d0)
{
	_outl(_3dfxiobase + 0xE8, d0 & 0x7FFF);
}

static void SetDoubleScan(int flag)
{
	DWORD d0;

	if(flag == 0){
		d0 =  _inl(_3dfxiobase + 0x5C) & ~0x10;
		_outl(_3dfxiobase + 0x5C, d0);
	}else{
		d0 =  _inl(_3dfxiobase + 0x5C) | 0x10;
		_outl(_3dfxiobase + 0x5C, d0);
	}
}

static void Set3dfxInterlace(int flag, CRTCPARAM *cp)
{
	if(flag == 0){
		if(g_interlace_status != 0){
			Set3dfxBytePerLine( Get3dfxBytePerLine() / 2 );
			g_interlace_status = 0;
		}
	}else{
		if(g_interlace_status == 0){
			Set3dfxBytePerLine( Get3dfxBytePerLine() * 2 );
			g_interlace_status = 1;
		}
	}
}

static long Get3dfxVclkParam(long vclk, int *Num, int *DeN, int *PS)
{
	double desfreq;
	double temp;
	double tempdt;
	double dt;
	int d0;
	int d1;

	if(vclk > 135000000) vclk = 135000000;
	*PS 	= 1;
	if( 	 vclk >=  75000000){
		*PS  = 1;
	}else if(vclk >=  37500000){
		*PS  = 2;
	}else if(vclk >=  13500000){
		*PS  = 4;
	}else{
		*PS  = 8;
	}

	desfreq = vclk * (*PS);
	dt = 250000000.0;
	*Num = 0;
	*DeN = 0;
	d0 = 3;
	while(d0 < 258){
		d1 = 3;
		while(d1 < 66){
			temp   = 14318000 * (double)d0 / d1;
			tempdt = desfreq - temp;
			if(tempdt < 0) tempdt *= -1;
			if(tempdt < dt){
				*Num = d0;
				*DeN = d1;
				dt = tempdt;
			}
			d1 += 1;
		}
		d0 += 1;
	}
	logerror("Num %3d, DeN %3d, PS %1d\n", *Num, *DeN, *PS);
	logerror("14.318 * (Num/DeN) : %lf\n", 14.318 *((*Num)/(*DeN)));
	logerror("14.318 * (Num/DeN)/PS : %lf\n", 14.318 * ((double)(*Num)/(*DeN))
													/ (*PS));
	vclk = 14318000 * ((double)(*Num)/(*DeN) / (*PS));
	logerror("vclk : %lf\n", (double)vclk / 1000000.0);

	return vclk;
}

static void Set3dfxVClock(int Num, int DeN, int PS)
{
	DWORD _PS;
	DWORD d0;

	SelectPLL();

	_PS = 0;
	while(1){
		PS = PS >> 1;
		if(PS == 0) break;
		_PS += 1;
	}

	d0	= _PS;
	d0 |= (((DWORD)DeN - 2) & 0x3F) << 2;
	d0 |= (((DWORD)Num - 2) & 0xFF) << 8;
	_outl(_3dfxiobase + 0x40, d0);
}
